package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkArgument;
import static java.lang.Math.PI;
import static java.lang.Math.cos;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.sqrt;

import java.util.EnumSet;
import java.util.Set;
import java.util.function.ToDoubleBiFunction;

import com.google.common.collect.ArrayTable;
import com.google.common.collect.ImmutableTable;
import com.google.common.collect.Table;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.gmm.Imt;

/**
 * Conditional spectrum implementation.
 *
 * @author U.S. Geological Survey
 */
class ConditionalSpectrum {

  interface ImtCorrelation extends ToDoubleBiFunction<Imt, Imt> {}

  /*
   * Method 4: "Exact" CS with disaggregation weights.
   *
   * Result: Map<Imt, XySequence> (x=T, y=μ) Map<Imt, XySequence> (x=T, y=σ)
   *
   * Given target IMT T* and conditional Imt Tj
   *
   * For each source/gmm, we need epsilon at the target IMT, Maths.epsilon(...)
   * which we can obtain from the disaggregation
   *
   * Primary loop is target, iterate over all other SAs including Tj = T* (shoud
   * just yield μCond = μ + ε and σ = 0.0
   */

  static final Table<Imt, Imt, Double> ρTable = initCorrelations();

  /*
   * Create a table of SA IMT correlations. Table contains duplicate values as
   * all possible ordered pairs (e.g. Imt1:Imt2 and Imt2:Imt1) are included.
   */
  static Table<Imt, Imt, Double> initCorrelations() {
    BakerJayaram08 bj08 = new BakerJayaram08();
    Set<Imt> saImts = EnumSet.copyOf(Imt.mprsImts());
    saImts.remove(Imt.PGA);
    Table<Imt, Imt, Double> table = ArrayTable.create(saImts, saImts);
    for (Imt imt1 : saImts) {
      saImts.stream().forEach(imt2 -> table.put(
          imt1,
          imt2,
          Maths.round(bj08.applyAsDouble(imt1, imt2), 8)));
    }
    return ImmutableTable.copyOf(table);
  }

  /*
   * Compute the mean at the conditional IMT from the conditional SA mean, μ,
   * the conditional SA sigma, σ, the IMT correlation coefficient, ρ, and
   * epsilon, ε, at the target IMT.
   */
  private static double conditionalMean(double μ, double σ, double ρ, double ε) {
    return μ + ρ * ε * σ;
  }

  /**
   * Baker and Jayaram (2008) intensity measure correlation model for spectral
   * accelerations between 0.01 and 10s.
   *
   * <p><b>Reference:</b> J. W. Baker and N. Jayaram. (2008). Correlation of
   * spectral acceleration values from NGA ground motion models, Earthquake
   * Spectra, 24: 1, 299-317.
   */
  static class BakerJayaram08 implements ImtCorrelation {

    /**
     * Computes the correlation coefficient between two spectral acceleration
     * IMTs. The order of the two IMTs does not matter.
     *
     * @param imt1 the first IMT
     * @param imt2 the second IMT
     * @return pearson correlation coefficient between {@code imt1} and
     *         {@code imt2}
     */
    @Override
    public double applyAsDouble(Imt imt1, Imt imt2) {

      checkArgument(imt1.isSA());
      checkArgument(imt2.isSA());

      double t1 = imt1.period();
      double t2 = imt2.period();

      double tMin = min(t1, t2);
      double tMax = max(t1, t2);

      double c1 = (1.0 - cos(PI / 2.0 - log(tMax / max(tMin, 0.109)) * 0.366));
      double c2 = (tMax < 0.2)
          ? 1.0 - 0.105 * (1.0 - 1.0 / (1.0 + exp(100.0 * tMax - 5.0))) *
              (tMax - tMin) / (tMax - 0.0099)
          : Double.NaN;
      double c3 = (tMax < 0.109) ? c2 : c1;
      double c4 = c1 + 0.5 * (sqrt(c3) - c3) * (1.0 + cos(PI * tMin / 0.109));

      if (tMax <= 0.109) {
        return c2;
      } else if (tMin > 0.109) {
        return c1;
      } else if (tMax < 0.2) {
        return min(c2, c4);
      } else {
        return c4;
      }
    }
  }

  public static void main(String[] args) {
    BakerJayaram08 bj = new BakerJayaram08();
    // System.out.println(ρTable);
    System.out.println(bj.applyAsDouble(Imt.SA0P02, Imt.SA0P02));
  }

}
