package gov.usgs.earthquake.nshmp.calc;

import static gov.usgs.earthquake.nshmp.Text.NEWLINE;
import static gov.usgs.earthquake.nshmp.calc.DisaggDataset.SOURCE_CONSOLIDATOR;
import static gov.usgs.earthquake.nshmp.calc.DisaggDataset.SOURCE_SET_CONSOLIDATOR;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;
import java.util.function.Function;

import com.google.common.collect.ListMultimap;
import com.google.common.collect.Maps;
import com.google.common.collect.MultimapBuilder;
import com.google.common.collect.Multimaps;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListenableFuture;

import gov.usgs.earthquake.nshmp.data.Interpolator;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.model.SourceType;

/**
 * Hazard disaggregation. Given a {@link Hazard} result, this class will
 * disaggregate the results at all spectral periods supplied in the result at an
 * intensity measure level or return period of interest.
 *
 * @author U.S. Geological Survey
 */
public final class Disaggregation {

  /*
   * Developer notes
   * -------------------------------------------------------------------------
   * Auto-scaling of results (dataset bounds) based on hazard model.
   * -------------------------------------------------------------------------
   * Warnings if config does not span source set range.
   * -------------------------------------------------------------------------
   * Disaggregate on probability of occurrence instead of exceedance.
   * -------------------------------------------------------------------------
   * Revisit precision issues associated with integer based return period;
   * 2%in50 years is really 0.00040405414, not 1/2475 = 0.0004040404
   * -------------------------------------------------------------------------
   * One of the difficulties with disaggregation is deciding how to specify
   * magnitude and distance ranges, and respective discretizations, over which
   * to disaggregate, given the broad ranges of distances and magnitudes
   * supported by various models (e.g. 300km in the WUS NSHMP vs 1000km in the
   * CEUS). Invariably, situations will arise where some sources are outside a
   * user-specified range and the sum of contributions in a disaggregation data
   * matrix will not equal the target rate specified at the outset of a
   * calculation. One could query the model(s) being used (set broad limits) or
   * create lists of sources and their contributions in advance before building
   * disaggregation datasets (set calculation specific limits). The former may
   * make sense as a default setting in the absence of any user specified
   * settings, the latter complicates the code considerably.
   *
   * For the time being we require user-specified ranges and encourage high
   * resolution disaggregation data bins that can be preserved if the
   * contributing sources span only a small part of a disaggregation result
   * matrix. If the disaggregation result matrix is heavily populated, bins
   * could be consolidated prior to output.
   *
   * For data outside the defined ranges, we track the 'un-binned' or residual
   * rate. This is needed to compute mean r, m, and ε.
   *
   * Note that in a webservice environment, only relevant data will be returned
   * (zero-contribution bins are omitted) and the client will render plots based
   * on the data supplied, not based on the ranges specified for the calculation
   * itelf.
   * -------------------------------------------------------------------------
   * Issues related to disaggreagtion targets.
   *
   * Because hazard is computed at specific intensity measure levels, only when
   * a disaggregation is computed at one of those levels will the contributions
   * of the relevant sources equal the target rate specified by the total mean
   * hazard curve. Because of the convexity of the hazard curve in log space,
   * the 'true' total as derived from the relevant sources will be slightly
   * higher.
   */

  final Map<Imt, ImtDisagg> disaggs;
  final Site site;

  private Disaggregation(Map<Imt, ImtDisagg> disaggs, Site site) {
    this.disaggs = disaggs;
    this.site = site;
  }

  /**
   * Disaggregate probabilistic seismic {@code hazard} at a unique intensity
   * measure level for each IMT. Assumes hazard contains results for each IMT
   * identified in the {@code imtImls} map. The calculation configuration file
   * accompanying the {@code hazard} object is used to configure the
   * disaggregations.
   *
   * @param hazard to disaggregate.
   * @param imtImls per-IMT intensity measure levels at which to disaggregate
   *        {@code hazard}
   * @param exec executor service to use for disaggregation
   * @return a {@code Disaggregation} object
   */
  public static Disaggregation atImls(
      Hazard hazard,
      Map<Imt, Double> imtImls,
      Executor exec) {

    DisaggConfig.Builder cb = DisaggConfig.builder(hazard);
    HazardToDisagg transform = new HazardToDisagg(hazard);
    AsyncList<ImtDisagg> futureImtDisaggs = AsyncList.createWithCapacity(imtImls.size());

    for (Entry<Imt, Double> imtIml : imtImls.entrySet()) {
      Imt imt = imtIml.getKey();
      double lnIml = Math.log(imtIml.getValue());
      double rate = RATE_INTERPOLATER.findY(hazard.totalCurves.get(imt), lnIml);
      double returnPeriod = 1.0 / rate;
      DisaggConfig config = cb.imt(imt).iml(lnIml, rate, returnPeriod).build();
      futureImtDisaggs.add(toImtDisagg(transform, config, exec));
    }

    Map<Imt, ImtDisagg> imtDisaggs = toImtDisaggMap(futureImtDisaggs);
    return new Disaggregation(imtDisaggs, hazard.site);
  }

  private static Map<Imt, ImtDisagg> toImtDisaggMap(AsyncList<ImtDisagg> futureDisaggs) {
    try {
      ListenableFuture<List<ImtDisagg>> imtDisaggsFuture = Futures.allAsList(futureDisaggs);
      List<ImtDisagg> imtDisaggs = imtDisaggsFuture.get();
      Map<Imt, ImtDisagg> imtDisaggMap = Maps.newEnumMap(Imt.class);
      for (ImtDisagg imtDisagg : imtDisaggs) {
        imtDisaggMap.put(imtDisagg.config.imt, imtDisagg);
      }
      return Maps.immutableEnumMap(imtDisaggMap);
    } catch (ExecutionException | InterruptedException e) {
      throw new RuntimeException(e);
    }
  }

  private static ListenableFuture<ImtDisagg> toImtDisagg(
      HazardToDisagg function,
      DisaggConfig config,
      Executor exec) {

    return Futures.transform(
        Futures.immediateFuture(config),
        function::apply,
        exec);
  }

  private static final class HazardToDisagg implements Function<DisaggConfig, ImtDisagg> {

    private final Hazard hazard;

    HazardToDisagg(Hazard hazard) {
      this.hazard = hazard;
    }

    @Override
    public ImtDisagg apply(DisaggConfig config) {
      return new ImtDisagg(hazard, config);
    }
  }

  /* Hazard curves are already in log-x space. */
  static final Interpolator IML_INTERPOLATER = Interpolator.builder()
      .logy()
      .decreasingY()
      .build();

  /* Hazard curves are already in log-x space. */
  static final Interpolator RATE_INTERPOLATER = Interpolator.builder()
      .logy()
      .build();

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    for (Imt imt : disaggs.keySet()) {
      sb.append(disaggs.get(imt));
    }
    return sb.toString();
  }

  /**
   * Returns an object containing all disaggregation results that is suitable
   * for JSON serialization.
   *
   * @param imt of the disaggregation to retrieve.
   */
  public Object toJson(Imt imt) {
    return disaggs.get(imt).toJson(true, true, true, true);
  }

  /**
   * Returns an object containing custom disaggregation results that is suitable
   * for JSON serialization.
   *
   * @param imt of the disaggregation to retrieve.
   * @param summary include summary information, or not
   * @param gmms include results decomposed by GMM (values are scaled by GMM
   *        logic tree weight)
   * @param types include results decomposed by source type
   * @param epsilons include distance-magnitude-epsilon data that can be used to
   *        build a 3D stem plot
   */
  public Object toJson(Imt imt,
      boolean summary,
      boolean gmms,
      boolean types,
      boolean epsilons) {
    return disaggs.get(imt).toJson(summary, gmms, types, epsilons);
  }

  /**
   * Returns an object containing epsilon bin data suitable for JSON
   * serialization. This is exposed independent of JSON serialization of as web
   * services may need this metadata independent of disaggregation results.
   */
  public Object εBins() {
    return disaggs.values().iterator().next().config.εBins;
  }

  /* One per Imt in supplied Hazard. */
  static class ImtDisagg {

    final DisaggConfig config;
    final DisaggDataset totalDataset;
    final Map<Gmm, DisaggDataset> gmmDatasets;
    final Map<SourceType, DisaggDataset> typeDatasets;

    ImtDisagg(Hazard hazard, DisaggConfig config) {
      this.config = config;

      /*
       * Datasets are combined as follows:
       *
       * For each HazardCurveSet (RuptureSet), disaggregation is performed
       * across all relevant Gmms. These are preserved in ListMultimaps for
       * output of disaggregation by Gmm and SourceType. It's too much work to
       * consolidate ListMultimaps on the fly and keep track of all the nested
       * DisaggContributors, so lists are maintained of Gmm and SourceType
       * datasets, and the total across all Gmms that result from each call to
       * disaggregate(). The combination of multiple datasets for single
       * RuptureSets is then straightforward via static consolidators in
       * DisaggDataset.
       */

      int ruptureSetCount = hazard.ruptureSetCurves.size();
      ListMultimap<Gmm, DisaggDataset> gmmDatasetLists = MultimapBuilder
          .enumKeys(Gmm.class)
          .arrayListValues(ruptureSetCount)
          .build();
      ListMultimap<SourceType, DisaggDataset> typeDatasetLists = MultimapBuilder
          .enumKeys(SourceType.class)
          .arrayListValues(ruptureSetCount)
          .build();

      for (HazardCurveSet curveSet : hazard.ruptureSetCurves.values()) {
        XySequence ruptureSetCurve = curveSet.totalCurves.get(config.imt);
        double ruptureSetRate = RATE_INTERPOLATER.findY(ruptureSetCurve, config.iml);
        if (Double.isNaN(ruptureSetRate) || ruptureSetRate == 0.0) {
          // Consider logging statement below
          System.err.println("Skipping: " + curveSet.ruptures.name());
          continue;
        }
        Map<Gmm, DisaggDataset> ruptureSetDatasets = Disaggregator.disaggregate(
            curveSet,
            config,
            hazard.site);
        gmmDatasetLists.putAll(Multimaps.forMap(ruptureSetDatasets));
        DisaggDataset ruptureSetTotal = SOURCE_CONSOLIDATOR.apply(ruptureSetDatasets.values());
        typeDatasetLists.put(curveSet.ruptures.type(), ruptureSetTotal);
      }

      /* Combine RuptureSets across Gmms. */
      gmmDatasets = Maps.immutableEnumMap(Maps.transformValues(
          Multimaps.asMap(gmmDatasetLists),
          SOURCE_SET_CONSOLIDATOR::apply));

      /* Combine RuptureSets across SourceTypes. */
      typeDatasets = Maps.immutableEnumMap(Maps.transformValues(
          Multimaps.asMap(typeDatasetLists),
          SOURCE_SET_CONSOLIDATOR::apply));

      /* Combine RuptureSet totals. */
      totalDataset = SOURCE_SET_CONSOLIDATOR.apply(typeDatasets.values());
    }

    private static final String TOTAL_COMPONENT = "Total";
    private static final String GMM_COMPONENT = "GMM: ";
    private static final String TYPE_COMPONENT = "Source Type: ";

    @Override
    public String toString() {
      StringBuilder sb = new StringBuilder();
      sb.append(NEWLINE);
      for (DisaggExport export : buildExports(false, true, true, true, true)) {
        sb.append(export.toString());
        sb.append(NEWLINE);
      }
      return sb.toString();
    }

    /*
     * Method does not return a JSON String, but rather an appropriately
     * structured object that may be serialized directly or added to some other
     * object prior to serialization.
     */
    Object toJson(boolean summary, boolean gmmDisaggs, boolean typeDisaggs, boolean εData) {
      return buildExports(true, summary, gmmDisaggs, typeDisaggs, εData);
    }

    private List<DisaggExport> buildExports(
        boolean json,
        boolean summary,
        boolean gmmDisaggs,
        boolean typeDisaggs,
        boolean εData) {

      List<DisaggExport> exports = new ArrayList<>();
      DisaggExport total = new DisaggExport(
          totalDataset,
          totalDataset,
          config,
          TOTAL_COMPONENT,
          json,
          εData,
          summary);
      exports.add(total);
      if (gmmDisaggs) {
        for (Entry<Gmm, DisaggDataset> gmmEntry : gmmDatasets.entrySet()) {
          DisaggExport gmm = new DisaggExport(
              totalDataset,
              gmmEntry.getValue(),
              config,
              GMM_COMPONENT + gmmEntry.getKey().toString(),
              json,
              εData,
              summary);
          exports.add(gmm);
        }
      }
      if (typeDisaggs) {
        for (Entry<SourceType, DisaggDataset> typeEntry : typeDatasets.entrySet()) {
          DisaggExport type = new DisaggExport(
              totalDataset,
              typeEntry.getValue(),
              config,
              TYPE_COMPONENT + typeEntry.getKey().toString(),
              json,
              εData,
              summary);
          exports.add(type);
        }
      }
      return exports;
    }

  }

}
