package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.calc.ValueFormat.POISSON_PROBABILITY;
import static gov.usgs.earthquake.nshmp.data.MutableXySequence.emptyCopyOf;
import static java.nio.charset.StandardCharsets.UTF_8;
import static java.nio.file.StandardOpenOption.APPEND;
import static java.util.stream.Collectors.joining;
import static java.util.stream.Collectors.toCollection;
import static java.util.stream.Collectors.toList;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.OpenOption;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Function;
import java.util.logging.Logger;
import java.util.stream.IntStream;

import com.google.common.base.Stopwatch;
import com.google.common.collect.FluentIterable;
import com.google.common.collect.Iterables;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.internal.Parsing;
import gov.usgs.earthquake.nshmp.internal.Parsing.Delimiter;
import gov.usgs.earthquake.nshmp.model.HazardModel;
import gov.usgs.earthquake.nshmp.model.RuptureSet;
import gov.usgs.earthquake.nshmp.model.Source;
import gov.usgs.earthquake.nshmp.model.SourceTree;
import gov.usgs.earthquake.nshmp.model.SourceType;

/**
 * Hazard calculation result exporter.
 *
 * @author U.S. Geological Survey
 */
public final class HazardExport {

  /*
   * Developer notes
   *
   * Export of gmm and source type is unreliable for results where multiple
   * models may have been run (e.g. Hazard2018) because the model.types() and
   * model.gmms() methods will only reflect a single model.
   */

  private static final String GMM_DIR = "gmm";
  static final String TYPE_DIR = "source";
  private static final String CURVE_FILE_ASCII = "curves.csv";
  private static final String CURVE_FILE_ASCII_TRUNCATED = "curves-truncated.csv";
  private static final String VALUE_FMT = "%.8e";

  private final Logger log;
  private final Path dir;
  private final HazardModel model;
  private final CalcConfig config;
  private final boolean exportGmm;
  private final boolean exportSource;

  private final Function<Double, String> valueFormatter;

  private final Stopwatch batchWatch;
  private final Stopwatch totalWatch;
  private int batchCount = 0;
  private int resultCount = 0;

  private final boolean namedSites;
  private boolean used = false;

  private HazardExport(
      HazardModel model,
      CalcConfig config,
      List<Site> sites,
      Logger log) throws IOException {

    this.log = log;
    this.dir = createOutputDir(config.output.directory);
    this.model = model;
    this.config = config;
    this.exportGmm = config.output.dataTypes.contains(DataType.GMM);
    this.exportSource = config.output.dataTypes.contains(DataType.SOURCE);

    Function<Double, String> formatter = Parsing.formatDoubleFunction(VALUE_FMT);
    this.valueFormatter = (config.hazard.valueFormat == POISSON_PROBABILITY)
        ? Maths.annualRateToProbabilityConverter().andThen(formatter)
        : formatter;

    Site demoSite = sites.get(0);
    this.namedSites = demoSite.name() != Site.NO_NAME;

    this.batchWatch = Stopwatch.createStarted();
    this.totalWatch = Stopwatch.createStarted();

    init(sites);
  }

  /**
   * Create a new results handler.
   *
   * @param model being run
   * @param config that specifies output options and formats
   * @param sites reference to the sites to be processed (not retained)
   * @param log shared logging instance from calling class
   */
  public static HazardExport create(
      HazardModel model,
      CalcConfig config,
      List<Site> sites,
      Logger log) throws IOException {

    return new HazardExport(model, config, sites, log);
  }

  /* Initialize output directories. */
  private void init(List<Site> sites) throws IOException {

    for (Imt imt : config.hazard.imts) {

      Path imtDir = dir.resolve(imt.name());
      Files.createDirectory(imtDir);
      Path totalFile = imtDir.resolve(CURVE_FILE_ASCII);
      Path totalFileTruncated = imtDir.resolve(CURVE_FILE_ASCII_TRUNCATED);

      List<Object> headerValues = new ArrayList<>();
      if (namedSites) {
        headerValues.add(Site.Key.NAME);
      }
      headerValues.add(Site.Key.LON);
      headerValues.add(Site.Key.LAT);
      XySequence modelCurve = config.hazard.modelCurves().get(imt);
      headerValues.addAll(modelCurve.xValues().boxed().collect(toList()));

      String header = Parsing.join(headerValues, Delimiter.COMMA);
      writeLine(totalFile, header);
      writeLine(totalFileTruncated, header);

      if (exportSource) {
        Path typeParent = imtDir.resolve(TYPE_DIR);
        Files.createDirectory(typeParent);
        for (SourceType type : model.types()) {
          Path typeDir = typeParent.resolve(type.name());
          Files.createDirectory(typeDir);
          Path typeFile = typeDir.resolve(CURVE_FILE_ASCII);
          writeLine(typeFile, header);
        }
      }

      if (exportGmm) {
        Path gmmParent = imtDir.resolve(GMM_DIR);
        Files.createDirectory(gmmParent);
        for (Gmm gmm : model.gmms()) {
          Path gmmDir = gmmParent.resolve(gmm.name());
          Files.createDirectory(gmmDir);
          Path gmmFile = gmmDir.resolve(CURVE_FILE_ASCII);
          writeLine(gmmFile, header);
        }
      }
    }
  }

  /* Avoid clobbering exsting result directories via incrementing. */
  static Path createOutputDir(Path dir) throws IOException {
    int i = 1;
    Path incrementedDir = dir;
    while (Files.exists(incrementedDir)) {
      incrementedDir = incrementedDir.resolveSibling(dir.getFileName() + "-" + i);
      i++;
    }
    Files.createDirectories(incrementedDir);
    return incrementedDir;
  }

  /**
   * Write a hazard result.
   *
   * @param hazard to write
   */
  public void write(Hazard hazard) throws IOException {
    checkState(!used, "This result handler is expired");
    writeHazard(hazard);
    resultCount++;
    if (resultCount % 10 == 0) {
      batchCount++;
      log.info(String.format(
          "     batch: %s in %s – %s sites in %s",
          batchCount, batchWatch, resultCount, totalWatch));
      batchWatch.reset().start();
    }
  }

  /**
   * Stops all timers and sets the state of this {@code Results} instance to
   * 'used'; no more results may be added.
   */
  public void expire() {
    batchWatch.stop();
    totalWatch.stop();
    used = true;
  }

  /**
   * The number of hazard [and disagg] results passed to this handler thus far.
   */
  public int resultCount() {
    return resultCount;
  }

  /**
   * A string representation of the time duration that this result handler has
   * been running.
   */
  public String elapsedTime() {
    return totalWatch.toString();
  }

  /**
   * The target output directory established by this handler.
   */
  public Path outputDir() {
    return dir;
  }

  /* Write the supplied hazard and possible disagg results to file(s). */
  private void writeHazard(Hazard hazard) throws IOException {

    Set<Gmm> gmms = hazard.model.gmms();
    Set<SourceType> types = hazard.model.types();

    Location location = hazard.site.location();
    String locStr = (namedSites ? hazard.site.name() + "," : "") +
        location.longitude + "," + location.latitude + ",";

    Map<Imt, Map<SourceType, ? extends XySequence>> curvesBySource = exportSource
        ? curvesBySource(hazard)
        : null;

    Map<Imt, Map<Gmm, ? extends XySequence>> curvesByGmm = exportGmm
        ? curvesByGmm(hazard)
        : null;

    for (Entry<Imt, XySequence> imtEntry : hazard.totalCurves.entrySet()) {
      XySequence totalCurve = imtEntry.getValue();

      /* Utility empty line to write if source or gmm is missing at site */
      String emptyLine = locStr + "0.0" + ",0.0".repeat(totalCurve.size() - 1);

      Imt imt = imtEntry.getKey();
      Path imtDir = dir.resolve(imt.name());

      String totalLine = locStr + valueStr(totalCurve, valueFormatter);
      String totalLineTrunc = locStr + truncValueStr(totalCurve, valueFormatter, TRUNCATION_LIMIT);

      Path totalFile = imtDir.resolve(CURVE_FILE_ASCII);
      writeLine(totalFile, totalLine, APPEND);
      Path totalFileTrunc = imtDir.resolve(CURVE_FILE_ASCII_TRUNCATED);
      writeLine(totalFileTrunc, totalLineTrunc, APPEND);

      if (exportSource) {
        Map<SourceType, ? extends XySequence> sourceCurveMap = curvesBySource.get(imt);
        Path typeParent = imtDir.resolve(TYPE_DIR);
        for (SourceType type : types) {
          Path typeDir = typeParent.resolve(type.name());
          Path typeFile = typeDir.resolve(CURVE_FILE_ASCII);
          String typeLine = emptyLine;
          if (sourceCurveMap.containsKey(type)) {
            XySequence typeCurve = sourceCurveMap.get(type);
            typeLine = locStr + valueStr(typeCurve, valueFormatter);
          }
          writeLine(typeFile, typeLine, APPEND);
        }
      }

      if (exportGmm) {
        Map<Gmm, ? extends XySequence> gmmCurveMap = curvesByGmm.get(imt);
        Path gmmParent = imtDir.resolve(GMM_DIR);
        for (Gmm gmm : gmms) {
          Path gmmDir = gmmParent.resolve(gmm.name());
          Path gmmFile = gmmDir.resolve(CURVE_FILE_ASCII);
          String gmmLine = emptyLine;
          if (gmmCurveMap.containsKey(gmm)) {
            XySequence gmmCurve = gmmCurveMap.get(gmm);
            gmmLine = locStr + valueStr(gmmCurve, valueFormatter);
          }
          writeLine(gmmFile, gmmLine, APPEND);
        }
      }
    }
  }

  /*
   * Write the current list of {@code Disaggregation}s to file.
   */
  // @Deprecated
  // private void writeDisagg(Disaggregation disagg) throws IOException {
  //
  // /*
  // * Writing of Hazard results will have already created necessary Imt, Gmm,
  // * and SourceType directories.
  // */
  // String name = namedSites ? disagg.site.name() :
  // lonLatStr(disagg.site.location());
  // for (Entry<Imt, ImtDisagg> imtEntry : disagg.disaggs.entrySet()) {
  //
  // /* Write total dataset. */
  // ImtDisagg imtDisagg = imtEntry.getValue();
  // DisaggDataset ddTotal = imtDisagg.totalDataset;
  // DisaggConfig dc = imtDisagg.config;
  // DisaggExport exporter = new DisaggExport(ddTotal, ddTotal, dc, "Total",
  // false, true, true);
  // Path imtDir = dir.resolve(imtEntry.getKey().name());
  // Path totalDir = imtDir.resolve(DISAGG_DIR);
  // Files.createDirectories(totalDir);
  // exporter.toFile(totalDir, name);
  //
  // if (exportSource) {
  // for (Entry<SourceType, DisaggDataset> typeEntry :
  // imtDisagg.typeDatasets.entrySet()) {
  // SourceType type = typeEntry.getKey();
  // Path typeDir = imtDir.resolve(TYPE_DIR)
  // .resolve(type.name())
  // .resolve(DISAGG_DIR);
  // DisaggDataset ddType = typeEntry.getValue();
  // exporter = new DisaggExport(ddTotal, ddType, dc, type.toString(), false,
  // true, true);
  // exporter.toFile(typeDir, name);
  // }
  // }
  //
  // if (exportGmm) {
  // for (Entry<Gmm, DisaggDataset> gmmEntry : imtDisagg.gmmDatasets.entrySet())
  // {
  // Gmm gmm = gmmEntry.getKey();
  // Path gmmDir = imtDir.resolve(GMM_DIR)
  // .resolve(gmm.name())
  // .resolve(DISAGG_DIR);
  // DisaggDataset ddGmm = gmmEntry.getValue();
  // exporter = new DisaggExport(ddTotal, ddGmm, dc, gmm.toString(), false,
  // true, true);
  // exporter.toFile(gmmDir, name);
  // }
  // }
  // }
  // }

  private static String valueStr(
      XySequence curve,
      Function<Double, String> formatter) {

    return curve.yValues()
        .mapToObj(formatter::apply)
        .collect(joining(","));
  }

  private static String truncValueStr(
      XySequence curve,
      Function<Double, String> formatter,
      double cutoff) {

    double[] values = curve.yValues().toArray();
    int limit = truncationIndex(values, cutoff);
    return IntStream.range(0, values.length)
        .mapToObj(i -> (i <= limit) ? formatter.apply(values[i]) : "")
        .collect(joining(","));
  }

  /* No options will write a new file, supply APPEND to append. */
  static void writeLine(Path path, String line, OpenOption... options)
      throws IOException {

    Files.write(
        path,
        (line + System.lineSeparator()).getBytes(UTF_8),
        options);
  }

  /**
   * Derive maps of curves by {@code SourceType} for each {@code Imt}.
   */
  public static Map<Imt, Map<SourceType, ? extends XySequence>> curvesBySource(Hazard hazard) {

    EnumMap<Imt, Map<SourceType, MutableXySequence>> imtMap = Maps.newEnumMap(Imt.class);

    // initialize receiver
    Set<SourceType> types = hazard.ruptureSetCurves.keySet();
    for (Entry<Imt, XySequence> entry : hazard.curves().entrySet()) {
      imtMap.put(entry.getKey(), initCurves(types, MutableXySequence.copyOf(entry.getValue())));
    }

    for (Entry<SourceType, HazardCurveSet> curveSet : hazard.ruptureSetCurves.entries()) {
      for (Entry<Imt, XySequence> typeTotals : curveSet.getValue().totalCurves.entrySet()) {
        imtMap.get(typeTotals.getKey())
            .get(curveSet.getKey())
            .add(typeTotals.getValue());
      }
    }
    return Maps.immutableEnumMap(imtMap);
  }

  @Deprecated // get from config
  private static final double TRUNCATION_LIMIT = 1e-4;

  /**
   * Returns the index of the first value that is less than {@code cutoff}.
   * Method is used to truncate hazard curves below the supplied {@code cutoff}
   * and assumes {@code values} is a monotonically decreasing array.
   *
   * @param values for which to compute truncation index
   */
  public static int truncationIndex(double[] values, double cutoff) {
    for (int i = 0; i < values.length; i++) {
      if (values[i] < cutoff) {
        return i;
      }
    }
    return values.length - 1;
  }

  /**
   * Derive maps of curves by {@code Gmm} for each {@code Imt} in a
   * {@code Hazard} result.
   */
  public static Map<Imt, Map<Gmm, ? extends XySequence>> curvesByGmm(Hazard hazard) {

    EnumMap<Imt, Map<Gmm, MutableXySequence>> imtMap = Maps.newEnumMap(Imt.class);

    /* Initialize receiver */
    Iterable<RuptureSet<? extends Source>> sources = Iterables.transform(
        hazard.ruptureSetCurves.values(),
        CURVE_SET_TO_SOURCE_SET::apply);
    Set<Gmm> gmms = gmmSet(sources);
    for (Entry<Imt, XySequence> entry : hazard.curves().entrySet()) {
      imtMap.put(entry.getKey(), initCurves(gmms, MutableXySequence.copyOf(entry.getValue())));
    }

    for (HazardCurveSet curveSet : hazard.ruptureSetCurves.values()) {
      for (Entry<Imt, Map<Gmm, MutableXySequence>> imtEntry : curveSet.curveMap.entrySet()) {
        for (Entry<Gmm, MutableXySequence> gmmEntry : imtEntry.getValue().entrySet()) {
          imtMap.get(imtEntry.getKey()).get(gmmEntry.getKey()).add(gmmEntry.getValue());
        }
      }
    }

    return Maps.immutableEnumMap(imtMap);
  }

  /*
   * Scan the supplied source sets for the set of all GMMs used.
   *
   * Note that this is used when consolidating gmm result curves. Results may
   * not span the entire range of GMMs in a model so this is commonly a subset
   */

  private static Set<Gmm> gmmSet(Collection<SourceTree> trees) {
    return trees.stream()
        .flatMap(tree -> tree.gmms().gmms().stream())
        .collect(toCollection(() -> EnumSet.noneOf(Gmm.class)));
  }

  // Consider Stream (above)
  @Deprecated
  private static Set<Gmm> gmmSet(Iterable<RuptureSet<? extends Source>> ruptureSets) {
    return Sets.immutableEnumSet(
        FluentIterable.from(ruptureSets).transformAndConcat(
            new Function<RuptureSet<? extends Source>, Set<Gmm>>() {
              @Override
              public Set<Gmm> apply(RuptureSet<? extends Source> ruptures) {
                return ruptures.groundMotionModels().gmms();
              }
            }::apply));
  }

  /* Initalize a map of curves, one entry for each of the supplied enum key. */
  // Consider Stream
  private static <K extends Enum<K>> Map<K, MutableXySequence> initCurves(
      final Set<K> keys,
      final MutableXySequence model) {
    return FluentIterable.from(keys).toMap(
        new Function<K, MutableXySequence>() {
          @Override
          public MutableXySequence apply(K key) {
            return emptyCopyOf(model);
          }
        }::apply);
  }

  private static final Function<HazardCurveSet, RuptureSet<? extends Source>> CURVE_SET_TO_SOURCE_SET =
      new Function<HazardCurveSet, RuptureSet<? extends Source>>() {
        @Override
        public RuptureSet<? extends Source> apply(HazardCurveSet curves) {
          return curves.ruptures;
        }
      };

}
