package gov.usgs.earthquake.nshmp.calc;

import static gov.usgs.earthquake.nshmp.Text.LOG_INDENT;
import static gov.usgs.earthquake.nshmp.geo.BorderType.MERCATOR_LINEAR;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.io.IOException;
import java.nio.file.Path;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Stream;

import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.geo.Bounds;
import gov.usgs.earthquake.nshmp.geo.GriddedRegion;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Region;
import gov.usgs.earthquake.nshmp.geo.Regions;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson.Type;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.model.SiteData;

/**
 * Factory class that generates {@code Site} lists from different resources.
 *
 * @author U.S. Geological Survey
 */
public final class Sites {

  private Sites() {}

  /**
   * Create an unmodifiable list of sites from the comma-delimted site file
   * designated by {@code path}. Site parameter values in a sites file take
   * precedence over defaults from a config and model values in the supplied
   * siteData.
   *
   * @param path to comma-delimited site data file
   * @throws IOException if a problem is encountered
   */
  public static List<Site> fromCsv(
      Path path,
      CalcConfig config,
      SiteData siteData) throws IOException {

    DelimitedData dd = DelimitedData.comma(path);
    try (Stream<Record> stream = dd.records()) {
      return stream.map(record -> recordToSite(
          record, dd.columnKeys(), config, siteData))
          .collect(toUnmodifiableList());
    }
  }

  private static Site recordToSite(
      Record record,
      Set<String> keys,
      CalcConfig config,
      SiteData siteData) {

    /*
     * Site parameter precedence:
     *
     * (3) Defaults from site builder and config (2) site data (1) sites file
     *
     * If a sites file contains columns for basin depths, a 'null' value string
     * indicates a GMM should do whatever it defines as default behavior and
     * will override values that may be included in the site data of a model. If
     * the value is empty, then whatever defaults from (2) or (3) will be used.
     */

    Location loc = Location.create(
        record.getDouble(Site.Key.LON),
        record.getDouble(Site.Key.LAT));

    Site.Builder builder = initBuilder(loc, config, siteData);

    // Override with values from sites file
    if (keys.contains(Site.Key.NAME)) {
      builder.name(record.get(Site.Key.NAME));
    }
    if (keys.contains(Site.Key.VS30)) {
      builder.vs30(record.getDouble(Site.Key.VS30));
    }
    if (keys.contains(Site.Key.VS_INF)) {
      builder.vsInferred(record.getBoolean(Site.Key.VS_INF));
    }
    if (keys.contains(Site.Key.Z1P0)) {
      builder.z1p0(record.getDouble(Site.Key.Z1P0));
    }
    if (keys.contains(Site.Key.Z2P5)) {
      builder.z2p5(record.getDouble(Site.Key.Z2P5));
    }

    return builder.build();
  }

  /*
   * Initialize a builder with location, config defaults and possible site
   * specific values for basin and other terms.
   */
  private static Site.Builder initBuilder(
      Location loc,
      CalcConfig config,
      SiteData siteData) {

    Site.Builder builder = Site.builder()
        .location(loc)
        .vs30(config.site.vs30)
        .vsInferred(config.site.vsInferred)
        .z1p0(config.site.z1p0)
        .z2p5(config.site.z2p5);

    SiteData.Values sdValues = siteData.get(loc);
    sdValues.z1p0.ifPresent(builder::z1p0);
    sdValues.z2p5.ifPresent(builder::z2p5);

    return builder;
  }

  /**
   * Create an unmodifiable list of sites from the GeoJSON site file designated
   * by {@code path}. The GeoJSON may either be a feature collection of point
   * features or a polygon feature and optional bounds feature.
   *
   * <p>A GeoJSON of points may include site parameters as properties of each
   * feature; these will override any default values from a config and model
   * values in the supplied siteData. Put any default values for use with
   * GeoJSON map polygon in the supplied {@code config}.
   *
   * @param path to GeoJson site data file
   * @throws IOException if a problem is encountered
   */
  public static List<Site> fromGeoJson(
      Path path,
      CalcConfig defaults,
      SiteData siteData) throws IOException {

    FeatureCollection fc = GeoJson.from(path).toFeatureCollection();
    List<Feature> features = fc.features();

    if (features.get(0).type() == Type.POINT) {
      return fromPoints(features, defaults, siteData);
    } else {
      return fromPolygon(features, defaults, siteData);
    }
  }

  private static List<Site> fromPoints(
      List<Feature> features,
      CalcConfig defaults,
      SiteData siteData) {

    return features.stream()
        .map(feature -> fromPoint(feature, defaults, siteData))
        .collect(toUnmodifiableList());
  }

  private static Site fromPoint(Feature feature, CalcConfig config, SiteData siteData) {
    Location loc = feature.asPoint();
    Site.Builder builder = initBuilder(loc, config, siteData);
    Properties props = feature.properties();
    props.getString(Site.Key.NAME).ifPresent(builder::name);
    props.getDouble(Site.Key.VS30).ifPresent(builder::vs30);
    props.getBoolean(Site.Key.VS_INF).ifPresent(builder::vsInferred);
    props.getDouble(Site.Key.Z1P0).ifPresent(builder::z1p0);
    props.getDouble(Site.Key.Z2P5).ifPresent(builder::z2p5);
    return builder.build();
  }

  private static List<Site> fromPolygon(
      List<Feature> features,
      CalcConfig config,
      SiteData siteData) {

    Optional<Bounds> mapBounds = Optional.empty();
    String boundsName = "";

    /* Possibly set map extents. */
    int mapRegionIndex = 0;
    if (features.size() > 1) {
      mapRegionIndex = 1;
      Feature mapPoly = features.get(0);
      LocationList mapPolyBorder = mapPoly.asPolygonBorder();
      mapBounds = Optional.of(GeoJson.validateBoundingBox(mapPolyBorder).bounds());
      boundsName = mapPoly.properties().getString(Site.Key.NAME).orElse("Map Extents");
    }

    Feature sitesPoly = features.get(mapRegionIndex);
    LocationList sitesPolyBorder = sitesPoly.asPolygonBorder();
    Properties properties = sitesPoly.properties();
    String mapName = properties.getString(Site.Key.NAME).orElse("Unnamed Map");

    /*
     * We special case a 5-coordinate border that defines a mercator recangle so
     * as to create a region that includes sites on the north and east borders.
     */

    Region calcRegion = null;
    try {
      Bounds b = GeoJson.validateBoundingBox(sitesPolyBorder).bounds();
      calcRegion = Regions.createRectangular(mapName, b.min, b.max);
    } catch (IllegalArgumentException iae) {
      calcRegion = Regions.create(mapName, sitesPolyBorder, MERCATOR_LINEAR);
    }

    double spacing = properties.getDouble("spacing").orElseThrow();

    Region mapRegion = calcRegion;
    if (mapBounds.isPresent()) {
      Bounds b = mapBounds.orElseThrow();
      Region r = Regions.createRectangular(boundsName, b.min, b.max);
      mapRegion = Regions.intersectionOf(mapName, r, calcRegion);
    }

    GriddedRegion region = Regions.toGridded(
        mapRegion,
        spacing, spacing,
        GriddedRegion.ANCHOR_0_0);

    return region.nodes().stream()
        .map(loc -> initBuilder(loc, config, siteData))
        .map(Site.Builder::build)
        .collect(toUnmodifiableList());
  }

  private static final int TO_STRING_LIMIT = 5;
  private static final String SITE_INDENT = LOG_INDENT + "       ";

  /** Site list string representation for logs. */
  public static String toString(List<Site> sites) {
    StringBuilder sb = new StringBuilder();
    sb.append("Sites [size=")
        .append(sites.size())
        .append("]");
    sites.stream()
        .limit(TO_STRING_LIMIT)
        .forEach(site -> sb.append(SITE_INDENT).append(site));
    if (sites.size() > TO_STRING_LIMIT) {
      int delta = sites.size() - TO_STRING_LIMIT;
      sb.append(SITE_INDENT).append("... and ").append(delta).append(" more ...");
    }
    return sb.toString();
  }

}
