package gov.usgs.earthquake.nshmp.data;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static java.util.stream.Collectors.toUnmodifiableList;
import static java.util.stream.Collectors.toUnmodifiableMap;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import com.google.common.base.CharMatcher;
import com.google.common.base.Splitter;

/**
 * Container class for streaming delimited data.
 *
 * <p><ul><li>Each row in a {@code DelimitedData} object is streamed as a
 * {@link Record}.
 *
 * <li>The {@link #columnKeys()} method returns the set of identifiers used to
 * access the values in a record.
 *
 * <li> A delimited data resource may contain comment rows marked with
 * {@code '#'} that are accessible via the {@link #comments()} method.
 *
 * <li>The first non-comment row <i>must</i> contain column identifiers.
 *
 * <li>Empty column <i>values</i> are supported in comma-delimited data (e.g.
 * {@code '-117.1,34.0,,,'}) but not whitespace-delimited data.
 *
 * <li>When using this class with {@code Path} based resources, wrap calls to
 * {@link #records()} in a try-with-resources statement or similar control
 * structure (see {@link Files#lines(Path)}).</ul>
 *
 * <p>Example usage:
 *
 * <pre>
 * try (Stream<Record> records = DelimitedData.comma(path).records()) {
 *   records.forEach(System.out::println);
 * }
 * </pre>
 *
 * @author U.S. Geological Survey
 */
public class DelimitedData {

  private static Splitter COMMA = Splitter.on(CharMatcher.is(',')).trimResults();
  private static Splitter WHITESPACE = Splitter.on(CharMatcher.whitespace())
      .omitEmptyStrings()
      .trimResults();

  private final Supplier<Stream<String>> lines;
  private final Map<String, Integer> keyMap;
  private final Splitter splitter;

  private DelimitedData(Supplier<Stream<String>> lines, Splitter splitter) {
    this.lines = lines;
    this.splitter = splitter;
    this.keyMap = readHeader();
  }

  /**
   * Create a new {@code DelimitedData} instance from a comma-delimited file.
   *
   * @param path a path to a comma-delimited file
   * @throws IllegalArgumentException if {@code path} is invalid
   * @throws IllegalStateException if header row is missing or contains empty or
   *         duplicate fields
   */
  public static DelimitedData comma(Path path) {
    return new DelimitedData(supplyFromPath(path), COMMA);
  }

  /**
   * Create a new {@code DelimitedData} instance from a list of comma-delimited
   * strings.
   *
   * @param lines a list of comma-delimited strings
   * @throws IllegalStateException if header row is missing or contains empty or
   *         duplicate fields
   */
  public static DelimitedData comma(List<String> lines) {
    return new DelimitedData(supplyFromList(lines), COMMA);
  }

  /**
   * Create a new {@code DelimitedData} instance from a whitespace-delimited
   * file. Whitespace includes spaces and tabs. Empty column values are not
   * supported in whitespace-delimited data.
   *
   * @param path a path to a whitespace-delimited file
   * @throws IllegalArgumentException if {@code path} is invalid
   * @throws IllegalStateException if header row is missing or contains empty or
   *         duplicate fields
   * @see CharMatcher#whitespace()
   */
  public static DelimitedData whitespace(Path path) {
    return new DelimitedData(supplyFromPath(path), WHITESPACE);
  }

  /**
   * Create a new {@code DelimitedData} instance from a list of
   * whitespace-delimited strings.
   *
   * @param lines a list of whitespace-delimited strings
   * @throws IllegalStateException if header row is missing or contains empty or
   *         duplicate fields
   */
  public static DelimitedData whitespace(List<String> lines) {
    return new DelimitedData(supplyFromList(lines), WHITESPACE);
  }

  /* Create a supplier of lines from a list of strings. */
  private static Supplier<Stream<String>> supplyFromList(List<String> lines) {
    checkNotNull(lines);
    return () -> lines.stream();
  }

  /* Create a supplier of lines from a path. */
  private static Supplier<Stream<String>> supplyFromPath(Path path) {
    checkArgument(
        Files.exists(checkNotNull(path)),
        "Invalid csv path: %s", path);

    Supplier<Stream<String>> streamer = () -> {
      try {
        return Files.lines(path);
      } catch (IOException ioe) {
        throw new UncheckedIOException("Could not read [" + path + "]", ioe);
      }
    };
    return streamer;
  }

  /**
   * Return the set of column keys that may be used to access values in a
   * {@link Record}.
   */
  public Set<String> columnKeys() {
    return keyMap.keySet();
  }

  /**
   * Return a list of all comment lines. Each call to this method will read the
   * entire file or list.
   */
  public List<String> comments() {
    return lines.get()
        .filter(Predicate.not(DelimitedData::filterComments))
        .collect(toUnmodifiableList());
  }

  /**
   * Return a stream of the records (rows) in this {@code DelimitedData} object.
   * For {@code Path} based resources, callers should wrap references to the
   * returned stream in a try-with-resources statement or similar control
   * structure (see {@link Files#lines(Path)}) to ensure proper closing of the
   * underlying resource.
   *
   * @throws IllegalStateException if a record is encountered while streaming
   *         that is not the same size as the header row
   */
  public Stream<Record> records() {
    return lines.get()
        .filter(DelimitedData::filterComments)
        .skip(1)
        .map(Record::new);
  }

  /* Comment skipping predicate. */
  private static boolean filterComments(String line) {
    return !line.startsWith("#");
  }

  /*
   * Extract header line (first non-comment line) from file and return map of
   * header strings to column index.
   */
  private Map<String, Integer> readHeader() {
    Optional<String> header = lines.get()
        .filter(DelimitedData::filterComments)
        .findFirst();
    checkState(header.isPresent(), "Header row missing");
    List<String> keyList = splitter.splitToList(header.orElseThrow());
    Set<String> keySet = Set.copyOf(keyList);
    checkState(
        keyList.size() == keySet.size(),
        "Header row contains duplicate columns/keys: %s", keyList);
    checkState(
        !keyList.contains(""),
        "Header row contains unlabeled columns", keyList);

    return IntStream.range(0, keyList.size())
        .boxed()
        .collect(toUnmodifiableMap(
            keyList::get,
            Function.identity()));
  }

  private void checkKey(String key) {
    if (!keyMap.containsKey(key)) {
      throw new NoSuchElementException(
          "Specified key [" + key + "] is not a valid data field");
    }
  }

  /** An entry (row) in a delimited-data file. */
  public class Record {

    private List<String> values;

    private Record(String line) {
      this.values = splitter.splitToList(line);
      checkState(
          values.size() == keyMap.size(),
          "Bad record: %s", line);
    }

    /**
     * Return the specified record value as a {@code boolean}.
     * @param key for value
     */
    public boolean getBoolean(String key) {
      return Boolean.parseBoolean(get(key));
    }

    /**
     * Return the specified record value as a {@code double}.
     * @param key for value
     */
    public double getDouble(String key) {
      return Double.parseDouble(get(key));
    }

    /**
     * Return the specified record value as an optional {@code double}.
     * @param key for value
     */
    public OptionalDouble getOptionalDouble(String key) {
      return get(key).isBlank()
          ? OptionalDouble.empty()
          : OptionalDouble.of(getDouble(key));
    }

    /**
     * Return the specified record value as a string.
     * @param key for value
     */
    public String get(String key) {
      checkKey(key);
      return values.get(keyMap.get(key));
    }

    @Override
    public String toString() {
      return values.toString();
    }
  }
}
