package gov.usgs.earthquake.nshmp.data;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkElementIndex;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.IntervalData.checkDataState;
import static gov.usgs.earthquake.nshmp.data.IntervalData.indexOf;
import static gov.usgs.earthquake.nshmp.data.IntervalData.keys;
import static java.util.stream.Collectors.toList;

import java.util.Arrays;
import java.util.List;

/**
 * An array of immutable, double-valued data that is arranged according to
 * increasing and uniformly spaced double-valued keys. Interval arrays are used
 * to represent binned data (or histograms), and so while row keys are bin
 * centers, indexing is managed internally using bin edges. This simplifies
 * issues related to rounding/precision errors that occur when indexing and
 * binning according to explicit double values.
 *
 * <p>To create an instance of an {@code IntervalArray}, use a {@link Builder}.
 *
 * <p>Internally, an {@code IntervalArray} is backed by a {@code double[]} where
 * a 'row' maps to an index in the backing array.
 *
 * <p>Note that interval arrays are not intended for use with very high
 * precision data and keys are currently limited to a precision of 4 decimal
 * places. This may change in the future.
 *
 * @author U.S. Geological Survey
 * @see IntervalData
 * @see IntervalTable
 * @see IntervalVolume
 */
public final class IntervalArray {

  private final double rowMin;
  private final double rowMax;
  private final double rowΔ;
  private final double[] rows;
  private final double[] data;

  private IntervalArray(Builder builder) {
    rowMin = builder.rowMin;
    rowMax = builder.rowMax;
    rowΔ = builder.rowΔ;
    rows = builder.rows;
    data = builder.data;
  }

  /**
   * Return the value of the bin that maps to the supplied row value. Do not
   * confuse this method with {@link #get(int)} by row index.
   *
   * @param rowValue of bin to retrieve
   * @throws IndexOutOfBoundsException if value is out of range
   */
  public double get(double rowValue) {
    int iRow = indexOf(rowMin, rowΔ, rowValue, rows.length);
    return get(iRow);
  }

  /**
   * Return the value of the bin that maps to the supplied row index. Do not
   * confuse this method with {@link #get(double)} by row value.
   *
   * @param rowIndex of bin to retrieve
   * @throws IndexOutOfBoundsException if index is out of range
   */
  public double get(int rowIndex) {
    return data[rowIndex];
  }

  /**
   * An immutable view of {@code this} as an {@link XySequence}.
   */
  public XySequence values() {
    return new ArrayXySequence(rows, data);
  }

  /**
   * The lower edge of the lowermost bin.
   */
  public double rowMin() {
    return rowMin;
  }

  /**
   * The upper edge of the uppermost bin.
   */
  public double rowMax() {
    return rowMax;
  }

  /**
   * The row bin discretization.
   */
  public double rowΔ() {
    return rowΔ;
  }

  /**
   * A lazily created copy of the row keys (bin centers).
   */
  public List<Double> rows() {
    return Arrays.stream(rows).boxed().collect(toList());
  }

  /**
   * The sum of the values in this array.
   */
  public double sum() {
    return DoubleData.sum(data);
  }

  /**
   * The index of the bin with smallest value.
   */
  public int minIndex() {
    return Indexing.minIndex(data);
  }

  /**
   * The index of the bin with largest value.
   */
  public int maxIndex() {
    return Indexing.maxIndex(data);
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    IntervalData.appendArrayKeys(sb, "", rows());
    IntervalData.appendArrayValues(sb, values().yValues().boxed().collect(toList()));
    return sb.toString();
  }

  /**
   * A builder of immutable {@code IntervalArray}s.
   *
   * <p>Rows must be specified before any data can be added. Note that any
   * supplied {@code max} values may not correspond to the final upper edge of
   * the uppermost bins if {@code max - min} is not evenly divisible by
   * {@code Δ} .
   */
  public static final class Builder {

    private double[] data;

    private double rowMin;
    private double rowMax;
    private double rowΔ;
    private double[] rows;

    private boolean built = false;

    private Builder() {}

    /**
     * Create a new builder with the structure and content identical to that of
     * the supplied array.
     *
     * @param array to copy
     */
    public static Builder copyOf(IntervalArray array) {
      Builder builder = copyStructure(array);
      builder.data = Arrays.copyOf(
          array.data,
          array.data.length);
      builder.init();
      return builder;
    }

    /**
     * Create a new builder with a structure identical to that of the supplied
     * model.
     *
     * @param model interval array
     */
    public static Builder fromModel(IntervalArray model) {
      Builder builder = copyStructure(model);
      builder.init();
      return builder;
    }

    private static Builder copyStructure(IntervalArray from) {
      Builder to = new Builder();
      to.rowMin = from.rowMin;
      to.rowMax = from.rowMax;
      to.rowΔ = from.rowΔ;
      to.rows = from.rows;
      return to;
    }

    /**
     * Define the array intervals.
     *
     * @param min lower edge of lowermost row bin
     * @param max upper edge of uppermost row bin
     * @param delta bin discretization
     */
    public static Builder withRows(double min, double max, double delta) {
      Builder b = new Builder();
      b.rowMin = min;
      b.rowMax = max;
      b.rowΔ = delta;
      b.rows = keys(min, max, delta);
      b.init();
      return b;
    }

    private void init() {
      if (data == null) {
        data = new double[rows.length];
      }
    }

    /**
     * Return the index of the row that would contain the supplied value.
     * @param row value
     */
    public int rowIndex(double row) {
      return indexOf(rowMin, rowΔ, row, rows.length);
    }

    /**
     * Set the value at the specified row. Be careful not to confuse this with
     * {@link #set(int, double)}.
     *
     * @param row key
     * @param value to set
     */
    public Builder set(double row, double value) {
      return set(rowIndex(row), value);
    }

    /**
     * Set the value at the specified row. Be careful not to confuse this with
     * {@link #set(double, double)}.
     *
     * @param row index
     * @param value to set
     */
    public Builder set(int row, double value) {
      data[row] = value;
      return this;
    }

    /**
     * Add to the existing value at the specified row. Be careful not to confuse
     * this with {@link #add(int, double)}.
     *
     * @param row key
     * @param value to add
     */
    public Builder add(double row, double value) {
      return add(rowIndex(row), value);
    }

    /**
     * Add to the existing value at the specified row. Be careful not to confuse
     * this with {@link #add(double, double)}.
     *
     * @param row index
     * @param value to add
     */
    public Builder add(int row, double value) {
      data[row] += value;
      return this;
    }

    /**
     * Add supplied values to this builder. Assumes that supplied value array is
     * the same size or smaller than the internal builder data array.
     *
     * @param values to add
     * @throws IndexOutOfBoundsException if values overrun array
     */
    public Builder add(double[] values) {
      checkElementIndex(values.length - 1, rows.length,
          "Supplied values overrun array");
      for (int i = 0; i < values.length; i++) {
        data[i] += values[i];
      }
      return this;
    }

    /**
     * Add the y-values of the supplied sequence to this builder. Assumes that
     * supplied sequence is the same size or smaller than the internal builder
     * data array.
     *
     * @param sequence to add
     * @throws IndexOutOfBoundsException if values overrun array
     */
    public Builder add(XySequence sequence) {
      // safe covariant cast
      return add(((ArrayXySequence) sequence).ys);
    }

    /**
     * Add the values in the supplied interval array to this builder. This
     * operation is very efficient if this builder and the supplied array are
     * sourced from the same model.
     *
     * @param array to add
     * @throws IllegalArgumentException if the rows of the supplied array do not
     *         match those of this array
     * @see #fromModel(IntervalArray)
     */
    public Builder add(IntervalArray array) {
      validateArray(array);
      DoubleData.uncheckedAdd(data, (array).data);
      return this;
    }

    /*
     * Check hash codes of row arrays in case fromModel or copyOf has been used,
     * otherwise check array equality.
     */
    IntervalArray validateArray(IntervalArray that) {
      checkArgument(Arrays.equals(this.rows, that.rows));
      return that;
    }

    /**
     * Add each value-pair of the supplied sequence to the appropriate interval.
     *
     * @param sequence of values to add
     */
    public Builder addEach(XySequence sequence) {
      sequence.stream().forEach(xy -> add(xy.x(), xy.y()));
      return this;
    }

    /**
     * Multiply ({@code scale}) all values in this builder.
     *
     * @param scale factor
     */
    public Builder multiply(double scale) {
      DoubleData.multiply(scale, data);
      return this;
    }

    /** Internal method to directly set data array. */
    Builder data(double[] data) {
      this.data = data;
      return this;
    }

    /*
     * Data is not copied on build() so we dereference data arrays to prevent
     * lingering builders from further modifying data.
     */
    private void dereference() {
      data = null;
      rows = null;
    }

    /**
     * Return a newly-created, immutable, interval data array populated with the
     * contents of this {@code Builder}.
     */
    public IntervalArray build() {
      checkState(built != true, "This builder has already been used");
      checkDataState(rows);
      IntervalArray array = new IntervalArray(this);
      dereference();
      built = true;
      return array;
    }
  }

}
