package gov.usgs.earthquake.nshmp.fault.surface;

import java.util.List;

import com.google.common.collect.Lists;

import gov.usgs.earthquake.nshmp.Faults;
import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.LocationVector;
import gov.usgs.earthquake.nshmp.geo.Locations;

/**
 * A {@code GriddedSurface} defined by an upper and lower trace whose spacing is
 * scaled to the be as close to a desired target spacing as possible over the
 * entire surface. The fault traces supplied to this surface <i>must</i> adhere
 * to the right-hand-rule (dip of surface is to right of direction of travel
 * along trace).
 *
 * @author U.S. Geological Survey
 */
public class ApproxGriddedSurface extends AbstractGriddedSurface {

  private LocationList upperTrace = null;
  private LocationList lowerTrace = null;

  private double avgDipRad;
  private double avgDepth;

  private Location centroid;

  private LocationList perimeter;

  /**
   * This constructor takes an upper and lower fault trace, re-samples these
   * according the the given aveGridSpacing to represent the first and last rows
   * of the surface, and then fills in the intermediate rows by evenly sampling
   * a straight line between the top and bottom point of each column. The number
   * of columns is the average length the top and bottom trace divided by the
   * aveGridSpacing (plus 1), and the number of rows is the average distance
   * between the top and bottom points in each column divided by the
   * aveGridSpacing (plus 1). and then
   * @param upperTrace
   * @param lowerTrace
   * @param spacing
   */
  public ApproxGriddedSurface(LocationList upperTrace,
      LocationList lowerTrace, double spacing) {

    // need to validate like builder methods in DefaultGriddedSurface
    strikeSpacing = spacing;
    dipSpacing = spacing;
    // sameGridSpacing = true;
    this.upperTrace = upperTrace;
    this.lowerTrace = lowerTrace;

    init();
  }

  private void init() {

    // compute num subsection of trace
    double aveTraceLength = (upperTrace.length() + lowerTrace.length()) / 2;
    int num = (int) Math.round(aveTraceLength / getAveGridSpacing());

    LocationList resampUpperTrace = Surfaces.resampleTrace(upperTrace, num);
    LocationList resampLowerTrace = Surfaces.resampleTrace(lowerTrace, num);

    double aveDist = 0;
    for (int i = 0; i < resampUpperTrace.size(); i++) {
      Location topLoc = resampUpperTrace.get(i);
      Location botLoc = resampLowerTrace.get(i);
      aveDist += Locations.linearDistanceFast(topLoc, botLoc);
    }
    aveDist /= resampUpperTrace.size();
    int nRows = (int) Math.round(aveDist / getAveGridSpacing()) + 1;

    this.setNumRowsAndNumCols(nRows, num + 1); // note the plus 1!

    // now set the surface locations
    for (int c = 0; c < resampUpperTrace.size(); c++) {
      Location topLoc = resampUpperTrace.get(c);
      Location botLoc = resampLowerTrace.get(c);
      LocationVector dir = LocationVector.create(topLoc, botLoc);
      double horzIncr = dir.Δh / (nRows - 1);
      double vertIncr = dir.Δv / (nRows - 1); // minus sign
      // because vertDist is pos up and depth is pos down

      LocationVector vIncr = LocationVector.create(dir.azimuth, horzIncr, vertIncr);

      // set first point on upper trace for column
      set(0, c, topLoc);

      Location prevLoc = topLoc;
      for (int r = 1; r < nRows; r++) {
        Location nextLoc = Locations.location(prevLoc, vIncr);
        set(r, c, nextLoc);
        prevLoc = nextLoc;
      }
    }

    // set dip
    LocationVector vFirst = LocationVector.create(upperTrace.first(), lowerTrace.first());
    LocationVector vLast = LocationVector.create(upperTrace.last(), lowerTrace.last());
    avgDipRad = (vFirst.plunge() + vLast.plunge()) / 2;

    avgDepth = upperTrace.depth();

    centroid = Locations.centroid(this);
    perimeter = perimeter();
  }

  private LocationList perimeter() {
    LocationList topTr = getRow(0);
    LocationList botTr = getRow(getNumRows() - 1).reverse();
    return LocationList.builder()
        .addAll(topTr)
        .addAll(botTr)
        .add(topTr.get(0))
        .build();
  }

  /**
   * @param args
   */
  public static void main(String[] args) {

    // //these traces come from the file "sub-ch9-z4.in" from the NSHMP 2008
    // model (South America subduction zone)
    // FaultTrace topTrace = new FaultTrace(null);
    //
    // topTrace.addLocation(-33.9450824793, -72.710497654, 10.);
    // topTrace.addLocation(-33.7588818362, -72.6388800048, 10.);
    // topTrace.addLocation(-33.5399999996, -72.5818237309, 10.);
    // topTrace.addLocation(-32.9000000001, -72.4949902347, 10.);
    // topTrace.addLocation(-31.9400000003, -72.4908227539, 10.);
    // topTrace.addLocation(-30.2999999998, -72.3911303709, 10.);
    // topTrace.addLocation(-29.9600000004, -72.3579357913, 10.);
    // topTrace.addLocation(-29.2320361332, -72.2279632571, 10.);
    // topTrace.addLocation(-28.0045263675, -71.9045263674, 10.);
    // topTrace.addLocation(-26.8599999996, -71.6779882813, 10.);
    // topTrace.addLocation(-25.9555615232, -71.584437256, 10.);
    // topTrace.addLocation(-25.1200000003, -71.4298706053, 10.);
    // topTrace.addLocation(-24.7841723632, -71.3841735843, 10.);
    // topTrace.addLocation(-23.0199999996, -71.2632910159, 10.);
    // topTrace.addLocation(-22.22, -71.1589331058, 10.);
    // topTrace.addLocation(-21.8942095629, -71.1322380371, 10.);
    //
    // FaultTrace botTrace = new FaultTrace(null);
    //
    // botTrace.addLocation(-34.3633763178, -71.3297920084, 50.);
    // botTrace.addLocation(-34.02, -71.1744982909, 50.);
    // botTrace.addLocation(-33.7199999999, -71.0928930662, 50.);
    // botTrace.addLocation(-33.4399999999, -71.0548120114, 50.);
    // botTrace.addLocation(-32.6400000002, -70.993572998, 50.);
    // botTrace.addLocation(-32.2813330081, -70.9813330075, 50.);
    // botTrace.addLocation(-31.9799999997, -70.9924999998, 50.);
    // botTrace.addLocation(-31.3460058597, -71.1060070798, 50.);
    // botTrace.addLocation(-31.093046875, -71.1330468748, 50.);
    // botTrace.addLocation(-30.8200000004, -71.1281042477, 50.);
    // botTrace.addLocation(-30.5200000003, -71.0938342283, 50.);
    // botTrace.addLocation(-29.6231591798, -70.9168408208, 50.);
    // botTrace.addLocation(-28.1739575199, -70.5260412594, 50.);
    // botTrace.addLocation(-26.872775879, -70.2672241207, 50.);
    // botTrace.addLocation(-26.2200000001, -70.1064562992, 50.);
    // botTrace.addLocation(-25.9844946293, -70.0644934082, 50.);
    // botTrace.addLocation(-25.7599999999, -70.0563464355, 50.);
    // botTrace.addLocation(-25.1999999999, -70.0945446773, 50.);
    // botTrace.addLocation(-24.9400000001, -70.0942907718, 50.);
    // botTrace.addLocation(-22.5600000003, -69.8766857908, 50.);
    // botTrace.addLocation(-22.0599999998, -69.855518799, 50.);
    // botTrace.addLocation(-21.5010148695, -69.8649200561, 50.);
    //
    // ApproxGriddedSurface surf = new
    // ApproxGriddedSurface(topTrace,botTrace,5);
    //// surf.writeXYZ_toFile("dev/scratch/ned/hereItIs.txt");
    //
    // System.out.println
    // ("AveGridSpacingAlongStrike="+surf.computeAveGridSpacingAlongStrike());
    // System.out.println
    // ("AveGridSpacingDownDip="+surf.computeAveGridSpacingDownDip());
    //
    // surf.test();

    // // lat aligned
    // Location top1 = new Location(20.0,0.0,10.0);
    // Location bot1 = new Location(40.0,0.0,50.0);
    //
    // // ln aligned
    // Location top2 = new Location(40.0,0.0,10.0);
    // Location bot2 = new Location(40.0,20.0,50.0);
    //
    // test1(top1, bot1);
    // test1(top2, bot2);
    //
    // test2(top1, bot1);
    // test2(top2, bot2);
    //
    //
  }

  @Override
  public double strike() {
    return Faults.strike(upperTrace);
  }

  @Override
  public double dip() {
    return avgDipRad * Maths.TO_DEGREES;
  }

  @Override
  public double dipDirection() {
    throw new UnsupportedOperationException("Implement me!");
  }

  @Override
  public double depth() {
    return avgDepth;
  }

  @Override
  public Location centroid() {
    return centroid;
  }

  public LocationList getRow(int row) {
    List<Location> locs = Lists.newArrayList();
    for (int col = 0; col < getNumCols(); col++) {
      locs.add(get(row, col));
    }
    return LocationList.copyOf(locs);
  }

  @Override
  public LocationList getPerimeter() {
    return perimeter;
  }

}
