package gov.usgs.earthquake.nshmp.fault.surface;

import static gov.usgs.earthquake.nshmp.geo.Locations.linearDistanceFast;
import static gov.usgs.earthquake.nshmp.geo.Locations.location;

import java.util.Iterator;
import java.util.List;

import com.google.common.collect.Lists;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.LocationVector;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.Region;

/**
 * Add comments here
 *
 * @author U.S. Geological Survey
 */
class Surfaces {

  /**
   * This returns brief info about this surface
   * @param surf
   */
  public static String getSurfaceInfo(GriddedSurface surf) {
    Location loc1 = surf.getLocation(0, 0);
    Location loc2 = surf.getLocation(0, surf.getNumCols() - 1);
    Location loc3 = surf.getLocation(surf.getNumRows() - 1, 0);
    Location loc4 = surf.getLocation(surf.getNumRows() - 1, surf.getNumCols() - 1);
    return new String("\tRup. Surf. Corner Locations (lat, lon, depth (km):" + "\n\n" + "\t\t" +
        (float) loc1.latitude + ", " + (float) loc1.longitude + ", " + (float) loc1.depth + "\n" +
        "\t\t" + (float) loc2.latitude + ", " + (float) loc2.longitude + ", " + (float) loc2.depth +
        "\n" + "\t\t" + (float) loc3.latitude + ", " + (float) loc3.longitude + ", " +
        (float) loc3.depth + "\n" + "\t\t" + (float) loc4.latitude + ", " + (float) loc4.longitude +
        ", " + (float) loc4.depth + "\n");
  }

  public static double getFractionOfSurfaceInRegion(GriddedSurface surface, Region region) {
    double numInside = 0;
    for (Location loc : surface) {
      if (region.contains(loc)) {
        numInside += 1;
      }
    }
    return numInside / surface.size();
  }

  /**
   * This returns the minimum distance as the minimum among all location pairs
   * between the two surfaces
   * @param surface1 RuptureSurface
   * @param surface2 RuptureSurface
   * @return distance in km
   */
  public static double getMinDistanceBetweenSurfaces(GriddedSurface surface1,
      GriddedSurface surface2) {
    Iterator<Location> it = surface1.iterator();
    double min3dDist = Double.POSITIVE_INFINITY;
    double dist;
    // find distance between all location pairs in the two surfaces
    while (it.hasNext()) { // iterate over all locations in this surface
      Location loc1 = it.next();
      Iterator<Location> it2 = surface2.getEvenlyDiscritizedListOfLocsOnSurface().iterator();
      while (it2.hasNext()) { // iterate over all locations on the user
        // provided surface
        Location loc2 = it2.next();
        dist = Locations.linearDistanceFast(loc1, loc2);
        if (dist < min3dDist) {
          min3dDist = dist;
        }
      }
    }
    return min3dDist;
  }

  /**
   * This resamples the trace into num subsections of equal length (final number
   * of points in trace is num+1). However, note that these subsections of are
   * equal length on the original trace, and that the final subsections will be
   * less than that if there is curvature in the original between the points
   * (e.g., corners getting cut).
   * @param trace
   * @param num - number of subsections
   */
  public static LocationList resampleTrace(LocationList trace, int num) {
    double resampInt = trace.length() / num;
    // FaultTrace resampTrace = new FaultTrace("resampled "+trace.name());
    List<Location> resampLocs = Lists.newArrayList();
    resampLocs.add(trace.first()); // add the first location
    double remainingLength = resampInt;
    Location lastLoc = trace.first();
    int NextLocIndex = 1;
    while (NextLocIndex < trace.size()) {
      Location nextLoc = trace.get(NextLocIndex);
      double length = linearDistanceFast(lastLoc, nextLoc);
      if (length > remainingLength) {
        // set the point
        // LocationVector dir = vector(lastLoc, nextLoc);
        // dir.setHorzDistance(dir.getHorzDistance() * remainingLength /
        // length);
        // dir.setVertDistance(dir.getVertDistance() * remainingLength /
        // length);
        LocationVector dirSrc = LocationVector.create(lastLoc, nextLoc);
        double hDist = dirSrc.Δh * remainingLength / length;
        double vDist = dirSrc.Δv * remainingLength / length;

        LocationVector dir = LocationVector.create(dirSrc.azimuth, hDist, vDist);
        Location loc = location(lastLoc, dir);
        resampLocs.add(loc);
        lastLoc = loc;
        remainingLength = resampInt;
        // Next location stays the same
      } else {
        lastLoc = nextLoc;
        NextLocIndex += 1;
        remainingLength -= length;
      }
    }

    // make sure we got the last one (might be missed because of numerical
    // precision issues?)
    double dist = linearDistanceFast(trace.last(), resampLocs.get(resampLocs.size() - 1));
    if (dist > resampInt / 2) {
      resampLocs.add(trace.last());
    }

    /* Debugging Stuff **************** */
    /*
     * // write out each to check System.out.println("RESAMPLED"); for(int i=0;
     * i<resampTrace.size(); i++) { Location l = resampTrace.getLocationAt(i);
     * System.out.println(l.getLatitude()+"\t"+
     * l.getLongitude()+"\t"+l.getDepth()); }
     *
     * System.out.println("ORIGINAL"); for(int i=0; i<trace.size(); i++) {
     * Location l = trace.getLocationAt(i); System.out.println(l.getLatitude(
     * )+"\t"+l.getLongitude()+"\t"+l.getDepth()); }
     *
     * // write out each to check System.out.println("target resampInt="
     * +resampInt+"\tnum sect="+num); System.out.println("RESAMPLED"); double
     * ave=0, min=Double.MAX_VALUE, max=Double.MIN_VALUE; for(int i=1;
     * i<resampTrace.size(); i++) { double d =
     * Locations.getTotalDistance(resampTrace.getLocationAt(i-1),
     * resampTrace.getLocationAt(i)); ave +=d; if(d<min) min=d; if(d>max) max=d;
     * } ave /= resampTrace.size()-1; System.out.println("ave="+ave+"\tmin="
     * +min+"\tmax="+max+"\tnum pts=" +resampTrace.size());
     *
     *
     * System.out.println("ORIGINAL"); ave=0; min=Double.MAX_VALUE;
     * max=Double.MIN_VALUE; for(int i=1; i<trace.size(); i++) { double d =
     * Locations.getTotalDistance(trace.getLocationAt(i-1),
     * trace.getLocationAt(i)); ave +=d; if(d<min) min=d; if(d>max) max=d; } ave
     * /= trace.size()-1; System.out.println("ave="+ave+"\tmin="+min+"\tmax="
     * +max+"\tnum pts=" +trace.size());
     *
     * /* End of debugging stuff *******************
     */

    // return FaultTrace.create("resampled " + trace.name(),
    // LocationList.create(resampLocs));
    return LocationList.copyOf(resampLocs);
  }

}
