package gov.usgs.earthquake.nshmp.geo.json;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;

import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonSerializationContext;
import com.google.gson.JsonSerializer;

import gov.usgs.earthquake.nshmp.Text;

/**
 * A GeoJSON feature collection.
 *
 * @author U.S. Geological Survey
 */
public final class FeatureCollection {

  private final GeoJson.Type type = GeoJson.Type.FEATURE_COLLECTION;

  private final double[] bbox;
  private final List<Feature> features;

  FeatureCollection(List<Feature> features, double[] bbox) {
    this.features = features;
    this.bbox = bbox;
  }

  /**
   * The top level GeoJSON type: FEATURE or FEATURE_COLLECTION.
   *
   * The 'type' field is overridden on deserialization; method is used
   * internally to check if a GeoJSON was deserialized to the appropriate type.
   */
  GeoJson.Type topLevelType() {
    return type;
  }

  /**
   * The bounding box value array; If bbox is {@code null}, returns
   * Optional.empty(), else returns an {@code Optional<double[]>} copy of the
   * bbox value array.
   */
  public Optional<double[]> bbox() {
    return (bbox == null)
        ? Optional.empty()
        : Optional.of(Arrays.copyOf(bbox, bbox.length));
  }

  /**
   * An immutable list of the features in this collection.
   */
  public List<Feature> features() {
    return List.copyOf(features);
  }

  /**
   * Serialize {@code FeatureCollection} to its equivalent JSON representation.
   *
   * @throws IllegalStateException if builder is empty
   */
  public String toJson() {
    String json = GeoJson.GSON_DEFAULT.toJson(this);
    return Text.cleanCoordinates(json);
  }

  /**
   * Returns the {@code FeatureCollection} as a {@code JsonElement}.
   */
  public JsonElement toJsonTree() {
    return GeoJson.GSON_DEFAULT.toJsonTree(this);
  }

  /**
   * Write the serialized form of this GeoJSON {@code FeatureCollection} to the
   * file at {@code path}.
   */
  public void write(Path path) throws IOException {
    Files.write(path, toJson().getBytes(StandardCharsets.UTF_8));
  }

  static final class Serializer implements JsonSerializer<FeatureCollection> {

    @Override
    public JsonElement serialize(
        FeatureCollection featureCollection,
        java.lang.reflect.Type typeOfSrc,
        JsonSerializationContext context) {

      /*
       * Serialize using Gson that includes Feature adapter but not
       * FeatureCollection adapter that would trigger recursive calls to this
       * serializer. This exists to strip the 'bbox' member if unused.
       */
      JsonObject jObj = GeoJson.GSON_FEATURE
          .toJsonTree(featureCollection)
          .getAsJsonObject();

      if (featureCollection.bbox == null) {
        jObj.remove("bbox");
      }
      return jObj;
    }
  }

}
