package gov.usgs.earthquake.nshmp.geo.json;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkState;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;

import com.google.common.base.CaseFormat;
import com.google.common.base.Charsets;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonIOException;
import com.google.gson.TypeAdapter;
import com.google.gson.annotations.JsonAdapter;
import com.google.gson.stream.JsonReader;
import com.google.gson.stream.JsonWriter;

import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.geo.Bounds;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;

/**
 * Entry point for creating and parsing <a href="http://geojson.org">GeoJSON</a>
 * features and feature collections.
 *
 * <p>To create a GeoJSON feature collection string, use a builder and add
 * features per the example below:
 *
 * <pre>
 * String geojson = GeoJson.builder()
 *     .add(Feature.point(Location.create(34, -117))
 *         .id("featureId")
 *         .properties(Map.of(
 *             "title", "Feature Title",
 *             "color", "#ff0080"))
 *         .build())
 *     .add(Feature.polygon(
 *         LocationList.create(
 *             Location.create(34, -117),
 *             Location.create(35, -118),
 *             Location.create(37, -116),
 *             Location.create(38, -117)))
 *         .id(3)
 *         .build())
 *     .build()
 *     .toJson();
 * </pre>
 *
 * <p>Where the GeoJSON string created is:
 *
 * <pre>
 * {
 *   "type": "FeatureCollection",
 *   "features": [
 *     {
 *       "type": "Feature",
 *       "id": "featureId",
 *       "geometry": {
 *         "type": "Point",
 *         "coordinates": [-117.0, 34.0]
 *       },
 *       "properties": {
 *         "title": "Feature Title",
 *         "color": "#ff0080"
 *       }
 *     },
 *     {
 *       "type": "Feature",
 *       "id": 3,
 *       "geometry": {
 *         "type": "Polygon",
 *         "coordinates": [
 *           [
 *             [-117.0, 34.0],
 *             [-118.0, 35.0],
 *             [-116.0, 37.0],
 *             [-117.0, 38.0]
 *           ]
 *         ]
 *       },
 *       "properties": {}
 *     }
 *   ]
 * }
 * </pre>
 *
 * <p>Note in the example above that features have their own geometry-specific
 * builders. Features supply {@link Feature#toJson()},
 * {@link Feature#toJsonTree()} and {@link Feature#write(Path)} methods to
 * output single-feature GeoJSON directly.
 *
 * <p>Parse GeoJSON to a feature or feature collection using static
 * {@code from*} methods as follows:
 *
 * <pre>
 * Feature f = GeoJson.from(String | Path | Url).toFeature();
 * FeatureCollection fc = GeoJson.from(String | Path | Url).toFeatureCollection();
 * </pre>
 *
 * <p>Once parsed, the feature geometry may be accessed as objects in the
 * {@code geo} package (e.g. {@code Location}, {@code LocationList}, etc...).
 * This requires some prior knowledge of the contents of the parsed GeoJSON.
 *
 * <p>This class does not allow the creation of empty feature collections or
 * features with null geometry.
 *
 * @author U.S. Geological Survey
 */
public abstract class GeoJson {

  private GeoJson() {}

  /**
   * Deserialize GeoJson to the supplied type.
   */
  abstract <T> T read(Class<T> T);

  /**
   * Return this GeoJson as a {@code Feature}.
   *
   * @throws UnsupportedOperationException if the content of this GeoJson is not
   *         a feature
   */
  public Feature toFeature() {
    Feature f = read(Feature.class);
    if (f.topLevelType() != Type.FEATURE) {
      throw new UnsupportedOperationException("GeoJSON is not a Feature");
    }
    return f;
  }

  /**
   * Return this GeoJson as a {@code FeatureCollection}.
   *
   * @throws UnsupportedOperationException if the content of this GeoJson is not
   *         a feature collection
   */
  public FeatureCollection toFeatureCollection() {
    FeatureCollection fc = read(FeatureCollection.class);
    if (fc.topLevelType() != Type.FEATURE_COLLECTION) {
      throw new UnsupportedOperationException("GeoJSON is not a FeatureCollection");
    }
    return fc;
  }

  /*
   * Developer notes:
   *
   * If depths are considered, should they be negative? Per the GeoJSON spec:
   * "Altitude or elevation MAY be included as an optional third element."
   *
   * Serialization note: Although we generally want to avoid null values in
   * GeoJSON properties members, we don't want to completely prevent their use.
   * However, when serializeNulls=true, all custom adapters are skipped so we
   * end up with optional GeoJSON member clutter, for example 'bbox' and 'id'.
   * The workaround is to use two distinct Gson instances (to avoid recursive
   * calls to delegate serializers) where we explicitly remove null fields using
   * JsonObject.remove().
   */

  /**
   * Create a GeoJSON from a string.
   *
   * @param json string to read
   */
  public static GeoJson from(String json) {
    return new FromString(json);
  }

  /**
   * Create a GeoJSON from a file path.
   *
   * @param json file path to read
   */
  public static GeoJson from(Path json) {
    return new FromPath(json);
  }

  /**
   * Create a GeoJSON from a URL. Use this method when reading resources from a
   * JAR.
   *
   * @param json URL to read
   */
  public static GeoJson from(URL json) {
    return new FromUrl(json);
  }

  /**
   * Checks that bounding box location list contains 5 coordinates defining a
   * mercator lat/lon rectangle.
   *
   * @param locations to check
   * @return the supplied locaiton list
   */
  public static LocationList validateBoundingBox(LocationList locations) {

    /* Rectangular linear ring check. */
    checkArgument(
        locations.size() == 5,
        "Extents polygon must contain 5 coordinates: %s",
        locations);

    /* Order agnostic rectilinear lat-lon check. */
    Location p1 = locations.get(0);
    Location p2 = locations.get(1);
    Location p3 = locations.get(2);
    Location p4 = locations.get(3);
    boolean rectangular = (p1.latitude == p2.latitude)
        ? (p3.latitude == p4.latitude &&
            p1.longitude == p4.longitude &&
            p2.longitude == p3.longitude)
        : (p1.latitude == p4.latitude &&
            p2.latitude == p3.latitude &&
            p1.longitude == p2.longitude &&
            p3.longitude == p4.longitude);
    checkArgument(
        rectangular,
        "Extents polygon does not define a lat-lon Mercator rectangle: %s",
        locations);

    return locations;
  }

  /**
   * Checks that the supplied location list defines a rectangular lat/lon area
   * and returns a slightly expanded version that will enclosepoints on the
   * borders of the area. Points coincident with east and north edges of
   * polygons are considered inside as a result.
   */
  public static LocationList expandRectangularArea(LocationList locations) {
    Bounds b = validateBoundingBox(locations).bounds();

    double lat1 = b.min.latitude;
    double lat2 = b.max.latitude;
    double lon1 = b.min.longitude;
    double lon2 = b.max.longitude;

    // checkArgument(lat1 != lat2, "Input lats cannot be the same");
    // checkArgument(lon1 != lon2, "Input lons cannot be the same");

    double minLat = Math.min(lat1, lat2);
    double minLon = Math.min(lon1, lon2);
    double maxLat = Math.max(lat1, lat2);
    double maxLon = Math.max(lon1, lon2);
    double offset = Locations.TOLERANCE;

    // ternaries prevent exceedance of max lat-lon values
    maxLat += (maxLat <= 90.0 - offset) ? offset : 0.0;
    maxLon += (maxLon <= 180.0 - offset) ? offset : 0.0;
    minLat -= (minLat >= -90.0 + offset) ? offset : 0.0;
    minLon -= (minLon >= -180.0 + offset) ? offset : 0.0;

    LocationList expanded = LocationList.of(
        Location.create(minLon, minLat),
        Location.create(minLon, maxLat),
        Location.create(maxLon, maxLat),
        Location.create(maxLon, minLat));

    // initBordered(locs, MERCATOR_LINEAR);
    return expanded;
  }

  /**
   * A reusable GeoJSON builder.
   */
  public static Builder builder() {
    return new Builder();
  }

  /* Default Gson with FeatureCollection adapter. */
  static final Gson GSON_DEFAULT = new GsonBuilder()
      .registerTypeAdapter(
          FeatureCollection.class,
          new FeatureCollection.Serializer())
      .setPrettyPrinting()
      .disableHtmlEscaping()
      .serializeNulls()
      .create();

  /* Gson with Feature adapter. */
  static final Gson GSON_FEATURE = new GsonBuilder()
      .registerTypeAdapter(
          Feature.class,
          new Feature.Serializer())
      .setPrettyPrinting()
      .disableHtmlEscaping()
      .serializeNulls()
      .create();

  /** A GeoJSON builder. */
  public static class Builder {

    private double[] bbox;
    private List<Feature> features;

    private boolean built = false;

    private Builder() {
      features = new ArrayList<>();
    }

    /**
     * Add a feature. Use static factory methods in {@link Feature} to create
     * new features.
     *
     * @param feature to add
     * @return this builder
     * @see Feature
     */
    public Builder add(Feature feature) {
      features.add(feature);
      return this;
    }

    /**
     * Set the optional {@code bbox} (bounding box) field of this feature
     * collection. See the GeoJSON <a href="http://geojson.org">
     * specification</a> for details on bounding boxes.
     *
     * @param bbox to set
     * @return this builder
     */
    public Builder bbox(double[] bbox) {
      checkArgument(bbox.length == 4, "'bbox' array must have 4 values");
      this.bbox = bbox;
      return this;
    }

    /**
     * Returns the {@code FeatureCollection}
     */
    public FeatureCollection build() {
      checkState(!built, "This builder has already been used");
      built = true;
      checkState(features.size() > 0, "GeoJSON is empty");
      return new FeatureCollection(features, bbox);
    }

  }

  /** GeoJSON type identifier. */
  @JsonAdapter(TypeSerializer.class)
  public enum Type {

    /** GeoJSON {@code Feature} object. */
    FEATURE,

    /** GeoJSON {@code FeatureCollection} object. */
    FEATURE_COLLECTION,

    /** GeoJSON {@code Point} geometry. */
    POINT,

    /** GeoJSON {@code MultiLineString} geometry */
    MULTI_LINE_STRING,

    /** GeoJSON {@code LineString} geometry. */
    LINE_STRING,

    /** GeoJSON {@code Polygon} geometry. */
    POLYGON;
  }

  /* Create a class with delegate to attach using JsonAdapter annotation. */
  static final class TypeSerializer extends TypeAdapter<Type> {

    TypeAdapter<Type> delegate = Text.enumSerializer(
        Text.enumStringConverter(Type.class, CaseFormat.UPPER_CAMEL));

    @Override
    public void write(JsonWriter out, Type value) throws IOException {
      delegate.write(out, value);
    }

    @Override
    public Type read(JsonReader in) throws IOException {
      return delegate.read(in);
    }
  }

  private static final class FromString extends GeoJson {

    final String json;

    FromString(String json) {
      this.json = json;
    }

    @Override
    <T> T read(Class<T> classOfT) {
      return GSON_DEFAULT.fromJson(json, classOfT);
    }
  }

  private static final class FromUrl extends GeoJson {

    final URL json;

    FromUrl(URL json) {
      this.json = json;
    }

    @Override
    <T> T read(Class<T> classOfT) {
      try (BufferedReader br =
          new BufferedReader(new InputStreamReader(json.openStream(), Charsets.UTF_8))) {
        return GSON_DEFAULT.fromJson(br, classOfT);
      } catch (IOException ioe) {
        throw new JsonIOException(ioe);
      }
    }
  }

  private static final class FromPath extends GeoJson {

    final Path json;

    FromPath(Path json) {
      this.json = json;
    }

    @Override
    <T> T read(Class<T> classOfT) {
      try (BufferedReader br = Files.newBufferedReader(json)) {
        return GSON_DEFAULT.fromJson(br, classOfT);
      } catch (IOException ioe) {
        throw new JsonIOException(ioe);
      }
    }
  }

}
