package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.ALASKA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.CASCADIA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.GLOBAL;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.SLAB;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.min;
import static java.lang.Math.pow;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Abstract implementation of the Abrahamson &amp; Gülerce NGA-Subduction ground
 * motion model. This model supports both slab and interface type events.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Implementation notes:</b><ul>
 *
 * <li>Implementation currently supports a global median model, as well as
 * Alaska and Cascadia regional models.</li>
 *
 * <li>Coefficients for 0.01s are used for PGA.</li>
 *
 * <li>Unlike BcHydro, this model does not include a forearc-backarc term.</li>
 *
 * <li>Mean model does not include aftershock term.</li>
 *
 * <li>{@code zTor} is interpreted as hypocentral depth and is only used for
 * slab events; it is limited to 120 km, consistent with other subduction
 * models.</li>
 *
 * <li>Basin scaling only amplifies long period ground motion when z2.5 is
 * greater than the reference value computed from Vs30.</li>
 *
 * <li>The aleatory variability model only includes the linear φ1 model
 * appropriate for the global, Alaska and Cascadia models.</li>
 *
 * </ul>
 *
 * <p><b>Reference:</b> Abrahamson, N., and Gülerce, Z., 2020, Regionalized
 * ground-motion models for subduction earthquakes based on the NGA-SUB
 * database. Report no. 2020/25. Berkeley, CA: PEER, 219 pp.
 *
 * <p><b>Component:</b> Geometric mean of two horizontal components
 *
 * @author U.S. Geological Survey
 * @see Gmm#AG_20_GLOBAL_INTERFACE
 * @see Gmm#AG_20_GLOBAL_SLAB
 * @see Gmm#AG_20_CASCADIA_INTERFACE
 * @see Gmm#AG_20_CASCADIA_ADJ_INTERFACE
 * @see Gmm#AG_20_CASCADIA_SLAB
 * @see Gmm#AG_20_CASCADIA_ADJ_SLAB
 * @see Gmm#AG_20_CASCADIA_INTERFACE_BASIN
 * @see Gmm#AG_20_CASCADIA_ADJ_INTERFACE_BASIN
 * @see Gmm#AG_20_CASCADIA_SLAB_BASIN
 * @see Gmm#AG_20_CASCADIA_ADJ_SLAB_BASIN
 * @see Gmm#AG_20_ALASKA_INTERFACE
 * @see Gmm#AG_20_ALASKA_SLAB
 * @see Gmm#AG_20_ALASKA_ADJ_INTERFACE
 * @see Gmm#AG_20_ALASKA_ADJ_SLAB
 */
public abstract class AbrahamsonGulerce_2020 implements GroundMotionModel {

  static final String NAME = "Abrahamson & Gülerce (2020)";

  /*
   * Developer notes:
   *
   * TODO per-implementation constraints? (slab vs interface depth limits)
   *
   * TODO NAA recommendation for PGV
   *
   * TODO vLin, b values for fSite are from CB14 with 0.6s and 6s values from
   * BcHydro. There is an inconsistency between CB14 and BcHydro at 0.02s, and
   * 0.03s is missing from BcHydro. vLin is rounded in CB14.
   *
   * TODO should Vs30 in non-lin sigma model be Vs*?
   *
   * TODO what is horizontal component of this and other NGA-Sub models?
   *
   * TODO It's not clear what the order of operations is on Δc1s eqn 3.5
   */

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(5.0, 9.5))
      .set(RRUP, Range.closed(0.0, 1000.0))
      .set(ZTOR, Earthquakes.SLAB_DEPTH_RANGE)
      .set(VS30, Range.closed(150.0, 1000.0))
      .build();

  static final CoefficientContainer COEFFS = new CoefficientContainer("AG20.csv");

  private static final double C1S = 7.5;
  private static final double C4 = 10.0;
  private static final double C = 1.88;
  private static final double N = 1.18;

  private static final double VSS_MAX = 1000.0; // Errata 2021/7/12
  private static final double ΔC1_SLAB = -0.3;
  private static final double VS30_ROCK = 1000.0;

  private static final double D0 = 0.47;

  private static final class Coefficients {

    final double c1i, c1s;
    final double a1, a2, a3, a4, a5, a6, a8, a9, a10, a11, a12, a13, a14, a39, a45;
    final double d1, d2, ρB, ρW;
    final double vlin, b;
    final double a1Adj;

    Coefficients(Imt imt, CoefficientContainer cc, SubductionZone zone) {
      Map<String, Double> coeffs = cc.get(imt);

      /* Regional adjustments (Global) */
      double a1r = coeffs.get("a1");
      double a1AdjRegional = 0.0;
      double a6r = 0.0;
      double a12r = 0.0;
      double c1sr = 7.5;

      if (zone == ALASKA) {
        a1r = coeffs.get("a31");
        a1AdjRegional = coeffs.get("cAk");
        a6r = coeffs.get("a24");
        a12r = coeffs.get("a17");
        c1sr = 7.9;
      } else if (zone == CASCADIA) {
        a1r = coeffs.get("a32");
        a1AdjRegional = coeffs.get("cCasc");
        a6r = coeffs.get("a25");
        a12r = coeffs.get("a18");
        c1sr = 7.1;
      }

      /* M-scaling break points */
      c1i = coeffs.get("c1i");
      c1s = c1sr;

      a1 = a1r;
      a1Adj = a1AdjRegional;
      a2 = coeffs.get("a2");
      a3 = coeffs.get("a3");
      a4 = coeffs.get("a4");
      a5 = coeffs.get("a5");
      a6 = coeffs.get("a6") + a6r;
      a8 = coeffs.get("a8");
      a9 = coeffs.get("a9");
      a10 = coeffs.get("a10");
      a11 = coeffs.get("a11");
      a12 = coeffs.get("a12") + a12r;
      a13 = coeffs.get("a13");
      a14 = coeffs.get("a14");
      a39 = coeffs.get("a39");
      a45 = coeffs.get("a45");
      d1 = coeffs.get("d1");
      d2 = coeffs.get("d2");
      ρB = coeffs.get("rhoB");
      ρW = coeffs.get("rhoW");
      vlin = coeffs.get("vlin");
      b = coeffs.get("b");
    }
  }

  private final Coefficients coeffs;
  private final Coefficients coeffsPGA;

  AbrahamsonGulerce_2020(Imt imt, SubductionZone zone) {
    coeffs = new Coefficients(imt, COEFFS, zone);
    coeffsPGA = new Coefficients(PGA, COEFFS, zone);
  }

  @Override
  public final LogicTree<GroundMotion> calc(GmmInput in) {

    double pgaRock = 0.0;
    if (in.vs30 < coeffs.vlin) {
      pgaRock = exp(calcMean(coeffsPGA, slab(), basin(), adjust(), 0.0,
          in.Mw, in.rRup, in.zTor, VS30_ROCK, in.z2p5));
    }
    double μ = calcMean(coeffs, slab(), basin(), adjust(), pgaRock,
        in.Mw, in.rRup, in.zTor, in.vs30, in.z2p5);

    double σ = calcSigma(coeffs, coeffsPGA, pgaRock, in.rRup, in.vs30);

    return GroundMotions.createTree(μ, σ);
  }

  abstract boolean slab();

  abstract boolean basin();

  abstract boolean adjust();

  private static double calcMean(
      Coefficients c,
      boolean slab,
      boolean basin,
      boolean adjust,
      double pgaRock,
      double Mw,
      double rRup,
      double zTor,
      double vs30,
      double z2p5) {

    /* Finite fault term (Eq. 3.2): ln(Rrup + Hff) */
    double lnRrupHff = log(rRup + C4 * exp((Mw - 6.0) * c.a9));

    /* Magnitude scaling */
    /* Eq. 3.4 */
    double C1 = (slab ? c.c1s : c.c1i);
    /* Eq. 3.3 */
    double c13m = c.a13 * (10 - Mw) * (10 - Mw);
    double c4s = c.a4 + (slab ? c.a45 : 0.0);
    // Norm uses Mw < c1, paper uses Mw <= c1
    double fMag = (Mw <= C1 ? c4s : c.a5) * (Mw - C1) + c13m;

    /* Additional slab scaling */
    /* Eq. 3.5, errata 2021/7/12 */
    double fSlab = 0.0;
    if (slab) {
      double Δc1s = C1S - c.c1s; // C1S - (c.c1s
      fSlab = c.a10 + (c.a4 + c.a45) * (c.c1s - C1S) + c.a14 * lnRrupHff;
    }

    /* Depth scaling: slab only */
    /* Eq. 3.6 */
    double fDepth = 0.0;
    if (slab) {
      // norm uses zTor < 50 and zTor < 200, paper uses zTor <= 50 and zTor >=
      // 50, this should not matter
      double ΔzTor = (zTor <= 200.0) ? (zTor - 50.0) : 150.0;
      fDepth += ((zTor <= 50.0) ? c.a8 : c.a11) * ΔzTor;
    }

    /* Site-response scaling */
    /* Eq. 3.8 */
    double vsS = getVsStar(vs30);
    /* Eq. 3.7 */
    double fSite = c.a12 * log(vsS / c.vlin);
    if (vsS < c.vlin) { // whether or not to use pgaRock
      fSite += -c.b * log(pgaRock + C) + c.b * log(pgaRock + C * pow((vsS / c.vlin), N));
    } else {
      /* pgaRock loop, vs=1000 > vlinPGA=865 */
      fSite += c.b * N * log(vsS / c.vlin);
    }

    /* Basin-depth scaling */
    // TODO check ge geq comparison operators; this is Cascadia model
    double fBasin = 0.0;
    if (basin && !Double.isNaN(z2p5)) {
      double z2p5m = z2p5 * 1000.0;
      /* Eq. 2.1 - this is lnZ2p5ref, paper uses all gt/lt, no ge/le */
      // Recommend using vs30 >= 570, otherwise the value for vs30 = 570
      // is 7.5983 (for vs30 = 569, value = 7.5999)
      /*
       * Note: PEER report (and xlsx) uses factor rounded to 0.88, fortran code
       * uses the calculated value. Using 0.878 would improve the match of
       * lnZ2p5ref to the fortran code (within 0.01%).
       */
      double lnZ2p5ref = (vs30 >= 570)
          ? 7.6
          : (vs30 > 200)
              ? 8.52 - 0.88 * log(vs30 / 200.0)
              : 8.52;

      // TODO commented out is PEER report but ln(z')>1 is odd
      // double zPrime = log((z2p5m + 50.0) / (exp(lnZ2p5ref) + 50.0));
      // fBasin = (zPrime > 1) ? c.a39 * zPrime : 0.0;
      /* Eq. 3.9 */
      double zPrime = (z2p5m + 50.0) / (exp(lnZ2p5ref) + 50.0);
      /* Eq. 3.11 - Cascadia basin scaling (Errata 2021/7/12) */
      fBasin = (log(zPrime) > 0) ? c.a39 * log(zPrime) : 0.0;
    }

    /* Eq. 3.1 */
    double μ = c.a1 +
        (c.a2 + c.a3 * (Mw - 7.0)) * lnRrupHff +
        c.a6 * rRup + fMag + fDepth +
        fSite + fSlab +
        fBasin; // + fAs * c.a15; no aftershock term

    if (adjust) {
      μ += c.a1Adj;
    }

    return μ;
  }

  private static double getVsStar(double vs30) {
    return min(vs30, VSS_MAX);
  }

  private static double calcSigma(
      Coefficients c,
      Coefficients cPga,
      double pgaRock,
      double rRup,
      double vs30) {

    double τlin = D0;
    double τSq = τlin * τlin;

    double φSq = φlinSq(c.d1, c.d2, rRup);

    double vsS = getVsStar(vs30);

    /* Add nonlinear effects. */
    if (vsS < c.vlin) {

      // TODO eq. 5.9 is vs30 <= c.vlin, but this is inconsistent with
      // fSite vs30 < c.vlin for nonlinear behavior

      /* φamp^2 = 0.3 * 0.3 = 0.09 */
      double φbSq = φSq - 0.09;

      // System.out.println(φbSq);

      double φlinSqPga = φlinSq(cPga.d1, cPga.d2, rRup);
      double φbSqPga = φlinSqPga - 0.09;

      double dSite = c.b * pgaRock *
          ((1.0 / (pgaRock + C * pow((vs30 / c.vlin), N))) - (1.0 / (pgaRock + C)));
      double dSiteSq = dSite * dSite;

      φSq += dSiteSq * φbSqPga + 2 * dSite * sqrt(φbSqPga) * sqrt(φbSq) * c.ρW;

      /* τlinPga = τlin = D0 */
      τSq += dSiteSq * τSq + 2 * dSite * τSq * c.ρB;

      // System.out.println(φSq);
      // System.out.println(τSq);
    }

    return sqrt(τSq + φSq);
  }

  private static double φlinSq(double d1, double d2, double rRup) {
    /* Eq. 5.2 */
    double φ = d1;
    if (rRup > 450.0) {
      φ += d2;
    } else if (rRup >= 150.0) {
      φ += d2 * (rRup - 150.0) / 300.0;
    }
    return φ;

  }

  private static abstract class Global extends AbrahamsonGulerce_2020 {
    static final SubductionZone ZONE = GLOBAL;
    static final String NAME = AbrahamsonGulerce_2020.NAME + " : " + ZONE;

    Global(Imt imt) {
      super(imt, ZONE);
    }

    @Override
    boolean slab() {
      return false;
    }

    @Override
    boolean basin() {
      return false;
    }

    @Override
    boolean adjust() {
      return false;
    }
  }

  private static abstract class Cascadia extends AbrahamsonGulerce_2020 {
    static final SubductionZone ZONE = CASCADIA;
    static final String NAME = AbrahamsonGulerce_2020.NAME + " : " + ZONE;

    Cascadia(Imt imt) {
      super(imt, ZONE);
    }

    @Override
    boolean slab() {
      return false;
    }

    @Override
    boolean basin() {
      return false;
    }

    @Override
    boolean adjust() {
      return true;
    }
  }

  private static abstract class Alaska extends AbrahamsonGulerce_2020 {
    static final SubductionZone ZONE = ALASKA;
    static final String NAME = AbrahamsonGulerce_2020.NAME + " : " + ZONE;

    Alaska(Imt imt) {
      super(imt, ZONE);
    }

    @Override
    boolean slab() {
      return false;
    }

    @Override
    boolean basin() {
      return false;
    }

    @Override
    boolean adjust() {
      return true;
    }
  }

  static final class GlobalInterface extends Global {
    static final String NAME = Global.NAME + " : " + INTERFACE;

    GlobalInterface(Imt imt) {
      super(imt);
    }
  }

  static final class GlobalSlab extends Global {
    static final String NAME = Global.NAME + " : " + SLAB;

    GlobalSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static class CascadiaAdjustedInterface extends Cascadia {
    static final String NAME = Cascadia.NAME + " (adj) " + " : " + INTERFACE;

    CascadiaAdjustedInterface(Imt imt) {
      super(imt);
    }
  }

  static class CascadiaAdjustedSlab extends Cascadia {
    static final String NAME = Cascadia.NAME + " (adj) " + " : " + SLAB;

    CascadiaAdjustedSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static class CascadiaInterface extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + INTERFACE;

    CascadiaInterface(Imt imt) {
      super(imt);
    }

    @Override
    boolean adjust() {
      return false;
    }
  }

  static class CascadiaSlab extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + SLAB;

    CascadiaSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }

    @Override
    boolean adjust() {
      return false;
    }
  }

  static final class CascadiaAdjustedInterfaceBasin extends CascadiaAdjustedInterface {
    static final String NAME = CascadiaAdjustedInterface.NAME + " (basin)";

    CascadiaAdjustedInterfaceBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class CascadiaAdjustedSlabBasin extends CascadiaAdjustedSlab {
    static final String NAME = CascadiaAdjustedSlab.NAME + " (basin)";

    CascadiaAdjustedSlabBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class CascadiaInterfaceBasin extends CascadiaInterface {
    static final String NAME = CascadiaInterface.NAME + " (basin)";

    CascadiaInterfaceBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class CascadiaSlabBasin extends CascadiaSlab {
    static final String NAME = CascadiaSlab.NAME + " (basin)";

    CascadiaSlabBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class AlaskaAdjustedInterface extends Alaska {
    static final String NAME = Alaska.NAME + " (adj) " + " : " + INTERFACE;

    AlaskaAdjustedInterface(Imt imt) {
      super(imt);
    }
  }

  static final class AlaskaAdjustedSlab extends Alaska {
    static final String NAME = Alaska.NAME + " (adj) " + " : " + SLAB;

    AlaskaAdjustedSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static final class AlaskaInterface extends Alaska {
    static final String NAME = Alaska.NAME + " : " + INTERFACE;

    AlaskaInterface(Imt imt) {
      super(imt);
    }

    @Override
    boolean adjust() {
      return false;
    }
  }

  static final class AlaskaSlab extends Alaska {
    static final String NAME = Alaska.NAME + " : " + SLAB;

    AlaskaSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }

    @Override
    boolean adjust() {
      return false;
    }
  }
}
