package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGV;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P02;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P03;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P05;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.pow;

import java.util.Map;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Abstract implementation of the subduction ground motion model created for BC
 * Hydro, Canada, by Addo, Abrahamson, &amp; Youngs (2012). An update to this
 * implementation is documented in Abrahamson, Gregor, &amp; Addo (2016). This
 * model supports both slab and interface type events.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Changes due to Abrahamson et al. (2016) update:</b><ul>
 *
 * <li>Added coefficients for 0.02s that are identical to PGA. 0.02s ground
 * motions had previously had been computed via interpolation between 0.01s and
 * 0.05s.</li>
 *
 * <li>Slab {@code zTor} adjusted from 125 km to 120 km.</li>
 *
 * <li>Changed interpolation of the {@code ΔC1} magnitude-scaling break term to
 * log spectral periods and linear values; prior implementation was linear;
 * change affects the following spectral periods: 0.4s, 0.6s (disabled), 0.75s,
 * 1.5s, and 2.5s wherein the {@code ΔC1} value has decreased by 0.01</li></ul>
 *
 * <p><b>Implementation notes:</b><ul>
 *
 * <li>Forearc/backarc term is considered.</li>
 *
 * <li>{@code zTor} is interpreted as hypocentral depth and is only used for
 * slab events; it is limited to 120 km, consistent with other subduction
 * models.</li>
 *
 * <li>Only the middle branch of the {@code ΔC1} magnitude-scaling break term
 * for interface events is considered; the {@code ΔC1} term for slab events is
 * fixed at -0.3.</li>
 *
 * <li>Support for spectral period 0.01s is provided using the same coefficients
 * as bounding periods PGA and 0.02s.</li>
 *
 * <li>Support for spectral periods 0.03s is provided via interpolation of
 * ground motion and sigma of adjacent periods.</li></li>
 *
 * <li>Support for basin effects is provided per documentation of the 2018
 * Conterminous U.S. NSHM update (<a
 * href="https://doi.org/10.1177%2F8755293019878199">Petersen et al.,
 * 2020</a>).</li>
 *
 * <li>Support for PGV is provided via the Abrahamson &amp; Bhasin (2020)
 * conditional PGV model for crustal earthquakes.</li></ul>
 *
 * <p><b>Reference:</b> Abrahamson, N., Gregor, N., and Addo, K., 2016, BC Hydro
 * ground motion prediction equations for subduction earthquakes: Earthquake
 * Spectra, v. 32, n. 1, p. 23-44.
 *
 * <p><b>doi:</b> <a href="http://dx.doi.org/10.1193/051712EQS188MR">
 * 10.1193/051712EQS188MR</a>
 *
 * <p><b>Reference:</b> Addo, K., Abrahamson, N., and Youngs, R., (BC Hydro),
 * 2012, Probabilistic seismic hazard analysis (PSHA) model—Ground motion
 * characterization (GMC) model: Report E658, v. 3, November.
 *
 * <p><b>Component:</b> Geometric mean of two horizontal components
 *
 * @author U.S. Geological Survey
 * @see Gmm#BCHYDRO_12_INTERFACE
 * @see Gmm#BCHYDRO_12_SLAB
 * @see Gmm#BCHYDRO_12_INTERFACE_BASIN
 * @see Gmm#BCHYDRO_12_SLAB_BASIN
 * @see Gmm#BCHYDRO_12_INTERFACE_BACKARC
 * @see Gmm#BCHYDRO_12_SLAB_BACKARC
 * @see Gmm#BCHYDRO_12_INTERFACE_BASIN_BACKARC
 * @see Gmm#BCHYDRO_12_SLAB_BASIN_BACKARC
 */
public abstract class BcHydro_2012 implements GroundMotionModel {

  static final String NAME = "BC Hydro (2012)";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(5.0, 9.5))
      .set(RRUP, Range.closed(0.0, 1000.0))
      .set(ZTOR, Earthquakes.SLAB_DEPTH_RANGE)
      .set(VS30, Range.closed(150.0, 1000.0))
      .build();

  static final CoefficientContainer COEFFS = new CoefficientContainer("BCHydro12.csv");

  private static final double C1 = 7.8;
  private static final double T3 = 0.1;
  private static final double T4 = 0.9;
  private static final double T5 = 0.0;
  private static final double T9 = 0.4;
  private static final double C4 = 10.0;
  private static final double C = 1.88;
  private static final double N = 1.18;
  private static final double VSS_MAX = 1000.0;
  private static final double SIGMA = 0.74;
  private static final double ΔC1_SLAB = -0.3;
  private static final double VS30_ROCK = 1000.0;

  private static final Map<Imt, Range<Imt>> INTERPOLATED_IMTS = Maps.immutableEnumMap(
      ImmutableMap.of(SA0P03, Range.closed(SA0P02, SA0P05)));

  private static final class Coefficients {

    final Imt imt;
    final double vlin, b, θ1, θ2, θ6, θ7, θ8, θ10, θ11, θ12, θ13, θ14, θ15, θ16, ΔC1mid;

    // not currently used
    // final double dC1lo, dC1hi;

    Coefficients(Imt imt, CoefficientContainer cc) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      vlin = coeffs.get("vlin");
      b = coeffs.get("b");
      θ1 = coeffs.get("t1");
      θ2 = coeffs.get("t2");
      θ6 = coeffs.get("t6");
      θ7 = coeffs.get("t7");
      θ8 = coeffs.get("t8");
      θ10 = coeffs.get("t10");
      θ11 = coeffs.get("t11");
      θ12 = coeffs.get("t12");
      θ13 = coeffs.get("t13");
      θ14 = coeffs.get("t14");
      θ15 = coeffs.get("t15");
      θ16 = coeffs.get("t16");

      ΔC1mid = coeffs.get("dC1mid");
    }
  }

  private final Coefficients coeffs;
  private final Coefficients coeffsPGA;
  private final CampbellBozorgnia_2014 cb14;

  // interpolatedGmm = null if !interpolated
  private final boolean interpolated;
  private final GroundMotionModel interpolatedGmm;
  private final Gmm subtype;

  BcHydro_2012(final Imt imt, Gmm subtype) {
    this.subtype = subtype;
    coeffs = new Coefficients(imt, COEFFS);
    coeffsPGA = new Coefficients(PGA, COEFFS);
    cb14 = (CampbellBozorgnia_2014) Gmm.CB_14_BASE.instance(imt);
    interpolated = INTERPOLATED_IMTS.containsKey(imt);
    interpolatedGmm = interpolated
        ? new InterpolatedGmm(subtype, imt, INTERPOLATED_IMTS.get(imt))
        : null;
  }

  @Override
  public final LogicTree<GroundMotion> calc(GmmInput in) {

    if (coeffs.imt == PGV) {
      GroundMotion pgv = UsgsPgvSupport.calcAB20Pgv(subtype, in);
      return GroundMotions.createTree(pgv.mean(), pgv.sigma());
    }

    if (interpolated) {
      return interpolatedGmm.calc(in);
    }

    double pgaRock =
        exp(calcMean(coeffsPGA, isSlab(), 0.0, in.Mw, in.rRup, in.zTor, VS30_ROCK, isBackarc()));
    double μAsk =
        calcMean(coeffs, isSlab(), pgaRock, in.Mw, in.rRup, in.zTor, in.vs30, isBackarc());

    /* Add (possibly depth-tapered) CB14 deep basin term. */
    if (deepBasinEffect()) {
      μAsk += cb14.deepBasinScaling(in.z2p5);
    }

    return GroundMotions.createTree(μAsk, SIGMA);
  }

  abstract boolean isSlab();

  abstract boolean deepBasinEffect();

  abstract boolean isBackarc();

  private static final double calcMean(
      final Coefficients c,
      final boolean slab,
      final double pgaRock,
      final double Mw,
      final double rRup,
      final double zTor,
      final double vs30,
      final boolean backarc) {

    double ΔC1 = (slab ? ΔC1_SLAB : c.ΔC1mid);
    double mCut = C1 + ΔC1;
    double t13m = c.θ13 * (10 - Mw) * (10 - Mw);
    double fMag = (Mw <= mCut ? T4 : T5) * (Mw - mCut) + t13m;

    // no depth term for interface events
    double fDepth = slab ? c.θ11 * (min(zTor, 120.0) - 60.0) : 0.0;

    double vsS = min(vs30, VSS_MAX);

    double fFaba = 0.0;
    if (backarc) {
      if (slab) {
        // slab modeled as point sources, so use of rRup instead of rHypo is
        // appropriate
        fFaba = c.θ7 + c.θ8 * log(max(rRup, 85.0) / 40.0);
      } else {
        fFaba = c.θ15 + c.θ16 * log(max(rRup, 100.0) / 40.0);
      }
    }

    double fSite = c.θ12 * log(vsS / c.vlin);
    if (vs30 < c.vlin) { // whether or not we use pgaRock
      fSite += -c.b * log(pgaRock + C) + c.b * log(pgaRock + C * pow((vsS / c.vlin), N));
    } else {
      // for pgaRock loop, vs=1000 > vlinPGA=865
      fSite += c.b * N * log(vsS / c.vlin);
    }

    return c.θ1 + T4 * ΔC1 +
        (c.θ2 + (slab ? c.θ14 : 0.0) + T3 * (Mw - 7.8)) *
            log(rRup + C4 * exp((Mw - 6.0) * T9)) +
        c.θ6 * rRup + (slab ? c.θ10 : 0.0) + fMag +
        fDepth +
        fFaba +
        fSite;
  }

  /*
   * Developer note: In most GMMs, subtype constructors, if present, need only
   * the IMT argument to initialize their parent. To support several
   * interpolated spectral periods, the parent also needs to know the specific
   * subtype Gmm identifier in order to obtain concrete instances of the
   * bounding spectral periods. In the case of cascading subtypes, multiple
   * constructors are needed, c(imt) and c(imt, subtype), to support
   * Gmm.instance and additional sub-subtypes, respectively.
   */

  static class Interface extends BcHydro_2012 {
    static final String NAME = BcHydro_2012.NAME + " : Interface";

    Interface(Imt imt) {
      super(imt, Gmm.BCHYDRO_12_INTERFACE);
    }

    protected Interface(Imt imt, Gmm subtype) {
      super(imt, subtype);
    }

    @Override
    final boolean isSlab() {
      return false;
    }

    @Override
    boolean deepBasinEffect() {
      return false;
    }

    @Override
    boolean isBackarc() {
      return false;
    }
  }

  static final class BasinInterface extends Interface {
    static final String NAME = Interface.NAME + " : Basin";

    BasinInterface(Imt imt) {
      super(imt, Gmm.BCHYDRO_12_INTERFACE_BASIN);
    }

    @Override
    final boolean deepBasinEffect() {
      return true;
    }
  }

  static final class BackarcBasinInterface extends Interface {
    static final String NAME = Interface.NAME + " : Basin : Backarc";

    BackarcBasinInterface(Imt imt) {
      super(imt, Gmm.BCHYDRO_12_INTERFACE_BASIN_BACKARC);
    }

    @Override
    final boolean deepBasinEffect() {
      return true;
    }

    @Override
    boolean isBackarc() {
      return true;
    }
  }

  static final class BackarcInterface extends Interface {
    static final String NAME = Interface.NAME + " : Backarc";

    BackarcInterface(Imt imt) {
      super(imt, Gmm.BCHYDRO_12_INTERFACE_BACKARC);
    }

    @Override
    final boolean deepBasinEffect() {
      return false;
    }

    @Override
    boolean isBackarc() {
      return true;
    }
  }

  static class Slab extends BcHydro_2012 {
    static final String NAME = BcHydro_2012.NAME + " : Slab";

    Slab(Imt imt) {
      super(imt, Gmm.BCHYDRO_12_SLAB);
    }

    Slab(Imt imt, Gmm subtype) {
      super(imt, subtype);
    }

    @Override
    final boolean isSlab() {
      return true;
    }

    @Override
    boolean deepBasinEffect() {
      return false;
    }

    @Override
    boolean isBackarc() {
      return false;
    }
  }

  static final class BasinSlab extends Slab {
    static final String NAME = Slab.NAME + " : Basin";

    BasinSlab(Imt imt) {
      super(imt, Gmm.BCHYDRO_12_SLAB_BASIN);
    }

    @Override
    final boolean deepBasinEffect() {
      return true;
    }
  }

  static final class BackarcBasinSlab extends Slab {
    static final String NAME = Slab.NAME + " : Basin : Backarc";

    BackarcBasinSlab(Imt imt) {
      super(imt, Gmm.BCHYDRO_12_SLAB_BASIN_BACKARC);
    }

    @Override
    final boolean deepBasinEffect() {
      return true;
    }

    @Override
    boolean isBackarc() {
      return true;
    }
  }

  static final class BackarcSlab extends Slab {
    static final String NAME = Slab.NAME + " : Backarc";

    BackarcSlab(Imt imt) {
      super(imt, Gmm.BCHYDRO_12_SLAB_BACKARC);
    }

    @Override
    final boolean deepBasinEffect() {
      return false;
    }

    @Override
    boolean isBackarc() {
      return true;
    }
  }
}
