package gov.usgs.earthquake.nshmp.gmm;

import static com.google.common.io.Resources.getResource;
import static com.google.common.io.Resources.readLines;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkWeights;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RJB;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.TABLE_DIR;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.min;
import static java.lang.Math.sqrt;
import static java.util.stream.Collectors.toMap;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import com.google.common.annotations.Beta;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.Text.Delimiter;
import gov.usgs.earthquake.nshmp.data.Interpolator;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable.Position;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the PEER NGA-East for USGS ground motion model. This
 * version of the model was developed specifically for USGS applications but is
 * identical to the final NGA-East model developed for the NRC. It is a
 * composite model that consists of 17 median ground motion models with period
 * dependent weights. This class also includes the implementation of the USGS
 * Updated Seed logic tree.
 *
 * <p>Calculation of hazard using this implementation deviates somewhat from the
 * current nshmp-haz PSHA pipeline and required implementation of a
 * {@code MultiScalarGroundMotion}. A {@code MultiScalarGroundMotion} stores
 * arrays of means and sigmas with associated weights.
 *
 * <p>This class manages implementations of 22 'seed' models, 19 of which were
 * used to generate (via Sammons mapping) the 17 NGA-East for USGS models and
 * associated weights, and 3 of which are updates. This class also handles USGS
 * logic tree of 14 of those seed models. Ground motions for most models are
 * computed via table lookups (SP16 is the exception).
 *
 * <p>On it's own, NGA-East is a hard rock model returning results for a site
 * class where Vs30 = 3000 m/s. To accomodate other site classes, the Stewart et
 * al. (2017) CEUS site amplification model is used. This model is applicable to
 * 200 ≤ vs30 ≤ 2000 m/s. In the current implementation, for vs30 &lt; 200, vs30
 * = 200 m/s; for vs30 &gt; 2000, vs30 = 3000 m/s.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Implementation note:</b> Mean values are NOT clamped per
 * {@link GmmUtils#ceusMeanClip(Imt, double)}.
 *
 * <p><b>Reference:</b> Goulet, C., Bozorgnia, Y., Kuehn, N., Al Atik, L.,
 * Youngs, R., Graves, R., Atkinson, G., 2017, NGA-East ground-motion models for
 * the U.S. Geological Survey national seismic hazard maps: PEER Report No.
 * 2017/03, 180 p.
 *
 * <p><b>Reference:</b> Goulet CA, Bozorgnia Y, Abrahamson N, et al. (2018)
 * Central and eastern North America ground— motion characterization—NGA-East
 * final report. Report no. 2018/08. Berkeley, CA: PEER, 817 pp.
 *
 * <p><b>Reference:</b> Stewart, J., Parker, G., Atkinson, G., Boore, D., and
 * Hashash, Y., and Silva, W., 2020, Ergodic site amplification model for
 * central and eastern North America: Earthquake Spectra 36(1): 42–68.
 *
 * <p><b>Reference:</b> Hashash, Y., Harmon, J., Ilhan, O., Parker, G., and
 * Stewart, 2020, Nonlinear site amplification model for ergodic seismic hazard
 * analysis in central and eastern North America. Earthquake Spectra 36(1):
 * 69–86.
 *
 * <p><b>Reference:</b>Stewart, J., Parker, G., Al Atik, L, et al. (2019)
 * Site-to-site standard deviation model for central and eastern North America.
 * UC E-Scholarship publication. Available at: <a
 * href="https://escholarship.org/uc/item/2sc5g220">
 * https://escholarship.org/uc/item/2sc5g220</a> (accessed 18 November 2019).
 *
 * <p><b>Component:</b> average horizontal (RotD50)
 *
 * @author U.S. Geological Survey
 * @see Gmm#NGA_EAST_USGS
 * @see Gmm#NGA_EAST_USGS_SEEDS
 * @see Gmm#NGA_EAST_SEED_1CCSP
 * @see Gmm#NGA_EAST_SEED_1CVSP
 * @see Gmm#NGA_EAST_SEED_2CCSP
 * @see Gmm#NGA_EAST_SEED_2CVSP
 * @see Gmm#NGA_EAST_SEED_B_A04
 * @see Gmm#NGA_EAST_SEED_B_AB14
 * @see Gmm#NGA_EAST_SEED_B_AB95
 * @see Gmm#NGA_EAST_SEED_B_BCA10D
 * @see Gmm#NGA_EAST_SEED_B_BS11
 * @see Gmm#NGA_EAST_SEED_B_SGD02
 * @see Gmm#NGA_EAST_SEED_FRANKEL
 * @see Gmm#NGA_EAST_SEED_GRAIZER
 * @see Gmm#NGA_EAST_SEED_GRAIZER16
 * @see Gmm#NGA_EAST_SEED_GRAIZER17
 * @see Gmm#NGA_EAST_SEED_HA15
 * @see Gmm#NGA_EAST_SEED_PEER_EX
 * @see Gmm#NGA_EAST_SEED_PEER_GP
 * @see Gmm#NGA_EAST_SEED_PZCT15_M1SS
 * @see Gmm#NGA_EAST_SEED_PZCT15_M2ES
 * @see Gmm#NGA_EAST_SEED_SP15
 * @see Gmm#NGA_EAST_SEED_SP16
 * @see Gmm#NGA_EAST_SEED_YA15
 */
@Beta
public abstract class NgaEastUsgs_2017 implements GroundMotionModel {

  /*
   * Developer notes:
   *
   * Note: supported periods are derived from sigma coefficient files. Several
   * supported periods have been commented out because they are not represented
   * in the site amplification model.
   *
   * When supplied with tables for the 17 usgs models, 0.01s is present with
   * values distinct from PGA, however the seed models are missing this period.
   * We therefore duplicate PGA for 0.01s ground motion values in the seed model
   * tables. Because a coefficient table (sigma coeffs in this case) is
   * referenced for supported IMTs, the ground motion tables across seed and
   * sammons models must be consistent.
   *
   * Missing PGV tables: Grazier, PEER_EX, PEER_GP, PZCT15_M1SS, PZCT15_M2ES
   *
   * Notes on sigma:
   *
   * NGA-East for USGS (2017) recommended use of an updated EPRI (2013) model
   * (Table 5.5), which is currently considered as one independent branch of an
   * aleatory variability logic tree.
   *
   * The NSHMP asked PEER panel to develop an improved φ_s2s model. This
   * necessitated including τ and φ_ss as independent terms. Only the 'global' τ
   * and φ_ss models are considered, per Linda Al Atik's recommendation. Because
   * no statistical uncertianty model was developed for φ_s2s, we only consider
   * the 'central' branch for τ and φ_ss terms. Sensitivity studies show that
   * consideration of only the 'global' and 'central' branches of τ and φ_ss has
   * little effect on the final σ. This 'panel' model is the second branch in
   * the aleatory variability logic tree.
   */

  static final String NAME = "NGA-East for USGS (2017)";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(4.0, 8.2))
      .set(RJB, Range.closed(0.0, 1500.0))
      .set(VS30, Range.closed(200.0, 3000.0))
      .build();

  /*
   * Sigma coefficients for global model from tables 5-1 (tau) and 5-2 (phi) and
   * EPRI model from table 5-5.
   */
  static CoefficientContainer COEFFS_SIGMA_PANEL;
  static CoefficientContainer COEFFS_SIGMA_EPRI;

  /* Immutable, ordered map of seed model weights. */
  static Map<String, Double> USGS_SEED_WEIGHTS;

  static {
    COEFFS_SIGMA_PANEL = new CoefficientContainer("nga-east-usgs-sigma-panel.csv");
    COEFFS_SIGMA_EPRI = new CoefficientContainer("nga-east-usgs-sigma-epri.csv");
    USGS_SEED_WEIGHTS = initSeedWeights();
  }

  private static Map<String, Double> initSeedWeights() {
    try {
      Map<String, Double> wtMap = readLines(
          getResource(TABLE_DIR + "nga-east-seed-weights.dat"),
          StandardCharsets.UTF_8)
              .stream()
              .skip(1)
              .map(line -> Text.splitToList(line, Delimiter.COMMA))
              .collect(toMap(
                  entry -> entry.get(0),
                  entry -> Double.valueOf(entry.get(1))));

      checkWeights(wtMap.values());
      return Map.copyOf(wtMap);

    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  private static final double[] SITE_AMP_SIGMA_WTS = { 0.185, 0.63, 0.185 };

  /* Panel and EPRI sigma models and wieghts */
  private static final String[] SIGMA_IDS = { "σ-epri", "σ-usgs" };
  private static final double[] SIGMA_WTS = { 0.8, 0.2 };

  /* φ_s2s constants */
  private static final double VΦ1 = 1200.0;
  private static final double VΦ2 = 1500.0;

  private static final class CoefficientsSigmaPanel {

    /* τ coefficients, global model, central branch */
    final double τ1, τ2, τ3, τ4;

    /* φ_ss coefficients, global model, central branch */
    final double a, b;

    /* φ_s2s coefficients */
    final double φs2s1, φs2s2;

    CoefficientsSigmaPanel(Imt imt, CoefficientContainer cc) {
      Map<String, Double> coeffs = cc.get(imt);
      τ1 = coeffs.get("t1");
      τ2 = coeffs.get("t2");
      τ3 = coeffs.get("t3");
      τ4 = coeffs.get("t4");
      a = coeffs.get("ss_a");
      b = coeffs.get("ss_b");
      φs2s1 = coeffs.get("s2s1");
      φs2s2 = coeffs.get("s2s2");
    }
  }

  /* Updated EPRI (2013); initial NGAE-East team recommendation. */
  private static final class CoefficientsSigmaEpri {

    final double τ_m5, φ_m5, τ_m6, φ_m6, τ_m7, φ_m7;

    CoefficientsSigmaEpri(Imt imt, CoefficientContainer cc) {
      Map<String, Double> coeffs = cc.get(imt);
      τ_m5 = coeffs.get("tau_M5");
      φ_m5 = coeffs.get("phi_M5");
      τ_m6 = coeffs.get("tau_M6");
      φ_m6 = coeffs.get("phi_M6");
      τ_m7 = coeffs.get("tau_M7");
      φ_m7 = coeffs.get("phi_M7");
    }
  }

  private final CoefficientsSigmaPanel σCoeffsPanel;
  private final CoefficientsSigmaEpri σCoeffsEpri;
  final Imt imt;

  NgaEastUsgs_2017(Imt imt) {
    σCoeffsPanel = new CoefficientsSigmaPanel(imt, COEFFS_SIGMA_PANEL);
    σCoeffsEpri = new CoefficientsSigmaEpri(imt, COEFFS_SIGMA_EPRI);
    this.imt = imt;
  }

  double sigmaEpri(double Mw) {
    return sigmaEpri(σCoeffsEpri, Mw);
  }

  double sigmaPanel(double Mw, double vs30) {
    return sigmaPanel(σCoeffsPanel, Mw, vs30);
  }

  /* σ branch 1: Updated EPRI (2013); initial recomendation. */
  private static double sigmaEpri(CoefficientsSigmaEpri c, double Mw) {
    double τ;
    double φ;
    if (Mw <= 5.0) {
      τ = c.τ_m5;
      φ = c.φ_m5;
    } else if (Mw <= 6.0) {
      τ = Interpolator.findY(5.0, c.τ_m5, 6.0, c.τ_m6, Mw);
      φ = Interpolator.findY(5.0, c.φ_m5, 6.0, c.φ_m6, Mw);
    } else if (Mw <= 7.0) {
      τ = Interpolator.findY(6.0, c.τ_m6, 7.0, c.τ_m7, Mw);
      φ = Interpolator.findY(6.0, c.φ_m6, 7.0, c.φ_m7, Mw);
    } else {
      τ = c.τ_m7;
      φ = c.φ_m7;
    }
    return Maths.hypot(φ, τ);
  }

  /* σ branch 2: Panel recommendation with φ_s2s. */
  private static double sigmaPanel(
      CoefficientsSigmaPanel c,
      double Mw,
      double vs30) {

    /* τ model; global branch only; Equation 5-1 */
    double τ = tau(Mw, c.τ1, c.τ2, c.τ3, c.τ4);

    /* φ_ss model; global branch only; Equation 5-2 */
    double φ_ss = phi_ss(Mw, c.a, c.b);

    /* φ_s2s model; single branch; Stewart et al. (2019) */
    double φ_s2s = phi_s2s(vs30, c.φs2s1, c.φs2s2);

    return Maths.hypot(τ, φ_ss, φ_s2s);
  }

  /* τ: Equation 5.1 */
  private static double tau(
      double Mw,
      double τ1,
      double τ2,
      double τ3,
      double τ4) {

    if (Mw <= 4.5) {
      return τ1;
    } else if (Mw <= 5.0) {
      return τ1 + (τ2 - τ1) * (Mw - 4.5) / 0.5;
    } else if (Mw <= 5.5) {
      return τ2 + (τ3 - τ2) * (Mw - 5.0) / 0.5;
    } else if (Mw <= 6.5) {
      return τ3 + (τ4 - τ3) * (Mw - 5.5);
    }
    return τ4;
  }

  /* φ_ss: Equation 5.2 */
  private static double phi_ss(
      double Mw,
      double a,
      double b) {

    if (Mw <= 5.0) {
      return a;
    } else if (Mw <= 6.5) {
      return a + (Mw - 5.0) * (b - a) / 1.5;
    }
    return b;
  }

  /* φ_s2s: Stewart et al. */
  private static double phi_s2s(
      double vs30,
      double φs2s1,
      double φs2s2) {

    if (vs30 < VΦ1) {
      return φs2s1;
    } else if (vs30 < VΦ2) {
      return φs2s1 - ((φs2s1 - φs2s2) / (VΦ2 - VΦ1)) * (vs30 - VΦ1);
    }
    return φs2s2;
  }

  /*
   * Base model used for sammons and seed model groups that share common site
   * class and sigma models.
   */
  static class Usgs17 extends NgaEastUsgs_2017 {

    static final int MODEL_COUNT = 17;
    static final String[] MEAN_IDS = IntStream.range(1, MODEL_COUNT + 1)
        .mapToObj(i -> "Sammons" + i)
        .collect(Collectors.toList())
        .toArray(new String[MODEL_COUNT]);

    final GroundMotionTable[] tables;
    final GroundMotionTable[] pgaTables;
    final SiteAmp siteAmp;
    final double[] μWts;

    /* Specifiy an array of models ids. */
    Usgs17(Imt imt) {
      super(imt);
      this.tables = GroundMotionTables.getNgaEast(imt);
      this.pgaTables = GroundMotionTables.getNgaEast(Imt.PGA);
      this.siteAmp = new SiteAmp(imt);
      this.μWts = GroundMotionTables.getNgaEastWeights(imt);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      Position p = tables[0].position(in.rRup, in.Mw);
      double[] μs = new double[MODEL_COUNT];
      for (int i = 0; i < MODEL_COUNT; i++) {
        double μ = tables[i].get(p);
        double μPga = exp(pgaTables[i].get(p));
        SiteAmp.Value fSite = siteAmp.calc(μPga, in.vs30);
        μs[i] = fSite.apply(μ);
      }
      double[] σs = new double[] {
          sigmaEpri(in.Mw),
          sigmaPanel(in.Mw, in.vs30) };
      return GroundMotions.createTree(
          MEAN_IDS, μs, μWts,
          SIGMA_IDS, σs, SIGMA_WTS);
    }
  }

  /* Guo and Chapman Gulf Coastal Plain Amplification (CPA) model. */
  static class Usgs17Cpa extends Usgs17 {

    static final String NAME = NgaEastUsgs_2017.NAME + " (Gulf Coast)";

    Usgs17Cpa(Imt imt) {
      super(imt);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      double cpa = Double.isNaN(in.z2p5)
          ? 0.0
          : log(ChapmanGuo_2021.cpaPsaRatio(imt, in.z2p5, in.Mw, in.rJB));
      Position p = tables[0].position(in.rRup, in.Mw);
      double[] μs = new double[MODEL_COUNT];
      for (int i = 0; i < MODEL_COUNT; i++) {
        double μ = tables[i].get(p);
        μs[i] = μ + cpa;
      }
      double[] σs = new double[] {
          sigmaEpri(in.Mw),
          sigmaPanel(in.Mw, 3000.0) };
      return GroundMotions.createTree(
          MEAN_IDS, μs, μWts,
          SIGMA_IDS, σs, SIGMA_WTS);
    }
  }

  /*
   * Implementation of USGS Seed model logic tree. All models but SP16 are table
   * based; SP16 is added to the median ground motion array last. NOTE that the
   * model ignores the SP16 aleatory variability model for consistency with the
   * other seed models.
   *
   * PGV: rather than use averaged tables for seed that don't provide PGV tables
   * (inital implementation), we delegate to UsgsPgvSupport. This requires quick
   * short circuiting of those Gmms lacking support. In doing so we use the
   * individual seed models which results in repeated position lookups. Both
   * sigma models considered include coefficients for PGV.
   */
  static class UsgsSeeds extends NgaEastUsgs_2017 {
    static final String NAME = "NGA-East Updated Seed Tree";
    static final String SEED_PREFIX = "NGA_EAST_SEED_";
    static final String SP16_ID = "SP16";
    static final Set<Gmm> noPgvSeeds = EnumSet.of(
        Gmm.NGA_EAST_SEED_PEER_GP,
        Gmm.NGA_EAST_SEED_GRAIZER16,
        Gmm.NGA_EAST_SEED_GRAIZER17,
        Gmm.NGA_EAST_SEED_PZCT15_M1SS,
        Gmm.NGA_EAST_SEED_PZCT15_M2ES);

    /* Valid Seed IDs for table based models; skips SP16 */
    static final List<String> SEED_TABLE_IDS; // size = 13

    /* All Gmms */
    static final List<Gmm> GMMS; // pgv helper, size = 14

    /* All ground motion mean model IDs and weights */
    static final String[] MEAN_IDS;
    static final double[] MEAN_WTS;

    final Map<Gmm, GroundMotionTable> tables;
    final Map<Gmm, GroundMotionTable> pgaTables;
    final ShahjoueiPezeshk_2016 sp16;
    final SiteAmp siteAmp;

    static {
      List<String> tableIds = new ArrayList<>(); // valid IDs for tables
      List<Gmm> gmms = new ArrayList<>();
      List<String> μIds = new ArrayList<>();
      List<Double> μWts = new ArrayList<>();
      USGS_SEED_WEIGHTS.entrySet().forEach(e -> {
        String seedId = e.getKey();
        if (!seedId.equals(SP16_ID)) {
          tableIds.add(seedId);
        }
        gmms.add(Gmm.valueOf(SEED_PREFIX + e.getKey().toUpperCase()));
        μIds.add("Seed (" + seedId + ")");
        μWts.add(e.getValue());
      });
      SEED_TABLE_IDS = List.copyOf(tableIds);
      GMMS = List.copyOf(gmms);
      MEAN_IDS = μIds.toArray(new String[μIds.size()]);
      MEAN_WTS = μWts.stream()
          .mapToDouble(Double::valueOf)
          .toArray();
    }

    UsgsSeeds(Imt imt) {
      super(imt);
      this.tables = getImtTables(imt);
      this.pgaTables = getImtTables(Imt.PGA);
      this.sp16 = (ShahjoueiPezeshk_2016) Gmm.NGA_EAST_SEED_SP16.instance(imt);
      this.siteAmp = new SiteAmp(imt);
    }

    private static Map<Gmm, GroundMotionTable> getImtTables(Imt imt) {
      return GroundMotionTables.getNgaEastSeeds(SEED_TABLE_IDS, imt)
          .entrySet().stream()
          .collect(toMap(
              e -> Gmm.valueOf(SEED_PREFIX + e.getKey().toUpperCase()),
              Entry::getValue));
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      Position p = tables.values().iterator().next().position(in.rRup, in.Mw);
      double[] μs = new double[GMMS.size()];
      for (int i = 0; i < GMMS.size(); i++) {
        Gmm seed = GMMS.get(i);
        if (seed == Gmm.NGA_EAST_SEED_SP16) {
          μs[i] = GroundMotions.combine(sp16.calc(in)).mean();
        } else if (imt == Imt.PGV && noPgvSeeds.contains(seed)) {
          // site is included when calling individual seed
          μs[i] = UsgsPgvSupport.calcAB20Pgv(seed, in).mean();
        } else {
          double μRock = tables.get(seed).get(p);
          double μPga = exp(pgaTables.get(seed).get(p));
          SiteAmp.Value fSite = siteAmp.calc(μPga, in.vs30);
          μs[i] = fSite.apply(μRock);
        }
      }
      double[] σs = new double[] {
          sigmaEpri(in.Mw),
          sigmaPanel(in.Mw, in.vs30) };
      return GroundMotions.createTree(
          MEAN_IDS, μs, MEAN_WTS,
          SIGMA_IDS, σs, SIGMA_WTS);
    }
  }

  /* Guo and Chapman Gulf Coastal Plain Amplification (CPA) model. */
  static class UsgsSeedsCpa extends UsgsSeeds {

    static final String NAME = UsgsSeeds.NAME + " (Gulf Coast)";

    UsgsSeedsCpa(Imt imt) {
      super(imt);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      GmmInput inRock = GmmInput.builder().fromCopy(in).vs30(3000).build();
      double cpa = Double.isNaN(in.z2p5)
          ? 0.0
          : log(ChapmanGuo_2021.cpaPsaRatio(imt, in.z2p5, in.Mw, in.rJB));
      Position p = tables.values().iterator().next().position(in.rRup, in.Mw);
      double[] μs = new double[GMMS.size()];
      for (int i = 0; i < GMMS.size(); i++) {
        Gmm seed = GMMS.get(i);
        if (seed == Gmm.NGA_EAST_SEED_SP16) {
          μs[i] = GroundMotions.combine(sp16.calc(inRock)).mean() + cpa;
        } else if (imt == Imt.PGV && noPgvSeeds.contains(seed)) {
          // use custom input to ensure hard rock from seeds
          μs[i] = UsgsPgvSupport.calcAB20Pgv(seed, inRock).mean() + cpa;
        } else {
          μs[i] = tables.get(seed).get(p) + cpa;
        }
      }
      double[] σs = new double[] {
          sigmaEpri(in.Mw),
          sigmaPanel(in.Mw, 3000) };
      return GroundMotions.createTree(
          MEAN_IDS, μs, MEAN_WTS,
          SIGMA_IDS, σs, SIGMA_WTS);
    }
  }

  static abstract class Seed extends NgaEastUsgs_2017 {
    static final String NAME = NgaEastUsgs_2017.NAME + " : Seed : ";

    final String id;
    final GroundMotionTable table;
    final GroundMotionTable pgaTable;
    final SiteAmp siteAmp;

    Seed(String id, Imt imt) {
      super(imt);
      this.id = id;
      this.table = GroundMotionTables.getNgaEastSeed(id, imt);
      this.pgaTable = GroundMotionTables.getNgaEastSeed(id, Imt.PGA);
      this.siteAmp = new SiteAmp(imt);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      double σEpri = sigmaEpri(in.Mw);
      double σPanel = sigmaPanel(in.Mw, in.vs30);
      Position p = table.position(in.rRup, in.Mw);
      double μPga = exp(pgaTable.get(p));
      SiteAmp.Value fSite = siteAmp.calc(μPga, in.vs30);
      double μ = fSite.apply(table.get(p));
      double σ = Maths.srssWeighted(
          new double[] { σEpri, σPanel },
          SIGMA_WTS);
      return GroundMotions.createTree(μ, σ);
    }
  }

  static final class Seed_1CCSP extends Seed {
    static final String ID = "1CCSP";
    static final String NAME = Seed.NAME + ID;

    Seed_1CCSP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_1CVSP extends Seed {
    static final String ID = "1CVSP";
    static final String NAME = Seed.NAME + ID;

    Seed_1CVSP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_2CCSP extends Seed {
    static final String ID = "2CCSP";
    static final String NAME = Seed.NAME + ID;

    Seed_2CCSP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_2CVSP extends Seed {
    static final String ID = "2CVSP";
    static final String NAME = Seed.NAME + ID;

    Seed_2CVSP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_ANC15 extends Seed {
    static final String ID = "ANC15";
    static final String NAME = Seed.NAME + ID;

    Seed_ANC15(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_a04 extends Seed {
    static final String ID = "B_a04";
    static final String NAME = Seed.NAME + ID;

    Seed_B_a04(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_ab14 extends Seed {
    static final String ID = "B_ab14";
    static final String NAME = Seed.NAME + ID;

    Seed_B_ab14(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_ab95 extends Seed {
    static final String ID = "B_ab95";
    static final String NAME = Seed.NAME + ID;

    Seed_B_ab95(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_bca10d extends Seed {
    static final String ID = "B_bca10d";
    static final String NAME = Seed.NAME + ID;

    Seed_B_bca10d(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_bs11 extends Seed {
    static final String ID = "B_bs11";
    static final String NAME = Seed.NAME + ID;

    Seed_B_bs11(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_sgd02 extends Seed {
    static final String ID = "B_sgd02";
    static final String NAME = Seed.NAME + ID;

    Seed_B_sgd02(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_Frankel extends Seed {
    static final String ID = "Frankel";
    static final String NAME = Seed.NAME + ID;

    Seed_Frankel(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_Graizer extends Seed {
    static final String ID = "Graizer";
    static final String NAME = Seed.NAME + ID;

    Seed_Graizer(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_HA15 extends Seed {
    static final String ID = "HA15";
    static final String NAME = Seed.NAME + ID;

    Seed_HA15(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_PEER_EX extends Seed {
    static final String ID = "PEER_EX";
    static final String NAME = Seed.NAME + ID;

    Seed_PEER_EX(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_PEER_GP extends Seed {
    static final String ID = "PEER_GP";
    static final String NAME = Seed.NAME + ID;

    Seed_PEER_GP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_PZCT15_M1SS extends Seed {
    static final String ID = "PZCT15_M1SS";
    static final String NAME = Seed.NAME + ID;

    Seed_PZCT15_M1SS(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_PZCT15_M2ES extends Seed {
    static final String ID = "PZCT15_M2ES";
    static final String NAME = Seed.NAME + ID;

    Seed_PZCT15_M2ES(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_SP15 extends Seed {
    static final String ID = "SP15";
    static final String NAME = Seed.NAME + ID;

    Seed_SP15(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_YA15 extends Seed {
    static final String ID = "YA15";
    static final String NAME = Seed.NAME + ID;

    Seed_YA15(Imt imt) {
      super(ID, imt);
    }
  }

  /*
   * Updated Graizer models.
   */

  static final class SeedUpdate_Graizer16 extends Seed {
    static final String ID = "Graizer16";
    static final String NAME = Seed.NAME + ID + " (updated)";

    SeedUpdate_Graizer16(Imt imt) {
      super(ID, imt);
    }
  }

  static final class SeedUpdate_Graizer17 extends Seed {
    static final String ID = "Graizer17";
    static final String NAME = Seed.NAME + ID + " (updated)";

    SeedUpdate_Graizer17(Imt imt) {
      super(ID, imt);
    }
  }

  /**
   * Stewart et al. site amplification model.
   *
   * The model is applicable to 200 ≤ vs30 ≤ 2000 m/s. In the current
   * implementation, for vs30 < 200, vs30 = 200 m/s; for vs30 > 2000, vs30 =
   * 3000 m/s.
   */
  static final class SiteAmp {

    static final String NAME = NgaEastUsgs_2017.NAME + " : Site Amplification";

    private static final CoefficientContainer COEFFS = new CoefficientContainer(
        "nga-east-usgs-siteamp.csv");

    private static final double V_MIN = 150.0;
    private static final double V_MAX = 3000.0;
    private static final double V_LIN_REF = 760.0;
    private static final double VL = 200.0;
    private static final double VU = 2000.0;

    private static final double VW1 = 600.0;
    private static final double VW2 = 400.0;
    private static final double WT1 = 0.767; // impedance model @ VW1
    private static final double WT2 = 0.1; // impedance model @ VW2
    private static final double WT_SCALE = (WT1 - WT2) / (log(VW1) - log(VW2)); // ≈1.65

    private final Coefficients c;

    private static final class Coefficients {

      final Imt imt;
      final double c, v1, v2, vf, σvc, σl, σu;
      final double f760i, f760g, f760iσ, f760gσ;
      final double f3, f4, f5, vc, σc;

      Coefficients(Imt imt, CoefficientContainer cc) {
        this.imt = imt;
        Map<String, Double> coeffs = cc.get(imt);
        c = coeffs.get("c");
        v1 = coeffs.get("V1");
        v2 = coeffs.get("V2");
        vf = coeffs.get("Vf");
        σvc = coeffs.get("sig_vc");
        σl = coeffs.get("sig_l");
        σu = coeffs.get("sig_u");
        f760i = coeffs.get("f760i");
        f760iσ = coeffs.get("f760is");
        f760g = coeffs.get("f760g");
        f760gσ = coeffs.get("f760gs");
        f3 = coeffs.get("f3");
        f4 = coeffs.get("f4");
        f5 = coeffs.get("f5");
        vc = coeffs.get("Vc");
        σc = coeffs.get("sig_c");
      }
    }

    SiteAmp(Imt imt) {
      c = new Coefficients(imt, COEFFS);
    }

    SiteAmp.Value calc(double pgaRock, double vs30) {

      /*
       * Developer notes:
       *
       * Vs30 values outside the range 200 < Vs30 < 3000 m/s are clamped to the
       * supported range.
       *
       * ---------------------------------------------------------------------
       *
       * Comments from R implementation:
       *
       * This function is used to compute site amplification in ln units for
       * inputs Vs30 (in m/s) and PGAr (peak acceleration for 3000 m/s reference
       * rock in g). Outputs:
       *
       * 1) Linear and nonlinear site amplification (FLin, FNonlin)
       *
       * 2) Standard deviation (σLin, σNonlin)
       *
       * 3) Total site amplification (fT) and total standard deviation (σT)
       *
       * Notes:
       *
       * 1) We assume fv and f760 are independent, so variance of linear site
       * amplication is the sum of variance f760 and f760;
       *
       * 2) The third value of σc coefficient, which is used to compute
       * nonlinear site amplication. 0.12 in report, and 0.1 in excel. Here we
       * use 0.12.
       *
       * 3) Standard deviation of fv calculation is updated by adding a new
       * coefficient vf instead of using v1.
       */

      /* Vs30 filtering */

      if (vs30 >= V_MAX) {
        return new Value(0.0, 0.0);
      } else if (vs30 < V_MIN) {
        vs30 = V_MIN;
      }

      /* Linear response */

      /*
       * Vs30 dependent f760 model: impedance vs. gradient
       *
       * The final model gives 2/3 weight to a weight scaling model of
       * [0.9i,0.1g] to [0.1i,0.9g] and 1/3 weight to a weight scaling model of
       * [0.5i,0.5g] to [0.1i,0.9g]. This results in a final weight scaling
       * model of [0.767i,0.233g] at 600 m/s and [0.1i,0.9g] at 400 m/s.
       */
      double wti = WT1; // impedance = 0.767 @ Vs30 > 600
      if (vs30 < VW2) {
        wti = WT2; // impedance = 0.1 @ Vs30 < 400
      } else if (vs30 < VW1) {
        wti = WT_SCALE * log(vs30 / VW2) + WT2;
      }
      double wtg = 1.0 - wti;

      double f760 = c.f760i * wti + c.f760g * wtg;
      double f760σ = c.f760iσ * wti + c.f760gσ * wtg;

      double fv = 0.0;
      if (vs30 <= c.v1) {
        fv = c.c * log(c.v1 / V_LIN_REF);
      } else if (vs30 <= c.v2) {
        fv = c.c * log(vs30 / V_LIN_REF);
      } else if (vs30 <= VU) {
        fv = c.c * log(c.v2 / V_LIN_REF);
      } else {
        /* Equivalent to equation 3 for 2000 < vs30 < 3000 */
        double f2000 = c.c * log(c.v2 / V_LIN_REF);
        fv = Interpolator.findY(log(VU), f2000, log(V_MAX), -f760, log(vs30));
      }

      double fvσ = 0.0;
      if (vs30 < c.vf) {
        double σT = c.σl - c.σvc;
        double vT = (vs30 - VL) / (c.vf - VL);
        fvσ = c.σl - 2.0 * σT * vT + σT * vT * vT;
      } else if (vs30 <= c.v2) {
        fvσ = c.σvc;
      } else if (vs30 <= VU) {
        double vT = (vs30 - c.v2) / (VU - c.v2);
        fvσ = c.σvc + (c.σu - c.σvc) * vT * vT;
      } else {
        fvσ = c.σu * (1.0 - log(vs30 / VU) / log(V_MAX / VU));
      }

      double fLin = fv + f760;
      double σLin = sqrt(fvσ * fvσ + f760σ * f760σ);

      /* Nonlinear response */

      double vRefNl = (c.imt.ordinal() >= Imt.SA0P4.ordinal()) ? V_MAX : V_LIN_REF;
      double rkRefTerm = log((pgaRock + c.f3) / c.f3);

      double fNonlin = 0.0;
      if (vs30 < c.vc) {
        double f2 = c.f4 * (exp(c.f5 * (min(vs30, vRefNl) - 360.0)) -
            exp(c.f5 * (vRefNl - 360.0)));
        fNonlin = f2 * rkRefTerm;
      }

      double σf2 = 0.0;
      if (vs30 < 300.0) {
        σf2 = c.σc;
      } else if (vs30 < 1000.0) {
        σf2 = c.σc - c.σc / log(1000.0 / 300.0) * log(vs30 / 300.0);
      }
      double σNonlin = σf2 * rkRefTerm;

      double fT = fLin + fNonlin;
      double σT = sqrt(σLin * σLin + σNonlin * σNonlin);

      return new Value(fT, σT);
    }

    /**
     * Wrapper class for site amplification and associated epistemic
     * uncertainty.
     */
    static final class Value {

      final double siteAmp;
      final double σ;

      Value(double siteAmp, double σ) {
        this.siteAmp = siteAmp;
        this.σ = σ;
      }

      double apply(double μ) {
        if (siteAmp == 0.0) {
          return μ;
        }
        double μAmp = μ + siteAmp;
        return log(SITE_AMP_SIGMA_WTS[0] * exp(μAmp + σ) +
            SITE_AMP_SIGMA_WTS[1] * exp(μAmp) +
            SITE_AMP_SIGMA_WTS[2] * exp(μAmp - σ));
      }
    }
  }

}
