package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P05;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P075;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P1;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA10P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA3P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA4P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA5P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA7P5;
import static java.lang.Math.exp;
import static java.lang.Math.log;

import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import com.google.common.annotations.Beta;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.google.common.collect.Range;
import com.google.common.primitives.Ints;

import gov.usgs.earthquake.nshmp.data.Interpolator;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the ground motion model by Wong et al. (2015) for deep
 * events (&gt;20 km) on the Big Island, Hawaii.
 *
 * <p><b>Implementation notes:</b> This model, as originally published, only
 * produces results for a site class with Vs30 = 428 m/s. Subsequently, Wong et
 * al. (2021) published a complementary model for shallow earthquakes that
 * included Vs30-specific sets of coefficients for both the shallow and deep
 * events, thereby expanding support to the full range of multi-period response
 * spectra (MPRS) site classes. For intermediate Vs30 values, this
 * implementation linearly interpolates ground motion between adjacent site
 * classes; ground motions are also clamped to the range 150 to 1500 m/s.
 *
 * To support the full MPRS response spectrum, coefficients, which are provided
 * in Hz, are mapped as closely as possible to their corrresponding periods.
 * This results in coefficients for 1.58 s being used for 1.5 s. Coefficients
 * for 0.07 and 0.08 seconds are provided, but this implementation interpolates
 * between 0.1 s and 0.05 s rather than interpolate coefficients for 0.075 s.
 * Spectral periods 4 s and 7.5 s are missing coefficients and are also
 * interpolated using adjacent periods.
 *
 * Lastly, per Wong et al. (2015), sigma at 2 s and above is fixed at the 2 s
 * value. This implementation further smoothes sigma to a fixed value of 0.8 for
 * all spectral periods.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Reference:</b> Wong, I.G., Silva, W.J., Darragh, R., Gregor, N., and
 * Dober, M., 2015, A ground motion prediction model for deep earthquakes
 * beneath the island of Hawaii: Earthquake Spectra, v. 31, p. 1763–1788.
 *
 * <p><b>Reference:</b>Wong, I.G., Darragh, R., Smith, S., Wu, Q., Silva, W.,
 * and Kishida, T., 2021, Ground Motion Models for Shallow Crustal and Deep
 * Earthquakes in Hawaii and Analyses of the 2018 M 6.9 Leilani Estates
 * Sequence: Earthquake Spectra, <i>in review</i>.
 *
 * <p><b>doi:</b><a href="http://doi.org/10.1193/012012EQS015M">
 * 10.1193/012012EQS015M</a>
 *
 * <p><b>Component:</b> average horizontal
 *
 * @author Peter Powers
 * @see Gmm#WONG_15
 */
@Beta
public final class WongEtAl_2015 implements GroundMotionModel {

  static final String NAME = "Wong et al. (2015)";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(5.0, 8.0))
      .set(RRUP, Range.closed(0.0, 300.0))
      .set(ZTOR, Range.closed(20.0, 60.0))
      .set(VS30, Range.closed(180.0, 1300.0))
      .build();

  private static final Map<Imt, Range<Imt>> INTERPOLATED_IMTS = Maps.immutableEnumMap(
      ImmutableMap.of(
          SA0P075, Range.closed(SA0P05, SA0P1),
          SA4P0, Range.closed(SA3P0, SA5P0),
          SA7P5, Range.closed(SA5P0, SA10P0)));

  /*
   * Developer note: c10 @ 1.995Hz (0.5s) missing negative sign per email from
   * Ivan on 07/23/2019.
   */

  static final CoefficientContainer COEFFS_150 = new CoefficientContainer("Wong21-150.csv");
  static final CoefficientContainer COEFFS_185 = new CoefficientContainer("Wong21-185.csv");
  static final CoefficientContainer COEFFS_260 = new CoefficientContainer("Wong21-260.csv");
  static final CoefficientContainer COEFFS_365 = new CoefficientContainer("Wong21-365.csv");
  static final CoefficientContainer COEFFS_428 = new CoefficientContainer("Wong21-428.csv");
  static final CoefficientContainer COEFFS_530 = new CoefficientContainer("Wong21-530.csv");
  static final CoefficientContainer COEFFS_760 = new CoefficientContainer("Wong21-760.csv");
  static final CoefficientContainer COEFFS_1080 = new CoefficientContainer("Wong21-1080.csv");
  static final CoefficientContainer COEFFS_1500 = new CoefficientContainer("Wong21-1500.csv");

  private static final double SIGMA = 0.8;

  private static class Coefficients {
    // Coefficient names updated to match Wong et al. 2021
    final double c1, c2, c4, c6, c7, c10;

    Coefficients(Imt imt, CoefficientContainer cc) {
      Map<String, Double> coeffs = cc.get(imt);
      c1 = coeffs.get("C1");
      c2 = coeffs.get("C2");
      c4 = coeffs.get("C4");
      c6 = coeffs.get("C6");
      c7 = coeffs.get("C7");
      c10 = coeffs.get("C10");
    }
  }

  private final Map<Integer, Coefficients> coeffsMap;
  private final List<Integer> vsKeys;

  private final boolean interpolated;
  private final GroundMotionModel interpolatedGmm;

  WongEtAl_2015(Imt imt) {
    coeffsMap = Map.of(
        150, new Coefficients(imt, COEFFS_150),
        185, new Coefficients(imt, COEFFS_185),
        260, new Coefficients(imt, COEFFS_260),
        365, new Coefficients(imt, COEFFS_365),
        428, new Coefficients(imt, COEFFS_428),
        530, new Coefficients(imt, COEFFS_530),
        760, new Coefficients(imt, COEFFS_760),
        1080, new Coefficients(imt, COEFFS_1080),
        1500, new Coefficients(imt, COEFFS_1500));
    vsKeys = coeffsMap.keySet().stream()
        .sorted()
        .collect(Collectors.toList());

    interpolated = INTERPOLATED_IMTS.containsKey(imt);
    interpolatedGmm = interpolated
        ? new InterpolatedGmm(Gmm.WONG_15, imt, INTERPOLATED_IMTS.get(imt))
        : null;
  }

  @Override
  public final LogicTree<GroundMotion> calc(GmmInput in) {
    if (interpolated) {
      return interpolatedGmm.calc(in);
    }

    /* clamp vs30 < 150 and vs30 > 1500 */
    int vsKey = Ints.constrainToRange((int) in.vs30, 150, 1500);

    if (coeffsMap.containsKey(vsKey)) {
      double μ = calcMean(coeffsMap.get(vsKey), in);
      return GroundMotions.createTree(μ, SIGMA);
    }

    int vsLoIndex = vsLoIndex(vsKey);
    int vsLo = vsKeys.get(vsLoIndex);
    int vsHi = vsKeys.get(vsLoIndex + 1);

    Coefficients coeffsLo = coeffsMap.get(vsLo);
    Coefficients coeffsHi = coeffsMap.get(vsHi);

    double μLo = calcMean(coeffsLo, in);
    double μHi = calcMean(coeffsHi, in);
    double μ = Interpolator.findX(μLo, vsLo, μHi, vsHi, in.vs30);

    return GroundMotions.createTree(μ, SIGMA);
  }

  private int vsLoIndex(int vsKey) {
    for (int i = 1; i < vsKeys.size(); i++) {
      if (vsKey < vsKeys.get(i)) {
        return i - 1;
      }
    }
    return vsKeys.size() - 1;
  }

  private static double calcMean(Coefficients c, GmmInput in) {
    double Mw = in.Mw;
    return c.c1 + c.c2 * Mw +
        (c.c6 + c.c7 * Mw) * log(in.rJB + exp(c.c4)) +
        c.c10 * (Mw - 6.0) * (Mw - 6.0);
  }

}
