package gov.usgs.earthquake.nshmp.internal;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkWeights;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.AbstractCollection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collection;
import java.util.EnumMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.DoubleStream;

import com.google.common.base.CharMatcher;
import com.google.common.base.Enums;
import com.google.common.base.Joiner;
import com.google.common.base.Splitter;
import com.google.common.collect.FluentIterable;
import com.google.common.collect.ImmutableSortedMap;
import com.google.common.collect.Iterables;
import com.google.common.collect.Iterators;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.primitives.Doubles;
import com.google.common.primitives.Ints;

import gov.usgs.earthquake.nshmp.data.Indexing;

/**
 * String, file, and XML parsing utilities.
 *
 * @author U.S. Geological Survey
 */
public final class Parsing {

  private static final Joiner.MapJoiner MAP_JOIN = Joiner.on(',').withKeyValueSeparator(":");
  private static final Splitter.MapSplitter MAP_SPLIT = Splitter.on(',').trimResults()
      .withKeyValueSeparator(":");
  private static final Splitter.MapSplitter MAP_MAP_SPLIT = Splitter.on(';').trimResults()
      .withKeyValueSeparator("::");

  /**
   * Convert a {@code Map<Enum, Double>} to a string with the form
   * {@code [ENUM_1 : 0.8, ENUM_2 : 0.2]}.
   *
   * @param map the {@code map} to convert
   * @return a string representation of the supplied {@code Map}
   */
  public static <T extends Enum<T>> String enumValueMapToString(Map<T, Double> map) {
    Map<String, String> strMap = Maps.newHashMap();
    for (Entry<T, Double> entry : map.entrySet()) {
      strMap.put(entry.getKey().name(), Double.toString(entry.getValue()));
    }
    return addBrackets(MAP_JOIN.join(strMap));
  }

  /**
   * Convert a string of the form {@code [ENUM_1 : 0.8, ENUM_2 : 0.2]} to a
   * {@code Map<Enum, Double>}. The returned map is immutable.
   *
   * @param s the string to parse
   * @param type {@code Class} to use as the key of the returned map
   * @throws IllegalArgumentException if supplied string is malformed or empty,
   *         or if weights do not sum to 1.0, within ±0.00001.
   * @throws NumberFormatException if supplied string does not contain parseable
   *         {@code double} values
   * @return a new immutable {@code Map<Enum, Double>} of identifiers and
   *         weights
   */
  public static <T extends Enum<T>> Map<T, Double> stringToEnumWeightMap(String s,
      Class<T> type) {
    Map<String, String> strMap = MAP_SPLIT.split(trimEnds(checkNotNull(s)));
    EnumMap<T, Double> wtMap = Maps.newEnumMap(type);
    Function<String, T> keyFunc = Enums.stringConverter(type);
    for (Entry<String, String> entry : strMap.entrySet()) {
      var key = checkNotNull(keyFunc.apply(entry.getKey().trim()));
      wtMap.put(key, Doubles.stringConverter().convert(entry.getValue()));
    }
    checkWeights(wtMap.values());
    return wtMap;
  }

  /**
   * Convert a string of the form
   * {@code [6.5 :: [1.0 : 0.8, 5.0 : 0.2]; 10.0 :: [1.0 : 0.2, 5.0 : 0.8]]} to
   * a {@code NavigableMap<Double, Map<Double, Double>}. The returned map and
   * nested maps are immutable and should only be accessed via
   * {@code Map.entrySet()} references or using keys derived from
   * {@code Map.keySet()} as {@code Double} comparisons are inherently
   * problematic.
   *
   * @param s the string to parse
   * @throws IllegalArgumentException if {@code s} is malformed or empty, or if
   *         weights do not sum to 1.0, within ±0.00001.
   * @throws NumberFormatException if {@code s} does not contain parseable
   *         {@code double} values
   * @return a new immutable {@code Map<Double, Double>} of values and their
   *         weights
   */
  public static NavigableMap<Double, Map<Double, Double>> stringToValueValueWeightMap(String s) {
    Map<String, String> strMap = MAP_MAP_SPLIT.split(trimEnds(checkNotNull(s)));
    ImmutableSortedMap.Builder<Double, Map<Double, Double>> builder = ImmutableSortedMap
        .naturalOrder();
    for (Entry<String, String> entry : strMap.entrySet()) {
      double key = checkNotNull(Doubles.stringConverter().convert(entry.getKey()));
      builder.put(key, stringToValueWeightMap(entry.getValue()));
    }
    return builder.build();
  }

  /**
   * Convert a string of the form {@code [1.0 : 0.4, 2.0 : 0.6]} to a
   * {@code NavigableMap<Double, Double>}. The returned map is immutable and
   * should only be accessed via {@code Map.entrySet()} references or using keys
   * derived from {@code Map.keySet()} as {@code Double} comparisons are
   * inherently problematic.
   *
   * @param s the string to parse
   * @throws IllegalArgumentException if {@code s} is malformed or if weights do
   *         not sum to 1.0, within ±0.00001.
   * @throws NumberFormatException if {@code s} does not contain parseable
   *         {@code double} values
   * @return a new immutable {@code Map<Double, Double>} of values and their
   *         weights
   */
  public static NavigableMap<Double, Double> stringToValueWeightMap(String s) {
    Map<String, String> strMap = MAP_SPLIT.split(trimEnds(checkNotNull(s)));
    ImmutableSortedMap.Builder<Double, Double> builder = ImmutableSortedMap.naturalOrder();
    for (Entry<String, String> entry : strMap.entrySet()) {
      double key = checkNotNull(Doubles.stringConverter().convert(entry.getKey()));
      double value = checkNotNull(Doubles.stringConverter().convert(entry.getValue()));
      builder.put(key, value);
    }
    NavigableMap<Double, Double> valMap = builder.build();
    checkWeights(valMap.values());
    return valMap;
  }

  /**
   * Returns a string containing the string representation of each of
   * {@code parts} joined with {@code delimiter}.
   *
   * @param parts the objects to join
   * @param delimiter the {@link Delimiter} to join on
   * @see Joiner
   */
  public static String join(Iterable<?> parts, Delimiter delimiter) {
    return delimiter.joiner().join(parts);
  }

  /**
   * Split a {@code sequence} into string components and make them available
   * through a (possibly-lazy) {@code Iterator}.
   *
   * @param sequence the sequence of characters to split
   * @param delimiter the {@link Delimiter} to split on
   * @see Splitter
   */
  public static Iterable<String> split(CharSequence sequence, Delimiter delimiter) {
    return delimiter.splitter().split(sequence);
  }

  /**
   * Split a {@code sequence} into string components and make them available
   * through an immutable {@code List}.
   *
   * @param sequence the sequence of characters to split
   * @param delimiter the {@link Delimiter} to split on
   */
  public static List<String> splitToList(CharSequence sequence, Delimiter delimiter) {
    return delimiter.splitter().splitToList(sequence);
  }

  /**
   * Split {@code sequence} into {@code Double} components and make them
   * available through an immutable {@code List}.
   *
   * @param sequence the sequence of characters to split
   * @param delimiter the {@link Delimiter} to split on
   */
  public static List<Double> splitToDoubleList(CharSequence sequence, Delimiter delimiter) {
    return FluentIterable
        .from(split(sequence, delimiter))
        .transform(Doubles.stringConverter())
        .toList();
  }

  /**
   * Delimiter identifiers, each of which can provide a {@link Joiner} and
   * {@link Splitter}.
   */
  public enum Delimiter {

    /** Colon (':') delimiter. */
    COLON(':'),

    /** Comma (',') delimiter. */
    COMMA(','),

    /** Dash ('-') delimiter. */
    DASH('-'),

    /** Period ('.') delimiter. */
    PERIOD('.'),

    /** Forward-slash ('/') delimiter. */
    SLASH('/'),

    /**
     * Whitespace (' ') delimiter.
     * @see CharMatcher#whitespace()
     */
    SPACE(' ', CharMatcher.whitespace()),

    /** Underscore ('_') delimiter. */
    UNDERSCORE('_');

    private Joiner joiner;
    private Splitter splitter;

    private Delimiter(char separator) {
      joiner = Joiner.on(separator).skipNulls();
      splitter = Splitter.on(separator).omitEmptyStrings().trimResults();
    }

    private Delimiter(char joinSeparator, CharMatcher splitMatcher) {
      joiner = Joiner.on(joinSeparator).skipNulls();
      splitter = Splitter.on(splitMatcher).omitEmptyStrings().trimResults();
    }

    /**
     * Returns a null-skipping {@link Joiner} on this {@code Delimiter}.
     * @see Joiner#skipNulls()
     */
    public Joiner joiner() {
      return joiner;
    }

    /**
     * Returns an empty-string-omitting and result-trimming {@link Splitter} on
     * this {@code Delimiter}.
     *
     * @see Splitter#omitEmptyStrings()
     * @see Splitter#trimResults()
     */
    public Splitter splitter() {
      return splitter;
    }
  }

  /**
   * Return a {@code String} representation of an {@code Iterable<Enum>} where
   * {@code Enum.name()} is used instead of {@code Enum.toString()}.
   *
   * @param iterable to process
   * @param enumClass
   */
  public static <E extends Enum<E>> String enumsToString(Iterable<E> iterable,
      Class<E> enumClass) {
    return addBrackets(FluentIterable
        .from(iterable)
        .transform(Enums.stringConverter(enumClass).reverse())
        .join(Delimiter.COMMA.joiner()));
  }

  /**
   * Convert an {@code Enum.name()} to a space-delimited presentation-friendly
   * string.
   *
   * @param e the {@code Enum} to generate label for
   * @param capitalize true if first letter of each word should be capitalized;
   *        false if letters should all be lowercase
   */
  public static String enumLabelWithSpaces(Enum<? extends Enum<?>> e, boolean capitalize) {
    return join(splitEnum(e, capitalize), Delimiter.SPACE);
  }

  /**
   * Convert an {@code Enum.name()} to a dash-delimited presentation-friendly
   * string.
   *
   * @param e the {@code Enum} to generate label for
   * @param capitalize true if fist letter of each word should be capitalized;
   *        false if letters should all be lowercase
   */
  public static String enumLabelWithDashes(Enum<? extends Enum<?>> e, boolean capitalize) {
    return join(splitEnum(e, capitalize), Delimiter.DASH);
  }

  /* Splits and possibly capitalizes an enum.name() */
  private static List<String> splitEnum(Enum<? extends Enum<?>> e, boolean capitalize) {
    Iterable<String> sources = split(e.name(), Delimiter.UNDERSCORE);
    List<String> results = Lists.newArrayList();
    for (String s : sources) {
      results.add(capitalize ? capitalize(s) : s.toLowerCase());
    }
    return results;
  }

  /**
   * Capitalize supplied string by converting the first {@code char} to
   * uppercase and all subsequent {@code char}s to lowercase.
   *
   * @param s the string to capitalize
   */
  public static String capitalize(String s) {
    return Character.toUpperCase(s.charAt(0)) + s.substring(1).toLowerCase();
  }

  /**
   * Convert a bracketed and comma-delimited string of numbers (e.g. [1.0, 2.0,
   * 3.0] to a {@code double[]}. This is the reverse of
   * {@link Arrays#toString(double[])} and {@link AbstractCollection#toString()}
   *
   * @param s the string to convert
   */
  public static double[] toDoubleArray(String s) {
    return Doubles.toArray(FluentIterable.from(split(trimEnds(s), Delimiter.COMMA))
        .transform(Doubles.stringConverter()).toList());
  }

  /**
   * Convert a {@code Collection<Double>} to a string of the same format
   * returned by {@link Arrays#toString(double[])} and
   * {@link AbstractCollection#toString()} , but will format the values using
   * the supplied format string. The supplied {@code format} should match that
   * expected by {@code String.format(String, Object...)}
   *
   * @param values the values to convert
   * @param format a format string
   */
  public static String toString(Collection<Double> values, String format) {
    return addBrackets(join(
        Iterables.transform(values, new FormatDoubleFunction(format, true)::apply),
        Delimiter.COMMA));
  }

  /**
   * Convert a {@code Collection<Double>} to a string of the same format
   * returned by {@link Arrays#toString(double[])} and
   * {@link AbstractCollection#toString()} , but will format the values using
   * the supplied format string. The supplied {@code format} should match that
   * expected by {@code String.format(String, Object...)}
   *
   * @param values the values to convert
   * @param format a format string
   * @param delimiter the delimiter to use
   * @param brackets {@code true} if hard brackets should be added to start and
   *        end, {@code false} otherwise
   */
  public static String toString(
      Collection<Double> values,
      String format,
      String delimiter,
      boolean brackets,
      boolean cleanZeros) {
    String base = Joiner.on(delimiter).join(
        Iterables.transform(
            values,
            new FormatDoubleFunction(format, cleanZeros)::apply));
    return brackets ? addBrackets(base) : base;
  }

  /**
   * Convert a {@code DoubleStream} to a string of the same format returned by
   * {@link Arrays#toString(double[])} and {@link AbstractCollection#toString()}
   * , but will format the values using the supplied format string. The supplied
   * {@code format} should match that expected by
   * {@code String.format(String, Object...)}
   *
   * @param stream the stream to convert
   * @param format a format string
   * @param delimiter the delimiter to use
   * @param brackets {@code true} if hard brackets should be added to start and
   *        end, {@code false} otherwise
   */
  public static String toString(
      DoubleStream stream,
      String format,
      String delimiter,
      boolean brackets,
      boolean cleanZeros) {
    List<Double> values = stream.boxed().collect(Collectors.toList());
    return toString(values, format, delimiter, brackets, cleanZeros);
  }

  /**
   * Convert an ordered list of non-repeating {@code Integer}s to a more compact
   * string form. For example,
   * {@code List<Integer>.toString() = "[1, 2, 3, 4, 10, 19, 18, 17, 16]"} would
   * instead be written as {@code "1:4,10,19:16"}.
   *
   * @param values to convert
   * @throws IllegalArgumentException if {@code values} is empty or if
   *         {@code values} contains adjacent repeating values
   * @see #rangeStringToIntList(String)
   */
  public static String intListToRangeString(List<Integer> values) {
    checkArgument(checkNotNull(values).size() > 0, "values may not be empty");
    // Given a list containing a single value, return that value as a string
    if (values.size() == 1) {
      return Integer.toString(values.get(0));
    }
    List<int[]> ranges = Lists.newArrayList();
    int start = values.get(0);
    int end = values.get(0);
    boolean buildingRange = false;
    boolean dir = true;

    for (int i = 1; i < values.size(); i++) {
      int next = values.get(i);
      checkArgument(next != end, "repeating value %s in %s", next, values);
      boolean currentDir = next > end;
      boolean terminateRange =
          // step > 1
          (Math.abs(next - end) != 1) ||
              // direction change
              (buildingRange && currentDir != dir) ||
              // end of list
              (i == values.size() - 1);

      if (terminateRange) {
        if (i == values.size() - 1) {
          // singleton trailing value
          if (Math.abs(next - end) == 1 && currentDir == dir) {
            ranges.add(new int[] { start, next });
          } else {
            ranges.add(new int[] { start, end });
            ranges.add(new int[] { next, next });
          }
        } else {
          ranges.add(new int[] { start, end });
        }
        start = next;
        end = next;
        buildingRange = false;
        continue;
      }

      // starting or continuing new range
      buildingRange = true;
      dir = currentDir;
      end = next;
    }
    return join(Iterables.transform(ranges, IntArrayToString.INSTANCE::apply), Delimiter.COMMA);
  }

  /**
   * Complement of {@link #intListToRangeString} converts a string of the form
   * {@code "1:4,10,19:16"} to an ordered {@code List<Integer>} (e.g.
   * {@code [1, 2, 3, 4, 10, 19, 18, 17, 16]}). This method should only be
   * called with strings created by {@link #intListToRangeString}, otherwise
   * results are undefined.
   *
   * @param s the string to convert
   * @see #intListToRangeString
   */
  public static List<Integer> rangeStringToIntList(String s) {
    Iterable<int[]> values = Iterables.transform(split(s, Delimiter.COMMA),
        StringToIntArray.INSTANCE::apply);
    return Ints.asList(Ints.concat(Iterables.toArray(values, int[].class)));
  }

  // internal use only - no argument checking
  // writes 2-element int[]s as 'a:b' or just 'a' if a==b
  private enum IntArrayToString implements Function<int[], String> {
    INSTANCE;

    @Override
    public String apply(int[] ints) {
      return (ints[0] == ints[1]) ? Integer.toString(ints[0]) : ints[0] + ":" + ints[1];
    }
  }

  // close complement of above, converts IntArrayToStrings to List<Integer>s
  private enum StringToIntArray implements Function<String, int[]> {
    INSTANCE;

    @Override
    public int[] apply(String s) {
      if (s.contains(":")) {
        Iterator<Integer> rangeIt = Iterators.transform(
            split(s, Delimiter.COLON).iterator(), Ints.stringConverter());
        return Indexing.indices(rangeIt.next(), rangeIt.next());
      }
      return new int[] { Integer.parseInt(s) };
    }
  }

  /**
   * Strip trailing zeros of a decimal number that has already been formatted as
   * a string. Method will leave a single zero after a decimal. Method will not
   * alter exponential forms, which may have a critical '0' in last position.
   *
   * @param s the string to clean
   */
  public static String stripZeros(String s) {
    if (s.charAt(1) == '.' &&
        (s.contains("e-") || s.contains("e+") || s.contains("E-") || s.contains("E+"))) {
      return s;
    }
    s = s.replaceAll("0*$", "");
    if (s.endsWith(".")) {
      s += "0";
    }
    return s;
  }

  /**
   * Returns a new {@code String} created by first trimming the supplied
   * {@code String} of leading and trailing whitespace (via
   * {@code String.trim()}), and then removing the first and last characters. At
   * a minimum, this method will return an empty string.
   *
   * @param s {@code String} to trim
   * @throws IllegalArgumentException if the size of the supplied string is
   *         &lt;2 after leading and trailing whitespace has been removed
   */
  public static String trimEnds(String s) {
    String trimmed = s.trim();
    checkArgument(trimmed.length() > 1, "\"%s\" is too short", s);
    return trimmed.substring(1, trimmed.length() - 1);
  }

  /*
   * Adds brackets to the supplied string.
   */
  static String addBrackets(String s) {
    return '[' + s + ']';
  }

  private static final Splitter.MapSplitter QUERY_SPLITTER = Splitter.on('&').trimResults()
      .withKeyValueSeparator('=');

  private static final Joiner.MapJoiner QUERY_JOINER = Joiner.on('&').withKeyValueSeparator("=");

  /**
   * Convert a {@code Map} to a {@code URL} name-value pair query string.
   * @param params the parameter map to convert
   */
  public static String mapToQuery(Map<String, String> params) {
    return "?" + QUERY_JOINER.join(params);
  }

  /**
   * Convert a {@code URL} name-value pair query string to a {@code Map}.
   * @param query the string to convert
   */
  public static Map<String, String> queryToMap(String query) {
    if (query.startsWith("?")) {
      query = query.substring(1);
    }
    return QUERY_SPLITTER.split(query);
  }

  /**
   * Read and return the {@code int} at {@code position} in a space-delimited
   * string.
   *
   * @param s the string to read from
   * @param position the position to read
   */
  public static int readInt(String s, int position) {
    return Integer.parseInt(Iterables.get(split(s, Delimiter.SPACE), position));
  }

  /**
   * Read and return the {@code double} at {@code position} in a space-delimited
   * string.
   *
   * @param s the string to read from
   * @param position the position to read
   */
  public static double readDouble(String s, int position) {
    return Double.valueOf(Iterables.get(split(s, Delimiter.SPACE), position));
  }

  /**
   * Strip a trailing comment from a string that starts with the supplied
   * character and return the cleaned string.
   *
   * @param s the string to process
   * @param c the comment indicator character
   */
  public static String stripComment(String s, char c) {
    int index = s.indexOf(c);
    return index != -1 ? s.substring(0, index) : s;
  }

  /**
   * Read the specified number of strings from an {@code Iterator} and return
   * them as a {@code List}.
   *
   * @param it the string {@code Iterator} to read from
   * @param n the number of strings to read
   */
  public static List<String> toLineList(Iterator<String> it, int n) {
    List<String> lines = new ArrayList<String>();
    for (int i = 0; i < n; i++) {
      lines.add(it.next());
    }
    return lines;
  }

  /**
   * Returns a {@link Function} for converting {@code double}s to formatted
   * strings. If a value to format is 0.0, the format string is ignored in favor
   * of always printing the often more compact string: "0.0".
   *
   * @param format a format string
   * @see String#format(String, Object...)
   */
  public static Function<Double, String> formatDoubleFunction(String format) {
    return new FormatDoubleFunction(format, true);
  }

  private static class FormatDoubleFunction implements Function<Double, String> {
    private String format;
    private boolean cleanZeros;

    private FormatDoubleFunction(String format, boolean cleanZeros) {
      this.format = format;
      this.cleanZeros = cleanZeros;
    }

    @Override
    public String apply(Double value) {
      return (cleanZeros && value == 0.0) ? "0.0" : String.format(format, value);
    }
  }

  /**
   * Reads a binary {@code byte} stream as a sequence of {@code double}s. Method
   * closes the supplied {@code InputStream} before returning.
   *
   * @param in the {@code InputStream} to read from
   * @param byteCount number of bytes to read
   * @return a {@code List<Double>}
   * @throws IOException if there is a problem reading the supplied stream
   */
  public static List<Double> readBinaryDoubleList(InputStream in, int byteCount)
      throws IOException {
    checkArgument(byteCount % 8 == 0, "byte count incompatible with doubles");
    checkArgument(byteCount > 0, "byte count too small");
    if (!(checkNotNull(in) instanceof BufferedInputStream)) {
      in = new BufferedInputStream(in);
    }
    DataInputStream din = new DataInputStream(in);
    int size = byteCount / 8;
    double[] vals = new double[size];
    for (int i = 0; i < size; i++) {
      vals[i] = din.readDouble();
    }
    in.close();
    return Doubles.asList(vals);
  }

  /**
   * Reads a binary {@code byte} stream as a sequence of {@code int}s. Method
   * closes the supplied {@code InputStream} before returning.
   *
   * @param in the {@code InputStream} to read from
   * @return a {@code List} of {@code List<Integer>}s
   * @throws IOException if there is a problem reading the supplied
   *         {@code InputStream}
   */
  public static List<List<Integer>> readBinaryIntLists(InputStream in) throws IOException {
    if (!(checkNotNull(in) instanceof BufferedInputStream)) {
      in = new BufferedInputStream(in);
    }
    DataInputStream din = new DataInputStream(in);
    int count = din.readInt();
    checkState(count > 0, "Number of lists must be > 0");
    List<List<Integer>> list = Lists.newArrayList();
    for (int i = 0; i < count; i++) {
      int[] vals = new int[din.readInt()];
      for (int j = 0; j < vals.length; j++) {
        vals[j] = din.readInt();
      }
      list.add(Ints.asList(vals));
    }
    in.close();
    return list;
  }

  /**
   * This method is currently unused and slated for removal.
   *
   * @param in
   * @param bitSetSize
   * @throws IOException
   */
  @Deprecated
  public static List<BitSet> readBinaryIntBitSets(InputStream in, int bitSetSize)
      throws IOException {
    if (!(checkNotNull(in) instanceof BufferedInputStream)) {
      in = new BufferedInputStream(in);
    }
    DataInputStream din = new DataInputStream(in);
    int count = din.readInt();
    checkState(count > 0, "Number of lists must be > 0");
    List<BitSet> list = Lists.newArrayList();
    for (int i = 0; i < count; i++) {
      int size = din.readInt();
      BitSet bits = new BitSet(bitSetSize);
      for (int j = 0; j < size; j++) {
        bits.set(din.readInt());
      }
      list.add(bits);
    }
    in.close();
    return list;
  }

}
