package gov.usgs.earthquake.nshmp.mfd;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkMagnitude;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.GR;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.GR_MMAX_GR;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.GR_MMAX_SINGLE;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.GR_TAPER;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.SINGLE;
import static gov.usgs.earthquake.nshmp.mfd.Mfds.checkGaussianMagnitudeCount;
import static gov.usgs.earthquake.nshmp.mfd.Mfds.checkGaussianSigma;
import static gov.usgs.earthquake.nshmp.mfd.Mfds.checkGaussianSigmaCount;
import static gov.usgs.earthquake.nshmp.mfd.Mfds.checkGrBvalue;
import static gov.usgs.earthquake.nshmp.mfd.Mfds.checkRate;

import java.util.Arrays;
import java.util.EnumSet;
import java.util.Optional;
import java.util.Set;
import java.util.function.Consumer;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.Sequences;
import gov.usgs.earthquake.nshmp.data.XyPoint;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.TaperedGr;

/**
 * Entry point for creating a magnitude frequency distribution (MFD). An MFD
 * defines the incremental rate of earthquakes at one or more magnitudes. MFD
 * values are backed by an immutable {@link XySequence}, however their
 * construction is varied.
 *
 * <p>Factory methods are provided to directly create incremental MFD builders
 * for different MFD {@link Type types}. Note that the {@link Builder} class
 * also includes factory methods to initialize builders from data or a copy of
 * an existing MFD. Once obtained, a builder can be used to scale and shape the
 * MFD prior to calling {@code build()}. Example:
 *
 * <pre>
 * Mfd mfd = Mfds.newGutenbergRighterBuilder(5.0, 7.5, 0.1, 1.0)
 *     .scaleToCumulativeRate(0.02)
 *     .transform(p -&gt; p.set((p.x() &gt; 6.5) ? p.y() * 0.667 : p.y()))
 *     .build()
 * </pre>
 *
 * In the example above, an MFD with a <a
 * href="https://en.wikipedia.org/wiki/Gutenberg–Richter_law">
 * Gutenberg–Richter</a> distribution of rates is initialized from {@code M=5.0}
 * to {@code M=7.5} with a step size of {@code 0.1} magnitude units. The
 * resultant MFD will have magnitudes
 * {@code [5.05, 5.15, 5.25, ... , 7.35, 7.45]} centered on the 0.1 magnitude
 * intervals spanning the magnitude range {@code [5.0..7.5]}. The MFD is then
 * scaled to the cumulative rate of 0.02 events/year, and lastly those rates
 * above {@code M=6.5} are scaled to ⅔ of their present value. Note that an MFD
 * builder executes methods in the order defined, so reversing
 * {@code scaleToCumulativeRate()} and {@code transform()} in the example above
 * results in two different MFDs. Note also that builders are reusable, but any
 * calls to {@link Builder#transform(Consumer)} can irrevocably change the state
 * of the builder between {@code build()} calls.
 *
 * <p>MFD {@link Properties} hold references to the parameters used to
 * initialize an MFD, including its {@link Type}. Properties objects may be
 * constructed directly and support a {@link Properties#toBuilder()} method. The
 * the {@code Mfd.Type} will be {@code INCREMENTAL} if the MFD was created
 * directly from magnitude and rate data or is the result of combining more than
 * one MFD via {@link Mfds#combine(java.util.Collection)}. Similarly, properties
 * <i>may</i> not align with an associated MFD downstream if a builder
 * initialized with a set of properties has been used to create multiple MFDs
 * with variably scaled rates or transformed in various ways. Note also that
 * when MFDs are {@link Mfds#combine(java.util.Collection) combined}, the
 * resulting combined MFD will be of type: {@link Type#INCR}.
 *
 * <p>Users of MFDs are encouraged take advantage of the conveniences of the
 * underlying {@code XySequence}, especially streaming support. For example, to
 * compute the cumulative rate of events in an MFD, one should use
 * {@code mfd.data().yValues().sum()}.
 *
 * @author U.S. Geological Survey
 * @see Mfds
 * @see XySequence
 */
public final class Mfd {

  /*
   * Developer notes:
   *
   * Builders: use of no-arg constructors ensure correct deserialization
   * behavior where existing fields such as GutenbergRichter.a and Single.rate
   * will not be overridden if absent from JSON.
   *
   * Builders: MFD properties are checked in toBuilder() methods allowing us to
   * verify values set via JSON deserialization. toBuilder() methods
   * additionally check that type is correctly set to defend against
   * deserialization errors.
   */

  private final XySequence data;
  private final Properties props;

  private Mfd(XySequence data, Properties props) {
    this.data = data;
    this.props = props;
  }

  /**
   * The immutable values representing this MFD.
   */
  public XySequence data() {
    return data;
  }

  /**
   * The properties used to initialize the MFD.
   */
  public Properties properties() {
    return props;
  }

  @Override
  public String toString() {
    return "MFD:" + props + Text.NEWLINE + data;
  }

  /**
   * Create an MFD from copies of the supplied arrays. The returned MFD will be
   * of type: {@link Type#INCR}.
   *
   * @param magnitudes to wrap in a MFD
   * @param rates to wrap in a MFD
   * @throws IllegalArgumentException if {@code magnitudes} is empty
   * @throws IllegalArgumentException if {@code magnitudes} and {@code rates}
   *         are not the same size
   * @throws IllegalArgumentException if {@code magnitudes} does not increase
   *         monotonically or contains repeated values
   * @throws IllegalArgumentException if {@code magnitudes} or {@code rates}
   *         contain invalid values per
   *         {@link Earthquakes#checkMagnitude(double)} and
   *         {@link Mfds#checkRate(double)}
   */
  public static Mfd create(double[] magnitudes, double[] rates) {
    return create(XySequence.create(magnitudes, rates));
  }

  /**
   * Create an MFD from the supplied sequence. The returned MFD will be of type:
   * {@link Type#INCR}. The supplied sequence will only be copied if necessary.
   *
   * @param xy data to wrap in an MFD
   * @throws IllegalArgumentException if sequence contains invalid magnitudes
   *         (x-values) or rates (y-values) per
   *         {@link Earthquakes#checkMagnitude(double)} and
   *         {@link Mfds#checkRate(double)}
   */
  public static Mfd create(XySequence xy) {
    return create(xy, new Properties());
  }

  /**
   * Create an MFD from the supplied sequence and properties. This method
   * assumes the supplied sequence has been created consistent with the supplied
   * properties. Users will have little use for this method; it is used
   * internally for magnitude array optimizations.
   *
   * @param xy data to wrap in an MFD
   * @param props
   * @throws IllegalArgumentException if sequence contains invalid magnitudes
   *         (x-values) or rates (y-values) per
   *         {@link Earthquakes#checkMagnitude(double)} and
   *         {@link Mfds#checkRate(double)}
   */
  public static Mfd create(XySequence xy, Properties props) {
    Mfds.checkValues(xy.xValues(), xy.yValues());
    return new Mfd(XySequence.copyOf(xy), props);
  }

  /**
   * Create an {@code Mfd.Builder} for a single magnitude with a rate of one.
   *
   * @param magnitude
   * @throws IllegalArgumentException if magnitude is outside the range
   *         {@code [-2.0..9.7]}
   */
  public static Mfd.Builder newSingleBuilder(double magnitude) {
    return new Properties.Single(magnitude).toBuilder();
  }

  /**
   * Create an {@code Mfd.Builder} for a doubly-truncated Gaussian MFD. Gaussian
   * MFDs are typically used to represent aleatory variability in magnitude.
   * When building the underlying magnitude distribution, this implementation
   * rounds each magnitude to 5 decimal places. This implementation constructs a
   * distribution that is {@code m ± 2σ} wide and represented by {@code nm}
   * evenly distributed magnitude values. The initial distribution integrates to
   * one. The MFD, once built, will be of type: {@link Type#SINGLE} as it
   * represents a single magnitude combined with a model of uncertainty. In the
   * context of a hazard model, a normal distribution of magnitudes is typically
   * used to represent aleatory variability.
   *
   * @param m the mean magnitude
   * @param nm the number of magnitudes in the resultant distribution
   * @param sigma the standard deviation
   * @param nSigma the number of standard deviations about {@code m}
   * @throws IllegalArgumentException if {@code m} is outside the range
   *         {@code [-2.0..9.7]}
   * @throws IllegalArgumentException if {@code nm} is outside the range
   *         {@code [1..21]}
   * @throws IllegalArgumentException if {@code sigma} is outside the range
   *         {@code (0..0.5]}
   * @throws IllegalArgumentException if {@code nSigma} is outside the range
   *         {@code [0..5]}
   */
  public static Mfd.Builder newGaussianBuilder(
      double m, int nm,
      double sigma, int nSigma) {
    return new Properties.Single(m).toGaussianBuilder(nm, sigma, nSigma);
  }

  /**
   * Create an {@code Mfd.Builder} for a truncated <a
   * href="https://en.wikipedia.org/wiki/Gutenberg–Richter_law">
   * Gutenberg–Richter</a> MFD. When building the underlying magnitude
   * distribution, this implementation rounds each magnitude to 4 decimal
   * places. Note that {@code mMin} and {@code mMax} define the lowermost and
   * uppermost magnitude bin edges in a magnitude distribution, respectively,
   * and that magnitude values in a Gutenberg-Richter MFD are defined as bin
   * centers. That is, if {@code mMin = 5.0} and {@code Δm = 0.1}, the first
   * magnitude in the distribution will be {@code mMin + Δm / 2 = 5.05}.
   *
   * @param b the Gutenberg-Richter b-value
   * @param mDelta the magnitude step of the distribution
   * @param mMin the minimum truncation magnitude
   * @param mMax the maximum truncation magnitude
   * @throws IllegalArgumentException if {@code mMin} or {@code mMax} is outside
   *         the range {@code [-2.0..9.7]}
   * @throws IllegalArgumentException if {@code mMin > mMax}
   * @throws IllegalArgumentException if {@code mDelta > mMax - mMin}
   * @throws IllegalArgumentException if {@code b} is outside the range
   *         {@code [0..3]}
   */
  public static Mfd.Builder newGutenbergRichterBuilder(
      double b,
      double mDelta,
      double mMin,
      double mMax) {
    return new Properties.GutenbergRichter(1.0, b, mDelta, mMin, mMax).toBuilder();
  }

  /**
   * Create an {@code Mfd.Builder} for a truncated <a
   * href="https://en.wikipedia.org/wiki/Gutenberg–Richter_law">
   * Gutenberg–Richter</a> MFD with an exponential taper. This implementation
   * uses the scalar seismic moment based distribution of Kagan (2002), Equation
   * 11 (see {@link Mfds#paretoRate(double, double, double, double)} for
   * reference) to compute scale factors for each magnitude bin that are then
   * applied to a standard Gutenberg–Richter distribution. When building the
   * underlying magnitude distribution, this implementation rounds each
   * magnitude to 4 decimal places. Note that {@code mMin} and {@code mMax}
   * define the lowermost and uppermost magnitude bin edges in a magnitude
   * distribution, respectively, and that magnitude values in a
   * Gutenberg-Richter MFD are defined as bin centers. That is, if
   * {@code mMin = 5.0} and {@code Δm = 0.1}, the first magnitude in the
   * distribution will be {@code mMin + Δm / 2 = 5.05}.
   *
   * @param mMin the minimum truncation magnitude
   * @param mMax the maximum truncation magnitude
   * @param mDelta the magnitude step of the distribution
   * @param b the Gutenberg-Richter b-value
   * @param mc the corner magnitude of the distribution
   * @throws IllegalArgumentException if {@code mMin}, {@code mMax}, or
   *         {@code mc} is outside the range {@code [-2.0..9.7]}
   * @throws IllegalArgumentException if {@code mMin > mMax} or {@code mc} is
   *         not {@code mMin < mc < mMax}
   * @throws IllegalArgumentException if {@code mDelta > mMax – mMin}
   * @throws IllegalArgumentException if {@code b} is outside the range
   *         {@code [0..3]}
   */
  public static Mfd.Builder newTaperedGutenbergRichterBuilder(
      double b,
      double mDelta,
      double mMin,
      double mMax,
      double mc) {
    return new Properties.TaperedGr(1.0, b, mDelta, mMin, mMax, mc).toBuilder();
  }

  /** Magnitude-frequency distribution (MFD) type identifier. */
  public enum Type {

    /**
     * An MFD with a single magnitude and rate.
     *
     * @see Mfd#newSingleBuilder(double)
     */
    SINGLE,

    /**
     * An MFD defining multiple magnitudes with varying rates.
     *
     * @see Mfd#create(double[], double[])
     * @see Mfd#create(XySequence)
     */
    INCR,

    /**
     * An incremental Gutenberg-Richter MFD.
     *
     * @see Mfd#newGutenbergRichterBuilder(double, double, double, double)
     */
    GR,

    /**
     * An incremental Gutenberg-Richter MFD with a rate scaling hint. This type
     * of MFD is used with NSHM grid sources to indicate that rates above
     * {@code M=6.5} (the minimum magnitude for finite faults) should be set to
     * zero if certain conditions are met. This type of MFD can not be created
     * directly.
     */
    GR_MMAX_GR,

    /**
     * An incremental Gutenberg-Richter MFD with a rate scaling hint. This type
     * of MFD is used with NSHM grid sources to indicate rates above some
     * 'characteristic' fault magnitude should be set to zero if certain
     * conditions are met. This type of MFD can not be created directly.
     */
    GR_MMAX_SINGLE,

    /**
     * A Gutenberg-Richter MFD with a tapered upper tail.
     *
     * @see Mfd#newTaperedGutenbergRichterBuilder(double, double, double,
     *      double, double)
     */
    GR_TAPER;
  }

  /**
   * Magnitude frequency distribution (MFD) builder. A builder may be obtained
   * via the factory methods in the {@link Mfd} class, or via factory methods
   * herein that copy existing data. An {@code Mfd.Builder} is an intermediate
   * class that can modify and shape an MFD prior to calling {@link #build()}.
   *
   * @see Mfd
   */
  public static class Builder {

    private final Properties props;
    private final MutableXySequence mfd;

    /**
     * Create an {@code Mfd.Builder} initialized with copies of the supplied
     * arrays. MFDs created from this builder will be of type:
     * {@link Type#INCR}.
     *
     * @param magnitudes to initialize builder with
     * @param rates to initialize builder with
     */
    public static Builder from(double[] magnitudes, double[] rates) {
      return new Builder(new Properties(), magnitudes, rates);
    }

    /**
     * Create an {@code Mfd.Builder} initialized with the supplied sequence.
     * MFDs created from this builder will be of type: {@link Type#INCR}.
     *
     * @param xy sequence to initialize builder with
     */
    public static Builder from(XySequence xy) {
      return new Builder(xy);
    }

    /**
     * Create an {@code Mfd.Builder} initialized with the supplied MFD. MFDs
     * created from this builder will be of the same type as the supplied MFD.
     *
     * @param mfd to initialize builder with
     */
    public static Builder from(Mfd mfd) {
      return new Builder(mfd);
    }

    private Builder(XySequence xy) {
      Mfds.checkValues(xy.xValues(), xy.yValues());
      this.props = new Properties();
      this.mfd = MutableXySequence.copyOf(xy);
    }

    private Builder(Mfd mfd) {
      this.props = mfd.props;
      this.mfd = MutableXySequence.copyOf(mfd.data);
    }

    private Builder(Properties props, double[] magnitudes) {
      this(props, magnitudes, new double[magnitudes.length]);
    }

    private Builder(Properties props, double[] magnitudes, double[] rates) {
      Mfds.checkValues(Arrays.stream(magnitudes), Arrays.stream(rates));
      this.props = props;
      this.mfd = MutableXySequence.create(magnitudes, Optional.of(rates));
    }

    /** Return a newly created {@code Mfd}. */
    public Mfd build() {
      return new Mfd(XySequence.copyOf(mfd), props);
    }

    /**
     * Scales the entire MFD by the supplied value, for example, a weight.
     *
     * @param scale to multiply MFD by
     * @return this {@code Builder} object
     */
    public Builder scale(double scale) {
      mfd.multiply(scale);
      return this;
    }

    /**
     * Scales the MFD being built to the supplied cumulative rate.
     *
     * @param cumulativeRate to scale MFD to
     * @return this {@code Builder} object
     */
    public Builder scaleToCumulativeRate(double cumulativeRate) {
      return scale(checkRate(cumulativeRate) / mfd.yValues().sum());
    }

    /**
     * Scales the MFD being built to the supplied incremental rate. The supplied
     * incremental rate is assumed to be that of the first, or minimum,
     * magnitude bin of the MFD under construction.
     *
     * @param incrementalRate to scale MFD to
     * @return this {@code Builder} object
     */
    public Builder scaleToIncrementalRate(double incrementalRate) {
      // needs sensitivity testing
      if (props.type == Type.GR_TAPER) {
        return TaperedGr.taperCorrection(this, incrementalRate);
      }
      return scale(checkRate(incrementalRate) / mfd.y(0));
    }

    /**
     * Scales the MFD being built to the supplied total moment rate.
     *
     * @param momentRate to scale MFD to
     * @return this {@code Builder} object
     */
    public Builder scaleToMomentRate(double momentRate) {
      return scale(checkRate(momentRate) / Mfds.momentRate(mfd));
    }

    /**
     * Transforms the rates in the MFD being built by applying the supplied
     * action to each {@link XyPoint}.
     *
     * @param action to perform on each MFD element
     * @return this {@code Builder} object
     */
    public Builder transform(Consumer<XyPoint> action) {
      mfd.transform(checkNotNull(action));
      return this;
    }
  }

  /**
   * Properties object associated with a MFD. A properties object wraps the
   * original parameters required to initialize a {@link Mfd.Builder}.
   */
  public static class Properties {

    private final Type type;

    /**
     * General purpose properties object for use with MFDs of type:
     * {@link Type#INCR}
     */
    public Properties() {
      this.type = Type.INCR;
    }

    Properties(Type type) {
      this.type = type;
    }

    /** The MFD type. */
    public Type type() {
      return type;
    }

    /**
     * Properties for a single MFD.
     *
     * @throws ClassCastException if properties are for other MFD type
     */
    public Single getAsSingle() {
      return (Single) this;
    }

    /**
     * Properties for a Gutenberg-Richter MFD.
     *
     * @throws ClassCastException if properties are for other MFD type
     */
    public GutenbergRichter getAsGr() {
      return (GutenbergRichter) this;
    }

    /**
     * Properties for a tapered Gutenberg-Richter MFD.
     *
     * @throws ClassCastException if properties are for other MFD type
     */
    public TaperedGr getAsGrTaper() {
      return (TaperedGr) this;
    }

    /** Return a MFD builder initialized with this properties object. */
    public Mfd.Builder toBuilder() {
      throw new UnsupportedOperationException();
    }

    @Override
    public String toString() {
      return type().toString();
    }

    /** Properties of a single magnitude MFD. */
    public static final class Single extends Properties {

      private final double m;
      private final double rate;

      /* No-arg constructor for deserialization. */
      private Single() {
        m = Double.NaN;
        rate = 1.0;
      };

      /**
       * Create a new properties object for a single MFD.
       *
       * @param magnitude of the distribution
       * @param rate of the sole magnitude in the distribution
       */
      public Single(double magnitude, double rate) {
        super(SINGLE);
        this.m = magnitude;
        this.rate = rate;
      }

      /**
       * Create a new properties object for a single MFD with a rate of one.
       *
       * @param magnitude of the distribution
       */
      public Single(double magnitude) {
        this(magnitude, 1.0);
      }

      /** The sole magnitude of the MFD */
      public double magnitude() {
        return m;
      }

      /** The rate of the sole magnitude. */
      public double rate() {
        return rate;
      }

      @Override
      public Builder toBuilder() {
        checkState(type().equals(SINGLE));
        /* Values checked in builder */
        return new Builder(
            this,
            new double[] { this.m },
            // this needs improvement
            new double[] { this.rate == 0.0 ? 1.0 : this.rate });
      }

      public Builder toGaussianBuilder(int nm, double σ, int nσ) {
        checkState(type().equals(SINGLE));
        checkGaussianMagnitudeCount(nm);
        checkGaussianSigma(σ);
        checkGaussianSigmaCount(nσ);
        /* Pre-check magnitudes; final arrays are checked in builder */
        double μ = checkMagnitude(this.m);
        double mMin = Maths.round(checkMagnitude(μ - nσ * σ), 6);
        double mMax = Maths.round(checkMagnitude(μ + nσ * σ), 6);
        double Δm = Maths.round((mMax - mMin) / (nm - 1), 6);
        double[] magnitudes = Sequences.arrayBuilder(mMin, mMax, Δm)
            .centered()
            .scale(5)
            .build();
        Builder builder = new Builder(this, magnitudes);
        builder.mfd.transform(p -> p.set(Maths.normalPdf(μ, σ, p.x())));
        builder.scaleToCumulativeRate(rate);
        return builder;
      }

      @Override
      public String toString() {
        String props = new StringBuilder()
            .append("magnitude=").append(m)
            .append(", rate=").append(rate)
            .toString();
        return Mfds.propsToString(type(), props);
      }
    }

    /** Properties of a Gutenberg–Richter MFD. */
    public static class GutenbergRichter extends Properties {

      private final double a;
      private final double b;
      private final double Δm;
      private final double mMin;
      private final double mMax;

      /* Dummy, no-arg constructor for deserialization. */
      private GutenbergRichter() {
        this.a = 1.0;
        this.b = Double.NaN;
        this.Δm = Double.NaN;
        this.mMin = Double.NaN;
        this.mMax = Double.NaN;
      }

      private GutenbergRichter(
          Type type, double a, double b,
          double Δm, double mMin, double mMax) {

        super(type);
        this.a = a;
        this.b = b;
        this.Δm = Δm;
        this.mMin = mMin;
        this.mMax = mMax;
      }

      /**
       * Create a new properties object for a Gutenberg-Richter MFD.
       *
       * @param a value of the distribution
       * @param b value of the distribution
       * @param mDelta magnitude bin width of the distribution
       * @param mMin maximum magnitude of the distribution
       */
      public GutenbergRichter(double a, double b, double mDelta, double mMin, double mMax) {
        this(GR, a, b, mDelta, mMin, mMax);
      }

      /**
       * Create a new Gutenberg-Richter MFD properties object from a copy of the
       * supplied properties and an updated maximum magnitude.
       *
       * @param gr properties to copy
       * @param mMax maximum magnitude
       */
      public static GutenbergRichter copyOfWithMax(GutenbergRichter gr, double mMax) {
        /* Supports common magnitude arrays for variable mMax MFDs. */
        return new GutenbergRichter(gr.type(), gr.a, gr.b, gr.Δm, gr.mMin, mMax);
      }

      /**
       * Create a new Gutenberg-Richter MFD properties object from a copy of the
       * supplied properties and an updated a-value.
       *
       * @param gr properties to copy
       * @param a value
       */
      public static GutenbergRichter copyOfWithRate(GutenbergRichter gr, double a) {
        /* Supports prescriptive a-values when mo-rate balancing. */
        return new GutenbergRichter(gr.type(), a, gr.b, gr.Δm, gr.mMin, gr.mMax);
      }

      /** The Gutenberg-Richter {@code a}-value. */
      public double a() {
        return a;
      }

      /** The Gutenberg-Richter {@code b}-value. */
      public double b() {
        return b;
      }

      /** The magnitude bin width. */
      public double Δm() {
        return Δm;
      };

      /** The minimum magnitude. */
      public double mMin() {
        return mMin;
      };

      /** The maximum magnitude. */
      public double mMax() {
        return mMax;
      };

      private static final Set<Type> GR_TYPES = EnumSet.of(
          GR,
          GR_MMAX_GR,
          GR_MMAX_SINGLE);

      @Override
      public Builder toBuilder() {
        checkState(GR_TYPES.contains(type()));
        checkGrBvalue(this.b);
        /* Pre-check magnitudes; final arrays are checked in builder */
        double mMin = checkMagnitude(this.mMin);
        double mMax = checkMagnitude(this.mMax);
        double[] magnitudes = Sequences.arrayBuilder(mMin, mMax, this.Δm)
            .scale(5)
            .build();
        Builder builder = new Builder(this, magnitudes);
        builder.transform(p -> p.set(Mfds.gutenbergRichterRate(a, b, p.x())));
        return builder;
      }

      @Override
      public String toString() {
        return Mfds.propsToString(type(), propsString());
      }

      private String propsString() {
        return new StringBuilder()
            .append("a=").append(a)
            .append(", b=").append(b)
            .append(", Δm=").append(Δm)
            .append(", mMin=").append(mMin)
            .append(", mMax=").append(mMax)
            .toString();
      }
    }

    /** Tapered Gutenberg–Richter MFD properties. */
    public static final class TaperedGr extends GutenbergRichter {

      private final double mc;

      /* Dummy, no-arg constructor for deserialization. */
      private TaperedGr() {
        this.mc = Double.NaN;
      }

      /**
       * Create a new properties object for a tapered Gutenberg-Richter MFD.
       *
       * @param a value of the distribution
       * @param b value of the distribution
       * @param mDelta magnitude bin width of the distribution
       * @param mMin minimum magnitude of the distribution
       * @param mc corner magnitude of the distribution
       */
      public TaperedGr(double a, double b, double mDelta, double mMin, double mMax, double mc) {
        super(GR_TAPER, a, b, mDelta, mMin, mMax);
        this.mc = mc;
      }

      /**
       * Create a new properties object for a tapered Gutenberg-Richter MFD with
       * a copy of the supplied properties and an updated maximum magnitude.
       *
       * @param gr properties to copy
       * @param mMax updated maximum magnitude
       */
      public TaperedGr(TaperedGr gr, double mMax) {
        /* Supports common magnitude arrays for variable mMax MFDs */
        this(gr.a(), gr.b(), gr.Δm(), gr.mMin(), mMax, gr.mc);
      }

      /** The corner magnitude. */
      public final double mc() {
        return mc;
      }

      @Override
      public Builder toBuilder() {
        checkState(type().equals(GR_TAPER));
        checkGrBvalue(b());
        /* Pre-check magnitudes; final arrays are checked in builder */
        double mMin = checkMagnitude(mMin());
        double mMax = checkMagnitude(mMax());
        double mc = checkMagnitude(this.mc);
        double[] magnitudes = Sequences.arrayBuilder(mMin, mMax, Δm())
            .scale(4)
            .build();
        Builder builder = new Builder(this, magnitudes);
        TaperFunction grTaper = new TaperFunction(Δm(), b(), mc);
        builder.transform(p -> p.set(taperedRate(p, grTaper)));

        /*
         * The tapering function actually does not recover the un-tapered rates;
         * it's close, but not exact. The difference, however, is uniform across
         * all m-bins. In order to get the same result by creating a tapered GR
         * directly with a given a-value and by creating a generic tapered MFD
         * that we then scaleToIncrementalRate() using the given a-value, we
         * correct the tapered MFD by the offset.
         *
         * Now, this is how we wished the nshmp-haz implementation behaved, but
         * since it doesn't we need an intermediate solution to match nshmp-haz
         * until further sensitivity testing through to hazard can be
         * undertaken. What we want: after calling scaleToIncrementalRate(), the
         * tapered GR MFD rates should match those that would've resulted from
         * the direct calculation.
         */
        double grRate = Mfds.gutenbergRichterRate(a(), b(), builder.mfd.x(0));
        double tgrRate = builder.mfd.y(0);
        double aCorrection = grRate / tgrRate;
        builder.scale(aCorrection);

        return builder;
      }

      private double taperedRate(XyPoint p, TaperFunction fn) {
        return Mfds.gutenbergRichterRate(a(), b(), p.x()) * fn.scale(p.x());
      }

      @Override
      public String toString() {
        return Mfds.propsToString(
            type(),
            super.propsString() + ", mc=" + mc);
      }

      /*
       * Clean after sensitivity testing for tapered rates when calling scale
       * toIncrementalRate() tapered GR values become exact matches to the
       * target rate as opposed to the slightly scaled rate resulting from
       * creating the tapered GR directly with a given a-value. We want to
       * replicate nshmp-haz behavior for now but get rid of it in the not too
       * distant future. This is basically reverting the scaling correction
       * applied in the builder.
       */
      private static Builder taperCorrection(Builder b, double rate) {
        TaperedGr p = b.props.getAsGrTaper();
        TaperFunction taperFn = new TaperFunction(p.Δm(), p.b(), p.mc);
        double mMin = b.mfd.x(0);
        double taperRate = checkRate(rate) * taperFn.scale(mMin);
        return b.scale(taperRate / b.mfd.y(0));
      }

    }

    private static final double M_MIN_MOMENT = Earthquakes.magToMoment(4.0);
    private static final double M_MAX_MOMENT = Earthquakes.magToMoment(9.05);

    /* Gutenberg–Richter tapering scale function */
    private static class TaperFunction {

      final double ΔmBy2;
      final double β;
      final double Mcm;

      TaperFunction(double Δm, double b, double mc) {
        ΔmBy2 = Δm / 2.0;
        β = b / 1.5;
        Mcm = Earthquakes.magToMoment(mc);
      }

      /* Resolve bin edges and return scale factor */
      double scale(double m) {
        double Mlo = Earthquakes.magToMoment(m - ΔmBy2);
        double Mhi = Earthquakes.magToMoment(m + ΔmBy2);
        double binRateTapered = paretoΔ(M_MIN_MOMENT, Mlo, Mhi, β, Mcm);
        double binRateUntapered = paretoΔ(M_MIN_MOMENT, Mlo, Mhi, β, M_MAX_MOMENT);
        return binRateTapered / binRateUntapered;
      }
    }

    /*
     * Compute the number of events in a tapered GR magnitude bin. All 'M'
     * values are scalar seismic moment.
     */
    private static double paretoΔ(
        double Mt,
        double Mlo,
        double Mhi,
        double β,
        double Mcm) {

      return Mfds.paretoRate(Mt, Mlo, β, Mcm) - Mfds.paretoRate(Mt, Mhi, β, Mcm);
    }
  }
}
