package gov.usgs.earthquake.nshmp.mfd;

import static gov.usgs.earthquake.nshmp.data.DoubleData.checkInRange;
import static java.util.stream.Collectors.toList;

import java.util.Collection;
import java.util.stream.DoubleStream;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.data.Indexing;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Type;

/**
 * Utility methods for working with magnitude frequency distributions (MFDs).
 * See the {@link Mfd} class for details on the creation of MFDs.
 *
 * @author U.S. Geological Survey
 * @see Mfd
 */
public final class Mfds {

  private Mfds() {}

  /**
   * Ensure {@code rate ≥ 0.0}.
   *
   * @param rate to validate
   * @return the validated rate
   * @throws IllegalArgumentException if {@code rate} value is less than 0.0
   */
  public static double checkRate(double rate) {
    return checkInRange(Range.atLeast(0.0), "Rate", rate);
  }

  static void checkValues(DoubleStream magnitudes, DoubleStream rates) {
    magnitudes.forEach(Earthquakes::checkMagnitude);
    rates.forEach(Mfds::checkRate);
  }

  /* Check Gaussian sigma. */
  static double checkGaussianSigma(double σ) {
    return checkInRange(Range.openClosed(0.0, 0.5), "Gaussian σ", σ);
  }

  /* Check Gaussian sigma count. */
  static int checkGaussianSigmaCount(int nσ) {
    return Indexing.checkInRange(Range.closed(0, 5), "Gaussian nσ", nσ);
  }

  /* Check Gaussian magnitude count. */
  static int checkGaussianMagnitudeCount(int nm) {
    return Indexing.checkInRange(Range.closed(1, 21), "Gaussian nm", nm);
  }

  /* Check Gaussian sigma. */
  static double checkGrBvalue(double b) {
    return checkInRange(Range.closed(0.0, 3.0), "Gutenberg-Richter b-value", b);
  }

  static String propsToString(Type type, String propsString) {
    return new StringBuilder(type.toString())
        .append(" {")
        .append(propsString)
        .append("}")
        .toString();
  }

  /**
   * Combine the supplied MFDs into a single MFD, summing the rates of duplicate
   * magnitudes. The resulting combined MFD will be of type: {@link Type#INCR}.
   *
   * @param mfds sequences to combine
   * @throws IllegalArgumentException if {@code mfds} is empty
   */
  public static Mfd combine(Collection<Mfd> mfds) {
    XySequence combined = XySequence.combine(mfds.stream()
        .map(Mfd::data)
        .collect(toList()));
    return Mfd.create(combined);
  }

  /**
   * Returns the <a href="https://en.wikipedia.org/wiki/Gutenberg–Richter_law">
   * Gutenberg–Richter</a> event rate for the supplied a- and b-values and
   * magnitude. This method follows the form
   * {@code N(M) = 10<sup>(a – bM)</sup>}. If one's {@code a}-values are
   * represented as the total number of events {@code > M}, sometimes denoted as
   * {@code A} and as NSHM a-grid values commonly are, be sure to supply
   * {@code log10(A)} instead.
   *
   * <p><b>Reference:</b> Gutenberg, B., Richter, C. F., 1956, Magnitude and
   * Energy of Earthquakes, Annali di Geofisica, v. 9, p. 1–15, doi:&lt;a
   * href="https://doi.org/10.4401/ag-5590" target=_top"&gt;10.4401/ag-5590<a>
   *
   * @param a value (log10 rate of M=0 events)
   * @param b value
   * @param M magnitude of interest
   * @return the rate of magnitude {@code M} events
   */
  public static double gutenbergRichterRate(double a, double b, double M) {
    return Math.pow(10, a - b * M);
  }

  /**
   * Returns the total moment rate of an incremental MFD.
   *
   * @param mfd for which to compute total moment rate
   */
  public static double momentRate(Mfd mfd) {
    return momentRate(mfd.data());
  }

  static double momentRate(XySequence xy) {
    return xy.stream()
        .mapToDouble(p -> p.y() * Earthquakes.magToMoment(p.x()))
        .sum();
  }

  /**
   * Returns the cumulative number of events with seismic moment greater than
   * {@code M} with an exponential taper applied. This is the complementary
   * Pareto distribution as recorded in Equation 11 of Kagan (2002).
   *
   * <p><b>Reference:</b> Kagan, Y.Y., 2002, Seismic moment distribution
   * revisited: I. Statistical results, Geophysical Journal International, v.
   * 148, p. 520—541, doi:<a
   * href="https://doi.org/10.1046/j.1365-246x.2002.01594.x">
   * 10.1046/j.1365-246x.2002.01594.x</a>
   *
   * @param Mt the scalar seismic moment of catalog completeness threshold
   * @param M the scalar seismic moment of the target magnitude
   * @param beta the index parameter of the distribution; {@code β = ⅔b}
   *        (Gutenberg–Richter {@code b}-value)
   * @param Mcm the scalar seismic moment of the upper bound corner magnitude
   */
  public static double paretoRate(
      double Mt,
      double M,
      double beta,
      double Mcm) {

    return Math.pow(Mt / M, beta) * Math.exp((Mt - M) / Mcm);
  }

}
