package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static com.google.common.base.Strings.padEnd;
import static gov.usgs.earthquake.nshmp.Text.checkName;

import java.util.function.Predicate;

import com.google.common.collect.FluentIterable;

import gov.usgs.earthquake.nshmp.geo.Location;

/**
 * Skeletal {@code RuptureSet} implementation.
 *
 * @author U.S. Geological Survey
 */
abstract class AbstractRuptureSet<T extends Source> implements RuptureSet<T> {

  private final String name;
  private final int id;
  private final SourceType type;
  private final TectonicSetting setting;
  private final GmmSet gmmSet;

  private Double weight;

  AbstractRuptureSet(Builder builder) {
    this.name = builder.name;
    this.id = builder.id;
    this.type = builder.type;
    this.setting = builder.setting;
    this.gmmSet = builder.gmmSet;
  }

  @Override
  public String name() {
    return name;
  }

  @Override
  public int id() {
    return id;
  }

  @Override
  public TectonicSetting setting() {
    return setting;
  }

  @Override
  public SourceType type() {
    return type;
  }

  @Override
  public GmmSet groundMotionModels() {
    return gmmSet;
  }

  /*
   * Ideally this can be removed in time but with the current calculation
   * pipeline we need to be able to carry a source tree branch/leaf weight with
   * each rupture set. We only allow the boxed non-final weight value to be set
   * once. We don't know the source tree leaf weights until after the full tree
   * has been built.
   */
  void setLeafWeight(double weight) {
    // only allow this to be set once
    checkState(this.weight == null);
    this.weight = weight;
  }

  @Override
  public double weight() {
    return weight;
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append(id + " " + padEnd(name(), 56, ' '));
    sb.append(padEnd("size: " + size(), 14, ' '));
    return sb.toString();
  }

  @Override
  public Iterable<T> iterableForLocation(Location loc) {
    return iterableForLocation(loc, gmmSet.maxDistance());
  }

  @Override
  public Iterable<T> iterableForLocation(Location loc, double distance) {
    Predicate<T> filter = distanceFilter(loc, distance);
    return FluentIterable.from(this).filter(filter::test);
  }

  static abstract class Builder {

    boolean built = false;

    String name;
    Integer id;
    SourceType type;
    TectonicSetting setting;
    GmmSet gmmSet;

    Builder(SourceType type) {
      this.type = type;
    }

    Builder name(String name) {
      this.name = checkName(name, "RuptureSet");
      return this;
    }

    Builder id(int id) {
      checkArgument(id > 0);
      this.id = id;
      return this;
    }

    Builder setting(TectonicSetting setting) {
      this.setting = checkNotNull(setting);
      return this;
    }

    Builder gmms(GmmSet gmmSet) {
      this.gmmSet = checkNotNull(gmmSet);
      return this;
    }

    void validateState(String label) {
      checkState(!built, "Single use builder");
      checkNotNull(name, "%s name", label);
      checkNotNull(id, "%s id", label);
      checkNotNull(type, "%s source type", label);
      checkNotNull(setting, "%s tectonic setting", label);
      checkNotNull(gmmSet, "%s ground motion models", label);
      built = true;
    }
  }

}
