package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.util.Iterator;
import java.util.List;
import java.util.function.Predicate;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Cluster source representation. Each cluster source wraps a
 * {@code FaultRuptureSet} containing one or more fault representations that
 * rupture as independent events but with a similar rate. For example, at New
 * Madrid, each ClusterRuptureSet has 5 ClusterSources, one for each position
 * variant of the model. For each position variant there is one FaultRuptureSet
 * containing the FaultSources in the cluster, each of which may have one, or
 * more, magnitude or other variants represented by its internal list of
 * {@code Mfd}s.
 *
 * <p>Cluster source hazard is calculated from the joint probabilities of ground
 * motions from the wrapped faults.
 *
 * <p>A {@code ClusterSource} cannot be created directly; it may only be created
 * by a private parser.
 *
 * @author U.S. Geological Survey
 */
public class ClusterRuptureSet extends AbstractRuptureSet<ClusterSource> {

  final ModelData data; // holds cluster rate tree
  private final List<ClusterSource> source;

  private ClusterRuptureSet(Builder builder) {
    super(builder);
    this.data = builder.data;
    this.source = List.of(builder.source);
  }

  @Override
  public int size() {
    return 1; // faultRuptureSets.size();
  }

  @Override
  public LogicTree<Mfd> mfdTree() {
    System.err.println("TODO Cluster MFD tree implementation");
    // only returning first cluster source
    throw new UnsupportedOperationException();
    // return source.mfdTree;
  }

  @Override
  @Deprecated // this is problemtic as we are not getting the rate weight
  public Iterator<ClusterSource> iterator() {
    return source.iterator();
  }

  @Override
  public Predicate<ClusterSource> distanceFilter(
      final Location loc,
      final double distance) {

    return new Predicate<ClusterSource>() {

      private final Predicate<FaultSource> filter =
          new FaultRuptureSet.DistanceFilter(loc, distance);

      @Override
      public boolean test(ClusterSource cs) {
        return cs.faults().stream()
            .map(frs -> frs.iterator().next())
            .anyMatch(filter);
      }
    };
  }

  /**
   * The closest point across the traces of all fault sources that participate
   * in this cluster, relative to the supplied site {@code Location}.
   */
  @Override
  public Location location(Location site) {
    return source.get(0).location(site);
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder */
  static class Builder extends AbstractRuptureSet.Builder {

    private ModelData data; // holds cluster rate-tree
    private ClusterSource source;

    private Builder() {
      super(SourceType.FAULT_CLUSTER);
    }

    @Override
    Builder name(String name) {
      super.name(name);
      return this;
    }

    @Override
    Builder id(int id) {
      super.id(id);
      return this;
    }

    /* Set MFD helper data. */
    Builder modelData(ModelData data) {
      this.data = checkNotNull(data);
      super.setting(data.tectonicSetting());
      super.gmms(data.gmms());
      return this;
    }

    Builder setSource(ClusterSource source) {
      checkState(this.source == null);
      this.source = checkNotNull(source);
      return this;
    }

    void validateAndInit(String label) {
      validateState(label);
      checkNotNull(data, "%s model data", label);
      checkNotNull(source, "%s cluster source", label);
    }

    ClusterRuptureSet build() {
      validateAndInit("ClusterRuptureSet.Builder");
      return new ClusterRuptureSet(this);
    }
  }

  /*
   * Data container class used during deserialization. ModelLoader uses these
   * data to initialize multiple cluster sources having identical fault sources
   * but scaled to unique rates.
   */
  static class Data {

    final String name;
    final int id;
    final ModelData data;
    final List<FaultRuptureSet> faultRuptureSets;

    Data(
        String name,
        int id,
        ModelData data,
        List<FaultRuptureSet> faultRuptureSets) {

      if (faultRuptureSets.size() > 1) {
        for (int i = 1; i < faultRuptureSets.size(); i++) {
          ModelTrees.checkTreeIdsAndWeights(
              faultRuptureSets.get(0).mfdTree,
              faultRuptureSets.get(i).mfdTree);
        }
      }

      this.name = name;
      this.id = id;
      this.data = data;
      this.faultRuptureSets = faultRuptureSets;
    }
  }

}
