package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.geo.Locations.distanceToLineFast;
import static gov.usgs.earthquake.nshmp.geo.Locations.distanceToSegmentFast;
import static gov.usgs.earthquake.nshmp.geo.Locations.horzDistanceFast;

import java.awt.geom.Area;
import java.awt.geom.Path2D;
import java.util.Iterator;

import gov.usgs.earthquake.nshmp.Faults;
import gov.usgs.earthquake.nshmp.fault.surface.GriddedSurface;
import gov.usgs.earthquake.nshmp.geo.BorderType;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.LocationVector;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.Region;
import gov.usgs.earthquake.nshmp.geo.Regions;

/**
 * Distance value wrapper.
 *
 * @author U.S. Geological Survey
 */
public final class Distance {

  /**
   * Maximum supported distance for PSHA calculations. Currently set to 1000 km,
   * the maximum known to be supported by across all implementated ground motion
   * models.
   */
  @Deprecated
  public static final double MAX = 1000.0; // should we poll GMMs

  public enum Type {
    R_JB,
    R_RUP,
    R_X;
  }

  public final double rJB, rRup, rX;

  private Distance(double rJB, double rRup, double rX) {
    this.rJB = rJB;
    this.rRup = rRup;
    this.rX = rX;
  }

  static Distance create(double rJB, double rRup, double rX) {
    return new Distance(rJB, rRup, rX);
  }

  /**
   * Compute distance metrics: rJB, rRup, and rX.
   * @param surface
   * @param loc
   */
  public static Distance compute(GriddedSurface surface, Location loc) {

    Location loc1 = loc;
    Location loc2;
    double distJB = Double.MAX_VALUE;
    // double distSeis = Double.MAX_VALUE;
    double distRup = Double.MAX_VALUE;

    double horzDist, vertDist, rupDist;

    // flag to project to seisDepth if only one row and depth is below
    // seisDepth
    // boolean projectToDepth = false;
    // if (surface.getNumRows() == 1 && surface.getLocation(0,0).depth() <
    // SEIS_DEPTH)
    // projectToDepth = true;

    // get locations to iterate over depending on dip
    Iterator<Location> it;
    if (surface.dip() > 89) {
      it = surface.getColumnIterator(0);
      // if (surface.getLocation(0,0).depth() < SEIS_DEPTH)
      // projectToDepth = true;
    } else {
      it = surface.iterator();
    }

    while (it.hasNext()) {

      loc2 = it.next();

      // get the vertical distance
      vertDist = Locations.vertDistance(loc1, loc2);

      // get the horizontal dist depending on desired accuracy
      horzDist = Locations.horzDistanceFast(loc1, loc2);

      if (horzDist < distJB) {
        distJB = horzDist;
      }

      rupDist = horzDist * horzDist + vertDist * vertDist;
      if (rupDist < distRup) {
        distRup = rupDist;
      }

    }

    distRup = Math.pow(distRup, 0.5);

    if (distJB < surface.getAveGridSpacing() && isDjbZero(surface.getPerimeter(), loc)) {
      distJB = 0;
    }

    double rX = getDistanceX(surface.getEvenlyDiscritizedUpperEdge(), loc);

    return Distance.create(distJB, distRup, rX);
  }

  private static double getDistanceX(LocationList trace, Location siteLoc) {

    // Point sources used to get processed through this method.
    // Theoretically we shouldn't get here now because any trace
    // will have already been validated to have 2 or more points.
    // Conceivably the points could be closer together than the
    // spacing of a gridded surface and I'm not sure what the
    // consequence of that would be, but we should make sure that
    // doesn't happen.
    checkState(trace.size() > 1, "Trace is too short");

    // @formatter:off

    /*
     *    P4                        P1
     *    |                          |
     *    |          dip dir         |
     *    |             |            |
     *    P3------************------P2
     *            <- strike --
     */

    // @formatter:on
    double strike = Faults.strikeRad(trace);
    double dipDir = Faults.dipDirectionRad(strike);

    LocationVector toP3 = LocationVector.create(strike, 1000.0, 0.0);
    LocationVector toP2 = LocationVector.reverseOf(toP3);
    LocationVector toP14 = LocationVector.create(dipDir, 1000.0, 0.0);

    Location p3 = Locations.location(trace.last(), toP3);
    Location p4 = Locations.location(p3, toP14);
    Location p2 = Locations.location(trace.first(), toP2);
    Location p1 = Locations.location(p2, toP14);

    LocationList region = LocationList.builder()
        .add(p1)
        .add(p2)
        .addAll(trace)
        .add(p3)
        .add(p4)
        .build();

    LocationList extendedTrace = LocationList.builder()
        .add(p2)
        .addAll(trace)
        .add(p3)
        .build();

    Region polygon = null;
    try {
      polygon = Regions.create("",
          LocationList.copyOf(region),
          BorderType.MERCATOR_LINEAR);
    } catch (Exception e) {
      e.printStackTrace();
      System.out.println("==== trace  ====");
      System.out.println(trace);
      // RegionUtils.locListToKML(extendedTrace, "distX_trace",
      // Color.ORANGE);
      System.out.println("==== region ====");
      System.out.println(region);
      // RegionUtils.locListToKML(LocationList.create(region),
      // "distX_region", Color.RED);
      System.exit(0);
    }
    boolean isInside = polygon.contains(siteLoc);

    double distToExtendedTrace = Locations.minDistanceToLine(
        siteLoc,
        LocationList.copyOf(extendedTrace));

    if (isInside || distToExtendedTrace == 0.0) { // zero values are always
      // on the hanging wall
      return distToExtendedTrace;
    }
    // else
    return -distToExtendedTrace;
    // }

    // return distanceX;
  }

  // this was an experiment gone wrong; it doesn't work because there
  // are wedges of areas where points are equidistant to two segments
  // and it is difficult to determine if one is off the endpoints of
  // a segment. See Kevins distance X tests in Quad Surface and possible
  // azimuth based solution in my notes. p powers
  private static double calcDistanceX(LocationList trace, Location loc) {
    if (trace.size() == 1) {
      return 0.0;
    }

    // Compare the distance to the closest segment to the distances to the
    // endpoints. If the closest segment distance is less than both endpoint
    // distances, use that segment as a line to compute rX, otherwise use
    // endpoints of the trace as a line to compute rX
    int minIndex = Locations.minDistanceIndex(loc, trace);
    double rSeg = distanceToSegmentFast(trace.get(minIndex),
        trace.get(minIndex + 1), loc);
    double rFirst = horzDistanceFast(trace.first(), loc);
    double rLast = horzDistanceFast(trace.last(), loc);

    return (rSeg < Math.min(rFirst, rLast)) ? distanceToLineFast(
        trace.get(minIndex), trace.get(minIndex + 1), loc)
        : distanceToLineFast(trace.first(), trace.last(), loc);
  }

  /*
   * This method is used to check small distJB values for continuous, smooth
   * surfaces; e.g. non-Frankel type surfaces. This was implemented to replace
   * using a Region.contains() which can fail when dipping faults have jagged
   * traces. This method borrows from Region using a java.awt.geom.Area to
   * perform a contains test, however no checking is done of the area's
   * singularity.
   *
   * The Elsinore fault was the culprit leading to this implementation. For a
   * near-vertical (85??) strike-slip fault, it is has an unrealistic ???90 jog
   * in it. Even this method does not technically give a 100% correct answer.
   * Drawing out a steeply dipping fault with a jog will show that the resultant
   * perimeter polygon has eliminated small areas for which distJB should be
   * zero. The areas are so small though that the hazard is not likely affected.
   */
  private static boolean isDjbZero(LocationList border, Location pt) {
    Path2D path = new Path2D.Double(Path2D.WIND_EVEN_ODD, border.size());
    boolean starting = true;
    for (Location loc : border) {
      double lat = loc.latitude;
      double lon = loc.longitude;
      // if just starting, then moveTo
      if (starting) {
        path.moveTo(lon, lat);
        starting = false;
        continue;
      }
      path.lineTo(lon, lat);
    }
    path.closePath();
    Area area = new Area(path);
    return area.contains(pt.longitude, pt.latitude);
  }

}
