package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static com.google.common.base.Strings.padEnd;
import static gov.usgs.earthquake.nshmp.Text.NEWLINE;
import static java.util.stream.Collectors.toCollection;
import static java.util.stream.Collectors.toUnmodifiableMap;

import java.awt.geom.Area;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;

import com.google.common.base.Strings;
import com.google.common.collect.ImmutableListMultimap;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.google.common.collect.SetMultimap;
import com.google.common.collect.Sets;
import com.google.common.collect.Streams;
import com.google.gson.annotations.SerializedName;

import gov.usgs.earthquake.nshmp.calc.CalcConfig;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionModel;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;

/**
 * A {@code HazardModel} is the top-level wrapper for earthquake {@link Source}
 * definitions and attendant {@link GroundMotionModel}s used in probabilisitic
 * seismic hazard analyses (PSHAs) and related calculations. A
 * {@code HazardModel} contains of a number of {@link RuptureSet}s that define
 * logical groupings of sources by {@link SourceType}. Each {@code RuptureSet}
 * carries with it references to the {@code GroundMotionModel}s and associated
 * weights to be used when evaluating hazard.
 *
 * @author U.S. Geological Survey
 * @see Source
 * @see RuptureSet
 * @see SourceType
 */
public final class HazardModel implements Iterable<SourceTree> {

  private final String name;
  private final CalcConfig config;
  private final Map<NehrpSiteClass, Double> siteClassMap;

  private final Multimap<SourceType, RuptureSet<? extends Source>> ruptureSetMap;
  private final Set<TectonicSetting> settings;
  private final Map<TectonicSetting, MapRegion> mapRegionMap;
  private final SiteData siteData;

  private final Multimap<TectonicSetting, SourceTree> treesMap;
  private final Map<Integer, SourceTree> idMap;

  private HazardModel(Builder builder) {
    this.name = builder.info.name;
    this.config = builder.config;
    this.siteClassMap = Maps.immutableEnumMap(builder.info.siteClassMap);
    this.ruptureSetMap = builder.ruptureSetMap;
    this.settings = Sets.immutableEnumSet(builder.settings);
    this.mapRegionMap = Maps.immutableEnumMap(builder.mapRegionMap);
    this.siteData = builder.siteData;
    this.treesMap = builder.treesMap;
    this.idMap = builder.idMap;
  }

  /**
   * Load a {@code HazardModel} from a directory specified by the supplied
   * {@code path}.
   *
   * <p>For more information on a HazardModel directory and file structure, see
   * the <a
   * href="https://github.com/usgs/nshmp-haz/wiki/Earthquake-Source-Models">
   * nshmp-haz wiki</a>.
   *
   * <p><b>Notes:</b> HazardModel loading is not thread safe. There are also a
   * wide variety of exceptions that may be encountered when loading a model.
   * Exceptions are generally logged and propagated back to calling application
   * to handle termination.
   *
   * @param path to {@code HazardModel} directory
   * @return a newly instantiated {@code HazardModel}
   */
  public static HazardModel load(Path path) {
    return ModelLoader.load(path);
  }

  /**
   * The number of {@code RuptureSet}s in this {@code HazardModel}.
   */
  public int size() { // TODO this should be the number of trees
    return ruptureSetMap.size();
  }

  @Override
  public Iterator<SourceTree> iterator() {
    return treesMap.values().iterator();
  }

  /**
   * The name of the model.
   */
  public String name() {
    return name;
  }

  /**
   * Return the default calculation configuration. This may be overridden.
   */
  public CalcConfig config() {
    return config;
  }

  /**
   * The set of {@code Gmm}s used by this model.
   */
  public Set<Gmm> gmms() {
    return Streams.stream(this)
        .flatMap(tree -> tree.gmms().gmms().stream())
        .collect(toCollection(() -> EnumSet.noneOf(Gmm.class)));
  }

  /**
   * The set of {@code SourceType}s included in this model.
   */
  public Set<SourceType> types() {
    return ruptureSetMap.keySet();
  }

  /**
   * The set of {@code TectonicSetting}s included in this model.
   */
  public Set<TectonicSetting> settings() {
    return settings;
  }

  /**
   * Return an {@code Optional} describing the source tree for the supplied ID
   * or an empty {@code Optional} if the ID is not present in the model.
   *
   * @param id to get source tree for
   */
  public Optional<SourceTree> tree(int id) {
    return Optional.ofNullable(idMap.get(id));
  }

  /**
   * The multimap of tectonic settings to source logic trees in this model.
   */
  public Multimap<TectonicSetting, SourceTree> trees() {
    return treesMap;
  }

  /**
   * The mappings of NEHRP site classes to V<sub><i>S</i>30</sub> values for
   * this model.
   */
  public Map<NehrpSiteClass, Double> siteClasses() {
    return siteClassMap;
  }

  /**
   * The site data used with this model.
   */
  public SiteData siteData() {
    return siteData;
  }

  /**
   * Returns a filter that will skip source sets in tectonic settings not
   * applicable to the supplied location.
   *
   * @param location of interest
   */
  public Predicate<RuptureSet<? extends Source>> settingFilter(
      CalcConfig config,
      Location location) {
    return new TectonicSettingFilter(this, config, location);
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder("HazardModel: ");
    sb.append(name).append(NEWLINE);
    for (SourceTree tree : this) {
      sb.append(padEnd(tree.type().toString(), 12, ' '))
          .append(tree.id())
          .append(" ")
          .append(Strings.padEnd(tree.name(), 56, ' '))
          .append("size: ")
          .append(tree.size())
          .append(NEWLINE)
          .append(tree)
          .append(NEWLINE);
    }
    return sb.toString();
  }

  /**
   * Returns a new {@link Builder}.
   */
  static Builder builder() {
    return new Builder();
  }

  static class Builder {

    static final String ID = "HazardModel.Builder";
    boolean built = false;

    private Info info;
    CalcConfig config; // also accessible during model initialization
    private SetMultimap<SourceType, RuptureSet<? extends Source>> ruptureSetMap;

    private ImmutableSetMultimap.Builder<SourceType, RuptureSet<? extends Source>> ruptureMapBuilder;

    private Set<TectonicSetting> settings = EnumSet.noneOf(TectonicSetting.class);
    private Map<TectonicSetting, MapRegion> mapRegionMap = new EnumMap<>(TectonicSetting.class);
    private SiteData siteData;

    private ImmutableListMultimap.Builder<TectonicSetting, SourceTree> treesMapBuilder;
    private Multimap<TectonicSetting, SourceTree> treesMap;
    private Map<Integer, SourceTree> idMap;

    // Not used but will be (uncertainty)
    private List<SourceTree> trees = new ArrayList<>();

    private Builder() {
      ruptureMapBuilder = ImmutableSetMultimap.builder();
      treesMapBuilder = ImmutableListMultimap.builder();
    }

    Builder info(Info info) {
      this.info = checkNotNull(info);
      return this;
    }

    Builder config(CalcConfig config) {
      this.config = checkNotNull(config);
      return this;
    }

    Builder siteData(SiteData siteData) {
      this.siteData = checkNotNull(siteData);
      return this;
    }

    Builder addMapRegion(TectonicSetting setting, MapRegion region) {
      mapRegionMap.put(
          checkNotNull(setting),
          checkNotNull(region));
      return this;
    }

    Builder addSourceTree(SourceTree tree) {
      trees.add(tree);
      settings.add(tree.setting());
      treesMapBuilder.put(tree.setting(), tree);
      addRuptureSetsFromTree(tree);
      return this;
    }

    private Builder addRuptureSet(RuptureSet<? extends Source> source) {
      ruptureMapBuilder.put(source.type(), source);
      return this;
    }

    private void validateState(String mssgID) {
      checkState(!built, "This %s instance as already been used", mssgID);
      checkState(info != null, "%s info not set", mssgID);
      checkState(config != null, "%s config not set", mssgID);
      checkState(ruptureSetMap.size() > 0, "%s has no source sets", mssgID);
      checkState(settings.size() > 0, "%s has no tectonic settings", mssgID);
      checkState(siteData != null, "%s site data not set", mssgID);
      checkState(treesMap.size() > 0, "%s has no source trees", mssgID);
      // mapRegions are purely optional
      built = true;
    }

    HazardModel build() {
      ruptureSetMap = ruptureMapBuilder.build();
      treesMap = treesMapBuilder.build();
      idMap = treesMap.values().stream()
          .collect(toUnmodifiableMap(
              SourceTree::id,
              Function.identity()));
      validateState(ID);
      return new HazardModel(this);
    }

    private void addRuptureSetsFromTree(SourceTree tree) {
      tree.forEach(branch -> addRuptureSet(branch.value()));
    }
  }

  static final class Info {

    final String name;
    @SerializedName("site-class-vs30")
    final Map<NehrpSiteClass, Double> siteClassMap;

    Info(String name, Map<NehrpSiteClass, Double> siteClassMap) {
      this.name = name;
      this.siteClassMap = siteClassMap;
    }
  }

  /*
   * A source filter that skips those in tectonic settings if the supplied
   * location is outside the map-region for the setting.
   */
  static class TectonicSettingFilter implements Predicate<RuptureSet<? extends Source>> {

    private Set<TectonicSetting> settings;

    TectonicSettingFilter(HazardModel model, CalcConfig config, Location loc) {
      // start with all settings included in model or specified in config
      settings = config.hazard.tectonicSettings.isEmpty()
          ? EnumSet.copyOf(model.settings())
          : EnumSet.copyOf(config.hazard.tectonicSettings);
      for (Entry<TectonicSetting, MapRegion> entry : model.mapRegionMap.entrySet()) {
        // remove those which are not applicable for the location
        if (!entry.getValue().contains(loc)) {
          settings.remove(entry.getKey());
        }
      }
    }

    @Override
    public boolean test(RuptureSet<? extends Source> ruptures) {
      return settings.contains(ruptures.setting());
    }
  }

  static final class MapRegion {

    final Feature feature;
    final Area area;

    MapRegion(Feature feature) {
      this.feature = feature;
      LocationList areaLocs = GeoJson.expandRectangularArea(feature.asPolygonBorder());
      this.area = new Area(Locations.toPath(areaLocs));
    }

    boolean contains(Location loc) {
      return area.contains(loc.longitude, loc.latitude);
    }

    @Override
    public String toString() {
      return feature.toJson();
    }
  }

}
