package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MFD_TREE;
import static java.util.stream.Collectors.collectingAndThen;
import static java.util.stream.Collectors.toList;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.function.Predicate;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.fault.surface.ApproxGriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.GriddedSurface;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.GutenbergRichter;
import gov.usgs.earthquake.nshmp.model.MfdConfig.AleatoryProperties;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Subduction interface rupture set.
 *
 * @author U.S. Geological Survey
 */
public class InterfaceRuptureSet extends AbstractRuptureSet<InterfaceSource> {

  /* May be original interface section or stitched. */
  final SourceFeature.Interface feature;
  final ModelData data;

  final LogicTree<Mfd> mfdTree;
  final Mfd mfdTotal;

  final List<Integer> sectionIds;// reference: actually needed?
  final GriddedSurface surface;
  private final List<InterfaceSource> source;

  InterfaceRuptureSet(Builder builder) {
    super(builder);
    this.feature = builder.feature;
    this.data = builder.data;

    this.mfdTree = builder.mfdTree;
    this.mfdTotal = builder.mfdTotal;

    this.sectionIds = builder.sectionIds;
    this.surface = builder.surface;
    this.source = List.of(createSource());
  }

  private InterfaceSource createSource() {
    return new InterfaceSource.Builder()
        .feature(feature)
        .config(data.sourceConfig().asInterface())
        .mfdTree(mfdTree)
        .mfdTotal(mfdTotal)
        .build();
  }

  @Override
  public int size() {
    return source.size(); // TODO should this be # ruptures
  }

  @Override
  public LogicTree<Mfd> mfdTree() {
    return mfdTree;
  }

  @Override
  public Iterator<InterfaceSource> iterator() {
    return source.iterator();
  }

  @Override
  public Predicate<InterfaceSource> distanceFilter(
      Location loc,
      double distance) {

    return new Predicate<InterfaceSource>() {
      private Predicate<Location> filter = Locations.distanceFilter(loc, distance);

      @Override
      public boolean test(InterfaceSource source) {
        return filter.test(source.upperTrace.first()) ||
            filter.test(source.upperTrace.last()) ||
            filter.test(source.lowerTrace.first()) ||
            filter.test(source.lowerTrace.last());
      }
    };
  }

  /**
   * The closest point on the fault trace, relative to the supplied site
   * {@code Location}.
   */
  @Override
  public Location location(Location site) {
    // update to check all traces
    return Locations.closestPoint(site,
        LocationList.builder()
            .addAll(feature.traces.get(0))
            .addAll(feature.traces.get(1))
            .build());
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder */
  static class Builder extends AbstractRuptureSet.Builder {

    private SourceFeature.Interface feature;
    private ModelData data;
    private List<Integer> sectionIds;

    private LogicTree<Mfd.Properties> mfdPropsTree;

    /* created on build */
    private LogicTree<Mfd> mfdTree;
    private Mfd mfdTotal;
    private GriddedSurface surface;

    private Builder() {
      super(SourceType.INTERFACE);
    }

    @Override
    Builder name(String name) {
      super.name(name);
      return this;
    }

    @Override
    Builder id(int id) {
      super.id(id);
      return this;
    }

    Builder sections(List<Integer> sections) {
      checkArgument(sections.size() >= 1);
      this.sectionIds = List.copyOf(sections);
      return this;
    }

    /* Set the MFD tree. */
    Builder mfdTree(LogicTree<Mfd.Properties> mfdTree) {
      this.mfdPropsTree = mfdTree;
      return this;
    }

    /* Set MFD helper data. */
    Builder modelData(ModelData data) {
      this.data = data;
      super.setting(data.tectonicSetting());
      super.gmms(data.gmms());
      return this;
    }

    InterfaceRuptureSet build() {
      validateAndInit("InterfaceRuptureSet.Builder");
      return new InterfaceRuptureSet(this);
    }

    private void validateAndInit(String label) {
      validateState(label);

      checkNotNull(data, "%s model data", label);
      checkNotNull(sectionIds, "%s section id list", label);
      checkNotNull(mfdPropsTree, "%s MFD logic tree", label);

      // checking props are always the same;
      // don't think this is always true at least for faults
      checkState(mfdPropsTree.stream()
          .map(Branch::value)
          .map(Mfd.Properties::type)
          .distinct().limit(2).count() <= 1);

      mfdTree = buildMfdTree(data, mfdPropsTree);
      mfdTotal = ModelTrees.reduceMfdTree(mfdTree);

      feature = createFeature();

      SourceConfig.Interface config = data.sourceConfig().asInterface();
      surface = new ApproxGriddedSurface(
          feature.traces.get(0),
          feature.traces.get(1),
          config.surfaceSpacing);
    }

    private SourceFeature.Interface createFeature() {
      Map<Integer, SourceFeature.Interface> featureMap =
          data.interfaceFeatureMap().orElseThrow();
      checkState(featureMap.keySet().containsAll(sectionIds));

      Feature f = joinFeatures(sectionIds.stream()
          .map(featureMap::get)
          .map(feature -> feature.source)
          .collect(toList()));
      return new SourceFeature.Interface(f, false);
    }

    /*
     * Join the traces in the supplied features into a new Feature and set name
     * and id of rupture-set. At present, interface sections have no geologic
     * properties such that we need to pick up properties from the first section
     * or override them with properties specified for the rupture-set. Moreover,
     * all current NSHM interface rupture-set sections have matching endpoints
     * and do not require coordinate/trace overrides.
     */
    private Feature joinFeatures(List<Feature> features) {
      Feature.Builder feature = (features.size() == 1)
          ? Feature.copyOf(features.get(0))
          : joinInterfaceSections(features);
      feature.id(id).properties(Map.of("name", name));
      return feature.build();
    }

    /*
     * Join the traces of the supplied sections in order, removing duplicate
     * points at the ends of sections.
     */
    private static Feature.Builder joinInterfaceSections(List<Feature> features) {
      int traceCount = features.get(0).asMultiLineString().size();
      List<LocationList> joinedTraces = new ArrayList<>();
      for (int i = 0; i < traceCount; i++) {
        LocationList trace = joinTracesAtIndex(features, i);
        joinedTraces.add(trace);
      }
      return Feature.multiLineString(joinedTraces);
    }

    /*
     * Join traces nested in MultiLineString geometries, throwing
     * IndexOutOfBoundsException if trace counts are not the same.
     */
    private static LocationList joinTracesAtIndex(List<Feature> features, int index) {
      return features.stream()
          .map(Feature::asMultiLineString)
          .map(list -> list.get(index))
          .flatMap(LocationList::stream)
          .distinct()
          .collect(collectingAndThen(
              toList(),
              LocationList::copyOf));
    }
  }

  private static LogicTree<Mfd> buildMfdTree(
      ModelData mfdData,
      LogicTree<Mfd.Properties> mfdPropertiesTree) {

    LogicTree.Builder<Mfd> mfdTree = LogicTree.builder(MFD_TREE);

    for (Branch<Mfd.Properties> branch : mfdPropertiesTree) {
      switch (branch.value().type()) {
        case SINGLE:
          createSingleMfds(mfdTree, mfdData, branch);
          break;
        case GR:
          createGrMfds(mfdTree, mfdData, branch);
          break;
        default:
          throw new UnsupportedOperationException(
              branch.value().type() + " MFD not supported");
      }
    }
    return mfdTree.build();
  }

  /* Interface GR */
  private static void createGrMfds(
      LogicTree.Builder<Mfd> mfdTree,
      ModelData mfdData,
      Branch<Mfd.Properties> branch) {

    MfdConfig mfdConfig = mfdData.mfdConfig().orElseThrow();
    Mfd.Properties.GutenbergRichter gr = branch.value().getAsGr();

    if (mfdConfig.epistemicTree.isEmpty()) {
      Mfd mfd = gr.toBuilder().build();
      mfdTree.addBranch(branch.id(), mfd, branch.weight());
      return;
    }

    // epistemic branches - note this is ignoring mMin 6.5
    double moRate = FaultRuptureSet.grMomentRate(gr);
    for (Branch<Double> epiBranch : mfdConfig.epistemicTree.orElseThrow()) {
      // see FaultRuptureSet; consider consolidating
      double mMaxEpi = gr.mMax() + epiBranch.value();
      double weightEpi = branch.weight() * epiBranch.weight();
      String id = ModelTrees.combineBranchIds(branch.id(), epiBranch.id());
      Mfd mfd = GutenbergRichter.copyOfWithMax(gr, mMaxEpi)
          .toBuilder()
          .scaleToMomentRate(moRate)
          .build();
      mfdTree.addBranch(id, mfd, weightEpi);
    }
  }

  /* Interface SINGLE */
  private static void createSingleMfds(
      LogicTree.Builder<Mfd> mfdTree,
      ModelData mfdData,
      Branch<Mfd.Properties> branch) {

    MfdConfig mfdConfig = mfdData.mfdConfig().orElseThrow();
    Mfd.Properties.Single props = branch.value().getAsSingle();

    if (mfdConfig.aleatoryProperties.isEmpty()) {
      /* rates deserilized from json; may not always hold */
      Mfd mfd = props.toBuilder().build();
      mfdTree.addBranch(branch.id(), mfd, branch.weight());
      return;
    }

    // aleatory branches - note this is ignoring mMin 6.5
    // any epistemic variation is set in MFD logic tree
    AleatoryProperties aleaProps = mfdConfig.aleatoryProperties.orElseThrow();
    double moRate = Earthquakes.magToMoment(props.magnitude()) * props.rate();
    Mfd mfd = Mfd.newGaussianBuilder(props.magnitude(),
        aleaProps.size, aleaProps.σ, aleaProps.nσ)
        .scaleToMomentRate(moRate)
        .build();
    mfdTree.addBranch(branch.id(), mfd, branch.weight());
  }

}
