package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.BiFunction;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.stream.Stream;

import gov.usgs.earthquake.nshmp.calc.CalcConfig;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Factory class for processing NSHM model files and directories. Class provides
 * methods to check for, process, or read all file types and directory contents
 * in a NSHM. Most read*() methods delegate file parsing to Deserialize and
 * return an Optional that may be empty if the requested file is not found. All
 * such methods take a parent directory as an argument.
 *
 * @author U.S. Geological Survey
 */
class ModelFiles {

  static final String SOURCE_TREE = "source-tree.json";
  static final String TREE_INFO = "tree-info.json";
  static final String CALC_CONFIG = CalcConfig.FILE_NAME;
  static final String MODEL_INFO = "model-info.json";
  static final String MAP_REGION = "map-region.geojson";
  static final String GMM_TREE = "gmm-tree.json";
  static final String GMM_CONFIG = "gmm-config.json";
  static final String FAULT_CONFIG = "fault-config.json";
  static final String GRID_CONFIG = "grid-config.json";
  static final String INTERFACE_CONFIG = "interface-config.json";
  static final String DECOLLEMENT_CONFIG = "decollement-config.json";
  static final String SLAB_CONFIG = "slab-config.json";
  static final String ZONE_CONFIG = "zone-config.json";
  static final String MFD_CONFIG = "mfd-config.json";
  static final String MFD_MAP = "mfd-map.json";
  static final String RATE_TREE = "rate-tree.json";
  static final String RATES_TREE = "rates-tree.json";
  static final String CLUSTER_SET = "cluster-set.json";
  static final String RUPTURE_SET = "rupture-set.json";
  static final String RUPTURE_SETS = "rupture-sets.json";
  static final String SECTIONS = "sections.geojson";
  static final String RUPTURES = "ruptures.csv";
  static final String FEATURES_DIR = "features";
  static final String GRID_DATA_DIR = "grid-data";
  static final String SITE_DATA_DIR = "site-data";
  static final String SOURCE_GROUP = "source-group.json";

  /* Check for source-tree or source-group logic tree. */
  static Optional<Path> checkSourceTree(Path dir) {
    return read(dir, SOURCE_TREE, Function.identity())
        .or(() -> read(dir, SOURCE_GROUP, Function.identity()));
  }

  /* Check for grid data directory. */
  static Optional<Path> checkGridDataDirectory(Path dir) {
    return read(dir, GRID_DATA_DIR, Function.identity());
  }

  /* Check for fault system ruputure data file. */
  static Optional<Path> checkRupturesFile(Path dir) {
    return read(dir, RUPTURES, Function.identity());
  }

  /* Read config that may accompany model and merge with defaults. */
  static CalcConfig readCalcConfig(Path dir) {
    try {
      CalcConfig.Builder config = CalcConfig.defaults();
      Path modelConfig = dir.resolve(CALC_CONFIG);
      if (Files.exists(modelConfig)) {
        config.extend(CalcConfig.from(modelConfig));
      }
      return config.build();
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  /* Site data directory. */
  static SiteData readSiteData(Path dir) {
    Optional<SiteData> siteData = read(dir, SITE_DATA_DIR, Deserialize::siteData);
    return siteData.orElse(SiteData.EMPTY);
  }

  /* Process GeoJSON files in supplied directory. */
  static void readGeoJsons(
      Path dir,
      ModelData data,
      BiFunction<Path, ModelData, SourceTree> mapper,
      Consumer<SourceTree> consumer) {

    try (Stream<Path> paths = Files.list(dir)) {
      paths.filter(ModelFiles::isGeoJson)
          .map(path -> mapper.apply(path, data))
          .forEach(consumer::accept);
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  static boolean isGeoJson(Path path) {
    String fileName = checkNotNull(path.getFileName()).toString();
    return fileName.endsWith(".geojson") && !fileName.equals(MAP_REGION);
  }

  /* Process nested directories in supplied directory. */
  static void readNestedDirectories(
      Path dir,
      ModelData data,
      BiFunction<Path, ModelData, List<SourceTree>> mapper,
      Consumer<List<SourceTree>> consumer) {

    try (Stream<Path> paths = Files.list(dir)) {
      paths.filter(ModelFiles::isValidDirectory)
          .map(path -> mapper.apply(path, data))
          .forEach(consumer::accept);
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  static boolean isValidDirectory(Path path) {
    try {
      String filename = checkNotNull(path.getFileName()).toString();
      return Files.isDirectory(path) &&
          !Files.isHidden(path) &&
          !filename.startsWith("~") &&
          !filename.equals(GRID_DATA_DIR) &&
          !filename.equals(FEATURES_DIR);
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  /* Model info. */
  static Optional<HazardModel.Info> readModelInfo(Path dir) {
    return read(dir, MODEL_INFO, Deserialize::modelInfo);
  }

  /* Map Region. */
  static Optional<HazardModel.MapRegion> readMapRegion(Path dir) {
    return read(dir, MAP_REGION, Deserialize::mapRegion);
  }

  /* A GMM tree and adjacent config, relative to parent. */
  static Optional<GmmSet> readGmms(Path dir) {
    Path gmmPath = dir.resolve(GMM_TREE);
    if (Files.exists(gmmPath)) {
      LogicTree<Gmm> tree = Deserialize.gmmTree(gmmPath);
      GmmConfig config = Deserialize.gmmConfig(gmmPath.resolveSibling(GMM_CONFIG));
      GmmSet gmms = new GmmSet.Builder().primaryModels(tree, config).build();
      return Optional.of(gmms);
    }
    return Optional.empty();
  }

  /* A source-tree or source-group logic tree. */
  static Optional<LogicTree<Path>> readSourceTree(Path dir) {
    return read(dir, SOURCE_TREE, Deserialize::sourceTree)
        .or(() -> read(dir, SOURCE_GROUP, Deserialize::sourceGroup));
  }

  /* Source tree info. */
  static Optional<SourceTree.Info> readTreeInfo(Path dir) {
    return read(dir, TREE_INFO, Deserialize::treeInfo);
  }

  /* Finite fault config. */
  static Optional<SourceConfig> readFaultConfig(Path dir) {
    return read(dir, FAULT_CONFIG, Deserialize::faultConfig);
  }

  /* Grid config. */
  static Optional<SourceConfig> readGridConfig(Path dir) {
    return read(dir, GRID_CONFIG, Deserialize::gridConfig);
  }

  /* Decollement config (same as interface). */
  static Optional<SourceConfig> readDecollementConfig(Path dir) {
    return read(dir, DECOLLEMENT_CONFIG, Deserialize::interfaceConfig);
  }

  /* Interface config (same as fault). */
  static Optional<SourceConfig> readInterfaceConfig(Path dir) {
    return read(dir, INTERFACE_CONFIG, Deserialize::interfaceConfig);
  }

  /* Intraslab config (same as grid). */
  static Optional<SourceConfig.Grid> readSlabConfig(Path dir) {
    return read(dir, SLAB_CONFIG, Deserialize::gridConfig);
  }

  /* Zone config (same as grid). */
  static Optional<SourceConfig.Grid> readZoneConfig(Path dir) {
    return read(dir, ZONE_CONFIG, Deserialize::gridConfig);
  }

  /* Fault section directory. */
  static Optional<Map<Integer, SourceFeature.NshmFault>> readFaultFeatures(Path dir) {
    return read(dir, FEATURES_DIR, Deserialize::faultFeatures);
  }

  /* Grid feature directory. */
  static Optional<Map<Integer, SourceFeature.Grid>> readGridFeatures(Path dir) {
    return read(dir, FEATURES_DIR, Deserialize::gridFeatures);
  }

  /* Interface section directory. */
  static Optional<Map<Integer, SourceFeature.Interface>> readInterfaceFeatures(Path dir) {
    return read(dir, FEATURES_DIR, Deserialize::interfaceFeatures);
  }

  /* Fault system section file. */
  static Optional<Map<Integer, SourceFeature.SystemSection>> readSystemFeatures(Path dir) {
    return read(dir, SECTIONS, Deserialize::systemFeatures);
  }

  /* Interface rupture-set. */
  static Optional<InterfaceRuptureSet> readInterfaceRuptureSet(Path dir, ModelData data) {
    return read(dir, RUPTURE_SET, data, Deserialize::interfaceRuptureSet);
  }

  /* Fault rupture-set. */
  static Optional<FaultRuptureSet> readFaultRuptureSet(Path dir, ModelData data) {
    return read(dir, RUPTURE_SET, data, Deserialize::faultRuptureSet);
  }

  /* Fault cluster-set data. */
  static Optional<ClusterRuptureSet.Data> readFaultClusterSet(Path dir, ModelData data) {
    return read(dir, CLUSTER_SET, data, Deserialize::clusterRuptureSetData);
  }

  /* Fault system rupture-set. */
  static Optional<SystemRuptureSet> readSystemRuptureSet(Path dir, ModelData data) {
    return read(dir, RUPTURE_SET, data, Deserialize::systemRuptureSet);
  }

  /* Grid rupture-set. */
  static Optional<List<GridRuptureSet>> readGridRuptureSets(Path dir, ModelData data) {
    return read(dir, RUPTURE_SETS, data, Deserialize::gridRuptureSets);
  }

  /* Slab rupture-set. */
  static Optional<List<SlabRuptureSet>> readSlabRuptureSets(Path dir, ModelData data) {
    return read(dir, RUPTURE_SETS, data, Deserialize::slabRuptureSets);
  }

  /* Map of MFD property logic trees. */
  static Optional<Map<String, LogicTree<Mfd.Properties>>> readMfdMap(Path dir) {
    return read(dir, MFD_MAP, Deserialize::mfdMap);
  }

  /* MFD config. */
  static Optional<MfdConfig> readMfdConfig(Path dir) {
    return read(dir, MFD_CONFIG, Deserialize::mfdConfig);
  }

  /* Rate-tree containing values. */
  static Optional<LogicTree<Double>> readRateTree(Path dir) {
    return read(dir, RATE_TREE, Deserialize::rateTree);
  }

  /* Grid rate-tree containing rate file names. */
  @Deprecated
  static Optional<LogicTree<Path>> readGridRatesTree(Path dir) {
    return read(dir, RATES_TREE, Deserialize::gridRatesTree);
  }

  /* Rate-tree containing deformation model ids. */
  static Optional<LogicTree<String>> readFaultRateTree(Path dir) {
    return read(dir, RATE_TREE, Deserialize::slipRateTree);
  }

  /*
   * Resolves the dir/file path. If it exists, method returns an optional
   * describing the result of applying a mapper to the path. To return the
   * resolved path, supply Function.identity() as a mapper.
   */
  private static <R> Optional<R> read(
      Path dir,
      String file,
      Function<Path, R> mapper) {

    Path path = dir.resolve(file);
    if (Files.exists(path)) {
      return Optional.of(mapper.apply(path));
    }
    return Optional.empty();
  }

  /* Same as above, but with a BiFunction mapper. */
  private static <R> Optional<R> read(
      Path dir,
      String file,
      ModelData data,
      BiFunction<Path, ModelData, R> mapper) {

    Path path = dir.resolve(file);
    if (Files.exists(path)) {
      return Optional.of(mapper.apply(path, data));
    }
    return Optional.empty();
  }

}
