package gov.usgs.earthquake.nshmp.model;

import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.mapping;
import static java.util.stream.Collectors.toList;

import java.nio.file.Path;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.stream.Collectors;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Factory class for querying source models.
 *
 * @author U.S. Geological Survey
 */
public class Models {

  static final Gson GSON = new GsonBuilder()
      .setPrettyPrinting()
      .create();

  public static void main(String[] args) {
    HazardModel model = ModelLoader.load(Path.of("../nshm-conus"));
    // HazardModel model = ModelLoader.load(Path.of("../nshm-hawaii"));

    // JSON of all tree info for a model
    System.out.println(GSON.toJson(trees(model)));
    System.out.println();

    // Cascadia interface tree
    SourceTree tree = model.tree(3199).orElseThrow();

    // see what the list of nodes for each branch looks like
    // we don't use this but it's good for illustrating
    // internal structure of a SourceTree
    tree.nodes().stream().forEach(System.out::println);
    System.out.println();

    // Serialized form of Object returned by tree() in this class
    System.out.println(GSON.toJson(tree(model, 3199)));
    System.out.println();

    System.out.println(GSON.toJson(tree(model, 2539)));
    System.out.println();

    // 2799 is huge wasatch tree
  }

  /**
   * Returns an object for JSON serialization with the name and ID of all source
   * logic tree groups in the supplied model organized by tectonic setting and
   * source type.
   *
   * @param model to extract logic tree data from
   */
  public static Object trees(HazardModel model) {
    return model.trees().asMap().entrySet().stream()
        .map(Models::toSettingGroup)
        .collect(toList());
  }

  /**
   * Returns an object for JSON serialization the details of a logic tree in the
   * supplied model.
   *
   * @param model to query
   * @param id of desired source tree
   * @throws NoSuchElementException if the id is not present in the model
   */
  public static Object tree(HazardModel model, int id) {
    return model.tree(id).map(Tree::new).orElseThrow();
  }

  private static SettingGroup toSettingGroup(Entry<TectonicSetting, Collection<SourceTree>> entry) {
    return new SettingGroup(
        entry.getKey(),
        toSourceGroup(entry.getValue()));
  }

  private static List<SourceGroup> toSourceGroup(Collection<SourceTree> trees) {
    return trees.stream()
        .collect(groupingBy(
            SourceTree::type,
            mapping(
                TreeInfo::new,
                Collectors.toList())))
        .entrySet().stream()
        .map(e -> new SourceGroup(e.getKey(), e.getValue()))
        .collect(toList());
  }

  static final class SettingGroup {

    final TectonicSetting setting;
    final List<SourceGroup> data;

    SettingGroup(TectonicSetting setting, List<SourceGroup> data) {
      this.setting = setting;
      this.data = data;
    }
  }

  static final class SourceGroup {

    final SourceType type;
    final List<TreeInfo> data;

    SourceGroup(SourceType type, List<TreeInfo> data) {
      this.type = type;
      this.data = data;
    }
  }

  static final class TreeInfo {

    final int id;
    final String name;

    TreeInfo(SourceTree tree) {
      this.id = tree.id();
      this.name = tree.name();
    }
  }

  static final class Tree {

    final int id;
    final String name;
    final TectonicSetting setting;
    final SourceType type;
    final List<SourceBranch> branches;

    Tree(SourceTree tree) {
      this.id = tree.id();
      this.name = tree.name();
      this.setting = tree.setting();
      this.type = tree.type();
      this.branches = tree.branches().entrySet().stream()
          .map(e -> new SourceBranch(
              e.getValue().toString(),
              tree.weights().get(e.getKey()),
              e.getKey().ruptures.mfdTree()))
          .collect(toList());
    }
  }

  static final class SourceBranch {

    final String name;
    final String path;
    final double weight;
    final List<MfdBranch> mfds;

    SourceBranch(String path, double weight, LogicTree<Mfd> mfds) {
      this.name = mfds.name();
      this.path = path;
      this.weight = weight;
      this.mfds = mfds.stream()
          .map(MfdBranch::new)
          .collect(toList());
    }
  }

  static final class MfdBranch {
    final String id;
    final double weight;
    final Mfd mfd;

    MfdBranch(Branch<Mfd> branch) {
      this.id = branch.id();
      this.weight = branch.weight();
      this.mfd = branch.value();
    }
  }
}
