package gov.usgs.earthquake.nshmp.model;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.NavigableMap;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.calc.HazardCalcs;
import gov.usgs.earthquake.nshmp.calc.InputList;
import gov.usgs.earthquake.nshmp.calc.Site;
import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureScaling;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.model.PointSource.DepthModel;

/**
 * Factory class for generating lists of {@code GmmInput}s for point sources.
 *
 * @author U.S. Geological Survey
 */
public class PointSources {

  /**
   * Using the supplied {@code Site} and the standard data that is used to build
   * point sources, create and return an {@code InputList}.
   *
   * @param site of interest
   * @param loc
   * @param mfd
   * @param mechWtMap
   * @param rupScaling
   * @param magDepthMap
   * @param maxDepth
   */
  public static InputList finiteInputs(
      Site site,
      SourceType sourceType,
      Location loc,
      Mfd mfd,
      Map<FocalMech, Double> mechWtMap,
      RuptureScaling rupScaling,
      NavigableMap<Double, Map<Double, Double>> magDepthMap,
      double maxDepth) {

    Source source = finiteSource(
        sourceType,
        loc,
        mfd,
        mechWtMap,
        rupScaling,
        magDepthMap,
        maxDepth);

    return HazardCalcs.sourceToInputs(site).apply(source);
  }

  /**
   * Using the supplied {@code Site}s and the standard data that is used to
   * build point sources, create and return a {@code List<InputList>}.
   *
   * @param sites
   * @param loc
   * @param mfd
   * @param mechWtMap
   * @param rupScaling
   * @param magDepthMap
   * @param maxDepth
   */
  public static List<InputList> finiteInputs(
      List<Site> sites,
      SourceType sourceType,
      Location loc,
      Mfd mfd,
      Map<FocalMech, Double> mechWtMap,
      RuptureScaling rupScaling,
      NavigableMap<Double, Map<Double, Double>> magDepthMap,
      double maxDepth) {

    Source source = finiteSource(
        sourceType,
        loc,
        mfd,
        mechWtMap,
        rupScaling,
        magDepthMap,
        maxDepth);

    List<InputList> inputsList = new ArrayList<>();
    for (Site site : sites) {
      InputList inputs = HazardCalcs.sourceToInputs(site).apply(source);
      inputsList.add(inputs);
    }
    return inputsList;
  }

  /**
   * Using the supplied {@code Site}s and a GridRuptureSet, from which the
   * standard data that is used to build point sources is derived, create and
   * return a {@code List<InputList>}.
   *
   */
  public static List<InputList> finiteInputs(
      List<Site> sites,
      SourceType sourceType,
      Location loc,
      Mfd mfd,
      Map<FocalMech, Double> mechWtMap,
      GridRuptureSet grid) {

    Source source = pointSource(
        sourceType,
        PointSourceType.FINITE,
        loc,
        mfd,
        mechWtMap,
        grid.rupScaling,
        grid.depthModel);

    return finiteInputs(sites, source);
  }

  private static List<InputList> finiteInputs(List<Site> sites, Source source) {

    List<InputList> inputsList = new ArrayList<>();
    for (Site site : sites) {
      InputList inputs = HazardCalcs.sourceToInputs(site).apply(source);
      inputsList.add(inputs);
    }
    return inputsList;
  }

  private static PointSource finiteSource(
      SourceType sourceType,
      Location loc,
      Mfd mfd,
      Map<FocalMech, Double> mechWtMap,
      RuptureScaling rupScaling,
      NavigableMap<Double, Map<Double, Double>> magDepthMap,
      double maxDepth) {

    DepthModel depthModel = DepthModel.create(
        magDepthMap,
        mfd.data().xValues().boxed().collect(Collectors.toList()),
        maxDepth);
    return new PointSourceFinite(sourceType, loc, mfd, mechWtMap, rupScaling, depthModel);
  }

  public static PointSource pointSource(
      SourceType sourceType,
      PointSourceType pointType,
      Location loc,
      Mfd mfd,
      Map<FocalMech, Double> mechWtMap,
      RuptureScaling rupScaling,
      DepthModel depthModel) {

    switch (pointType) {
      case POINT:
        return new PointSource(sourceType, loc, mfd, mechWtMap, rupScaling, depthModel);
      case FINITE:
        return new PointSourceFinite(sourceType, loc, mfd, mechWtMap, rupScaling, depthModel);
      default:
        throw new UnsupportedOperationException("FIXED_STRIKE point sources not supported");
    }

  }

}
