package gov.usgs.earthquake.nshmp.model;

import static gov.usgs.earthquake.nshmp.model.Deserialize.ID;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MODEL;
import static gov.usgs.earthquake.nshmp.model.Deserialize.NAME;
import static gov.usgs.earthquake.nshmp.model.Deserialize.SPACING;
import static java.util.stream.Collectors.toUnmodifiableMap;
import static java.util.stream.Collectors.toUnmodifiableSet;

import java.awt.geom.Area;
import java.math.BigDecimal;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.OptionalDouble;
import java.util.Set;
import java.util.stream.Stream;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.Properties;

/**
 * Site data wrapper class. This class provides one method {@code get(Location)}
 * that returns the superset of all site data parameters required by the current
 * suite of supported NSHMs. Currently this only includes the {@code z1.0} and
 * {@code z2.5} basin depth values.
 *
 * This class assumes that if a location 'hit test' succeeds on an area for
 * which there are site data, then non null values will exist for the location.
 *
 * @author U.S. Geological Survey
 */
public class SiteData {

  /*
   * Developer notes: This class will soon be expanded to support a variety of
   * site data such as sediment thickness. Web services should map into the
   * SiteData.Values class. We're only working with the two basin depth values
   * at present, but ultimately we'll probably want
   *
   * Consider having this class deliver defaults.
   */

  /** Empty site data instance. */
  public static final SiteData EMPTY = new SiteData(null) {
    @Override
    public Values get(Location loc) {
      return Values.EMPTY;
    }
  };

  private static Path BASIN_FILE = Path.of("basin/basins.geojson");

  private Set<Basin> basins;

  private SiteData(Set<Basin> basins) {
    this.basins = basins;
  }

  static SiteData create(Path dir) {
    Set<Basin> basins = Basin.init(dir);
    return new SiteData(basins);
  }

  /**
   * Return the site data values for the specified location.
   *
   * @param location of interest
   */
  public Values get(Location location) {
    Values.Builder builder = new Values.Builder();
    for (Basin basin : basins) {
      if (basin.contains(location)) {
        builder.basinValues(basin.get(location));
        break;
      }
    }
    return builder.build();
  }

  /** Optional site data values associated with a location. */
  public static class Values {

    private static final Values EMPTY = new Builder().build();

    /** Depth to the shear-wave velocity horizon of 1.0 km/s, in km. */
    public final OptionalDouble z1p0;

    /** Depth to the shear-wave velocity horizon of 2.5 km/s, in km. */
    public final OptionalDouble z2p5;

    private Values(Builder builder) {
      this.z1p0 = builder.z1p0;
      this.z2p5 = builder.z2p5;
    }

    private static class Builder {

      OptionalDouble z1p0 = OptionalDouble.empty();
      OptionalDouble z2p5 = OptionalDouble.empty();

      Builder basinValues(Basin.Values values) {
        this.z1p0 = OptionalDouble.of(values.z1p0);
        this.z2p5 = OptionalDouble.of(values.z2p5);
        return this;
      }

      Values build() {
        return new Values(this);
      }
    }
  }

  private static Location toLocation(Record record) {
    return Location.create(
        record.getDouble("lon"),
        record.getDouble("lat"));
  }

  /* Container for a basin region. */
  static final class Basin {

    final String name;
    final String id;
    final String model;
    final int scale;
    final Feature feature;
    final Area area;
    final Map<Location, Values> dataMap;

    private Basin(Path file, Feature feature) {
      Properties properties = feature.properties();
      this.name = properties.getString(NAME).orElseThrow();
      this.id = properties.getString(ID).orElseThrow();
      this.model = properties.getString(MODEL).orElseThrow();
      double spacing = properties.getDouble(SPACING).orElseThrow();
      this.scale = BigDecimal.valueOf(spacing).scale();
      this.feature = feature;
      this.area = new Area(Locations.toPath(feature.asPolygonBorder()));
      Path dataFile = file.resolveSibling(id + ".csv");
      this.dataMap = loadData(dataFile);
    }

    /* Return whether the basin contains the supplied location. */
    boolean contains(Location loc) {
      return area.contains(loc.longitude, loc.latitude);
    }

    Values get(Location loc) {
      Location roundedLoc = Location.create(
          Maths.round(loc.longitude, scale),
          Maths.round(loc.latitude, scale));
      return dataMap.get(roundedLoc);
    }

    static Set<Basin> init(Path dir) {
      Path file = dir.resolve(BASIN_FILE);
      if (Files.exists(dir)) {
        FeatureCollection fc = GeoJson.from(file).toFeatureCollection();
        Set<Basin> basins = fc.features().stream()
            .map(feature -> new Basin(file, feature))
            .collect(toUnmodifiableSet());
        return basins;
      }
      return Set.of();
    }

    private Map<Location, Values> loadData(Path file) {
      DelimitedData csv = DelimitedData.comma(file);
      try (Stream<Record> stream = csv.records()) {
        return stream.collect(toUnmodifiableMap(
            SiteData::toLocation,
            Basin::toValues));
      }
    }

    private static Values toValues(Record record) {
      return new Values(
          record.getDouble("z1p0"),
          record.getDouble("z2p5"));
    }

    static final class Values {

      final double z1p0;
      final double z2p5;

      Values(double z1p0, double z2p5) {
        this.z1p0 = z1p0;
        this.z2p5 = z2p5;
      }
    }
  }

}
