package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.GR;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.SINGLE;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MFD_TREE;
import static gov.usgs.earthquake.nshmp.model.Deserialize.RATE_TREE;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Optional;
import java.util.function.Predicate;

import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.GutenbergRichter;
import gov.usgs.earthquake.nshmp.mfd.Mfds;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;
import gov.usgs.earthquake.nshmp.tree.Trees;

/*
 * Areal source zone rupture set implementation.
 *
 * @author U.S. Geological Survey
 */
class ZoneRuptureSet extends AbstractRuptureSet<PointSource> {

  private final SourceFeature.Zone feature;
  private final ModelData data;

  private final List<Location> locations;
  private final List<Mfd> mfds;
  private final LogicTree<List<Mfd>> mfdsTree;
  private final LogicTree<Mfd> mfdTree;

  private GridRuptureSet delegate;

  /*
   * Developer notes:
   *
   * Move to setting rate-tree; if single value "rate" present, create singleton
   * rate-tree.
   *
   * Rates in rate tree will need to be distributed to grid rates that scale
   * with latitude.
   *
   * Zone sources: all GR or SINGLE, but not both
   */

  ZoneRuptureSet(Builder builder) {
    super(builder);
    this.feature = builder.feature;
    this.data = builder.data;
    this.locations = builder.locations;
    this.mfdsTree = builder.mfdsTree;
    this.mfdTree = builder.mfdTree;
    this.mfds = builder.mfds;
    // TODO handle SS weight
    this.delegate = createRuptureSet();
  }

  private GridRuptureSet createRuptureSet() {
    GridRuptureSet grid = GridLoader.createGrid(
        feature.name,
        feature.id,
        data,
        SourceType.ZONE,
        feature,
        locations,
        mfds,
        mfdsTree,
        Optional.empty(),
        data.gmms());
    return grid;
  }

  @Override
  public int size() {
    return locations.size();
  }

  @Override
  public LogicTree<Mfd> mfdTree() {
    throw new UnsupportedOperationException();
  }

  @Override
  public Predicate<PointSource> distanceFilter(Location loc, double distance) {
    return delegate.distanceFilter(loc, distance);
  }

  @Override
  public Iterator<PointSource> iterator() {
    return delegate.iterator();
  }

  @Override
  public Location location(Location site) {
    // Is this called a lot and should be precomputed??
    return Locations.centroid(locations);
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder. */
  static final class Builder extends AbstractRuptureSet.Builder {

    private SourceFeature.Zone feature;
    private ModelData data;
    private Path ratesPath;
    private LogicTree<Mfd.Properties> mfdPropertiesTree;

    private Optional<String> mfdTreeKey = Optional.empty();

    private List<Location> locations;
    private LogicTree<List<Mfd>> mfdsTree;
    private LogicTree<Mfd> mfdTree;
    private List<Mfd> mfds;

    private LogicTree<double[]> ratesTree;

    private Builder() {
      super(SourceType.ZONE);
    }

    // @Override
    // Builder name(String name) {
    // super.name(name);
    // return this;
    // }
    //
    // @Override
    // Builder id(int id) {
    // super.id(id);
    // return this;
    // }

    Builder feature(SourceFeature.Zone feature) {
      this.feature = feature;
      super.name(feature.name);
      super.id(feature.id);
      return this;
    }

    Builder modelData(ModelData data) {
      this.data = data;
      super.setting(data.tectonicSetting());
      super.gmms(data.gmms());
      return this;
    }

    // absolute path to rates
    Builder rates(Path ratesPath) {
      this.ratesPath = ratesPath;
      return this;
    }

    /* Set the MFD tree using a mfdMap key. */
    Builder mfdTree(String key) {
      this.mfdTreeKey = Optional.of(key);
      return this;
    }

    /* Set the MFD tree. */
    Builder mfdTree(LogicTree<Mfd.Properties> mfdTree) {
      checkArgument(ModelTrees.checkTreeTypes(mfdTree, GR) ||
          ModelTrees.checkTreeTypes(mfdTree, SINGLE),
          "Zone MFDs should all be of the same type");
      this.mfdPropertiesTree = mfdTree;
      return this;
    }

    private void validateAndInit(String label) {
      validateState(label);
      checkNotNull(data, "%s model data", label);
      checkNotNull(feature, "%s feature", label);
      checkNotNull(ratesPath, "%s rates file path", label);
      if (mfdTreeKey.isPresent()) {
        String key = mfdTreeKey.orElseThrow();
        mfdPropertiesTree = data.mfdMap().orElseThrow().get(key);
      }
      checkNotNull(mfdPropertiesTree, "%s MFD tree not set", label);
      initLocationAndRateLists();
      mfds = createTotalMfds();
      mfdsTree = createMfdsTree();
      mfdTree = Trees.transform(mfdsTree, Mfds::combine, Optional.empty());
    }

    private void initLocationAndRateLists() {
      /*
       * this will change to use JA's rate distribution code
       */
      List<Location> locations = new ArrayList<>();
      List<Double> rates = new ArrayList<>();
      GridLoader.readZoneRateFile(ratesPath, locations, rates);
      this.locations = List.copyOf(locations);
      double[] rateArray = rates.stream()
          .mapToDouble(Double::doubleValue)
          .toArray();
      this.ratesTree = LogicTree.singleton(RATE_TREE, "rate", rateArray);
    }

    /* Create a new rupture set. */
    ZoneRuptureSet build() {
      validateAndInit("ZoneRuptureSet.Builder");
      return new ZoneRuptureSet(this);
    }

    private List<Mfd> createTotalMfds() {
      double[] rates = reduceRatesTree();
      Mfd.Properties props0 = mfdPropertiesTree.get(0).value();
      // SINGLE
      if (props0.type().equals(Mfd.Type.SINGLE)) {
        Mfd model = ModelTrees.reduceSingleMfdTree(mfdPropertiesTree);
        return Arrays.stream(rates)
            .mapToObj(r -> Mfd.Builder.from(model).scale(r).build())
            .collect(toUnmodifiableList());
      }
      // else GR
      GutenbergRichter grProps = props0.getAsGr();
      return Arrays.stream(rates)
          .map(Math::log10)
          .mapToObj(a -> createGrMfd(grProps, a))
          .collect(toUnmodifiableList());
    }

    private Mfd createGrMfd(GutenbergRichter props, double a) {
      return GutenbergRichter.copyOfWithRate(props, a)
          .toBuilder()
          .build();
    }

    private double[] reduceRatesTree() {
      double[] reduced = new double[locations.size()];
      double[] scaled = new double[locations.size()];
      for (Branch<double[]> branch : ratesTree) {
        scaled = Arrays.copyOf(branch.value(), branch.value().length);
        DoubleData.multiply(branch.weight(), scaled);
        DoubleData.add(reduced, scaled);
      }
      return reduced;
    }

    // SINGLE only
    private LogicTree<List<Mfd>> createMfdsTree() {
      Mfd.Properties props0 = mfdPropertiesTree.get(0).value();
      // Revisit: hack short circuit for WUS zones
      if (props0.type().equals(Mfd.Type.SINGLE)) {
        LogicTree.Builder<List<Mfd>> mfdsTree = LogicTree.builder(MFD_TREE);
        for (Branch<Mfd.Properties> m : mfdPropertiesTree) {
          Mfd.Builder builder = m.value().getAsSingle().toBuilder();
          for (Branch<double[]> r : ratesTree) {
            List<Mfd> mfds = Arrays.stream(r.value())
                .mapToObj(builder::scaleToCumulativeRate)
                .map(Mfd.Builder::build)
                .collect(toUnmodifiableList());
            mfdsTree.addBranch(
                m.id() + "-" + r.id(),
                mfds,
                m.weight() * r.weight());
          }
        }
        return mfdsTree.build();
      }
      // else GR
      GutenbergRichter grProps = props0.getAsGr();
      LogicTree.Builder<List<Mfd>> mfdsTree = LogicTree.builder(MFD_TREE);
      for (Branch<Mfd.Properties> m : mfdPropertiesTree) {
        for (Branch<double[]> r : ratesTree) {
          List<Mfd> mfds = Arrays.stream(r.value())
              .map(Math::log10)
              .mapToObj(a -> createGrMfd(grProps, a))
              .collect(toUnmodifiableList());
          mfdsTree.addBranch(
              m.id() + "-" + r.id(),
              mfds,
              m.weight() * r.weight());
        }
      }
      return mfdsTree.build();
    }

  }

}
