package gov.usgs.earthquake.nshmp.tree;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import gov.usgs.earthquake.nshmp.Text;

/**
 * Basic logic tree implementation.
 *
 * @author U.S. Geological Survey
 */
class RegularTree<T> extends AbstractList<Branch<T>> implements LogicTree<T> {

  /*
   * Scale used for rounding cumulative weights to eliminate double precision
   * rounding errors when summing wieghts.
   */
  static final int WEIGHT_SCALE = 12;

  private final String name;
  private final List<Branch<T>> branches;
  private final double[] cumulativeWeights;

  RegularTree(Builder<T> builder) {
    this.name = builder.name;
    this.branches = List.copyOf(builder.branches);
    this.cumulativeWeights = Arrays.copyOf(
        builder.cumulativeWeights,
        builder.cumulativeWeights.length);
  }

  @Override
  public Branch<T> get(int index) {
    return branches.get(index);
  }

  @Override
  public String name() {
    return name;
  }

  @Override
  public int size() {
    return branches.size();
  }

  @Override
  public Branch<T> sample(double probability) {
    for (int i = 0; i < cumulativeWeights.length; i++) {
      if (probability < cumulativeWeights[i]) {
        return branches.get(i);
      }
    }
    return branches.get(cumulativeWeights.length - 1);
  }

  @Override
  public List<Branch<T>> sample(double[] probabilities) {
    List<Branch<T>> samples = new ArrayList<>(probabilities.length);
    for (double probability : probabilities) {
      samples.add(sample(probability));
    }
    return samples;
  }

  @Override
  public String toString() {
    return toString(name, branches);
  }

  static <T> String toString(String name, List<Branch<T>> branches) {
    StringBuilder sb = new StringBuilder(name);
    for (int i = 0; i < branches.size(); i++) {
      switch (i) {
        case 0:
          sb.append((branches.size() == 1) ? " ─ " : " ┬ ");
          break;
        default:
          sb.append(" ".repeat(name.length()));
          sb.append((i == branches.size() - 1) ? " └ " : " ├ ");
      }
      sb.append(branches.get(i));
      sb.append((i == branches.size() - 1) ? "" : Text.NEWLINE);
    }
    return sb.toString();
  }
}
