package gov.usgs.earthquake.nshmp;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;

final class FaultsTests {

  /* radian degrees conversion tolerance */
  private static final double RAD_CONV_TOL = 1e-15;

  private static final Location P1 = Location.create(0.0, 0.0);
  private static final Location P2 = Location.create(0.0, 1.0);
  private static final Location P3 = Location.create(1.0, 1.0);
  private static final Location P4 = Location.create(1.0, 0.0);

  @ParameterizedTest
  @ValueSource(doubles = { 0.0, 0.5, 1.0 })
  final void checkAseismicityFactor_inRange(double v) {
    assertEquals(v, Faults.checkAseismicityFactor(v), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -1.0, -0.01, 1.01, 10.0 })
  final void checkAseismicityFactor_outOfRange(double v) {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkAseismicityFactor(v);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 10.0, 45.0, 90.0 })
  final void testCheckDip_inRange(double v) {
    assertEquals(v, Faults.checkDip(v), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -30.0, 9.9, 90.1, 100.0 })
  final void testCheckDip_outOfRange(double v) {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkDip(v);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { -180.0, 0.0, 180.0 })
  final void testCheckRake_inRange(double v) {
    assertEquals(v, Faults.checkRake(v), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -360.0, -180.1, 180.1, 1000.0 })
  final void testCheckRake_outOfRange(double v) {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkRake(v);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 1e-6, 4.0, 60.0 })
  final void testCheckSlipRate_inRange(double v) {
    assertEquals(v, Faults.checkSlipRate(v), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -1.0, 0.0, 60.1, 100.0 })
  final void testCheckSlipRate_outOfRange(double v) {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkSlipRate(v);
    });
  }

  @ParameterizedTest
  @ValueSource(doubles = { 0.0, 60.0, 359.9 })
  final void testCheckStrike_inRange(double v) {
    assertEquals(v, Faults.checkStrike(v), 0.0);
  }

  @ParameterizedTest
  @ValueSource(doubles = { -1.0, -0.1, 360.0, 1000.0 })
  final void testCheckStrike_outOfRange(double v) {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkStrike(v);
    });
  }

  @Test
  final void testCheckTrace() {
    LocationList expect = LocationList.of(
        Location.create(0.0, 0.0),
        Location.create(1.0, 1.0));
    assertSame(expect, Faults.checkTrace(expect));
  }

  @Test
  final void testCheckTrace_invalid() {
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkTrace(LocationList.of(Location.create(0.0, 0.0)));
    });
  }

  @Test
  final void testProjectVertical() {
    assertEquals(2.0, Faults.projectVertical(30, 1), RAD_CONV_TOL);
    assertEquals(Maths.SQRT_2, Faults.projectVertical(45, 1), RAD_CONV_TOL);
    assertEquals(2 / Math.sqrt(3), Faults.projectVertical(60, 1), RAD_CONV_TOL);
    assertEquals(1.0, Faults.projectVertical(90, 1), RAD_CONV_TOL);
  }

  @Test
  final void testHypoDepth() {
    double hypoDepthExpected = 6.0;
    assertEquals(hypoDepthExpected, Faults.hypocentralDepth(90.0, 10.0, 1.0), 0.0);
    hypoDepthExpected = 1.0 + Math.sin(60.0 * Maths.TO_RADIANS) * 10.0 / 2.0;
    assertEquals(hypoDepthExpected, Faults.hypocentralDepth(60.0, 10.0, 1.0), 0.0);
    // 3-4-5 triangle
    hypoDepthExpected = 4.0 + 1.0;
    assertEquals(hypoDepthExpected,
        Faults.hypocentralDepth(Math.asin(4.0 / 5.0) * Maths.TO_DEGREES, 10.0, 1.0), 1e-8);
  }

  @Test
  final void testDipDirection() {
    LocationList locs = LocationList.of(P1, P2, P3);

    double strikeExpect = Locations.azimuth(P1, P3);
    double strikeRadExpect = Locations.azimuthRad(P1, P3);

    assertEquals(strikeExpect, Faults.strike(locs), 0.0);
    assertEquals(strikeExpect, Faults.strike(P1, P3), 0.0);
    assertEquals(strikeRadExpect, Faults.strikeRad(locs), 0.0);
    assertEquals(strikeRadExpect, Faults.strikeRad(P1, P3), 0.0);

    double dipDirExpect = 70.0;
    double dipDirRadExpect = dipDirExpect * Maths.TO_RADIANS;

    assertEquals(dipDirExpect, Faults.dipDirection(340.0), 0.0);
    assertEquals(
        dipDirRadExpect,
        Faults.dipDirectionRad(340.0 * Maths.TO_RADIANS), RAD_CONV_TOL);

    dipDirExpect = strikeExpect + 90.0;
    dipDirRadExpect = strikeRadExpect + Maths.PI_BY_2;

    assertEquals(dipDirExpect, Faults.dipDirection(locs), 0.0);
    assertEquals(dipDirExpect, Faults.dipDirection(P1, P3), 0.0);
    assertEquals(dipDirRadExpect, Faults.dipDirectionRad(locs), 0.0);
    assertEquals(dipDirRadExpect, Faults.dipDirectionRad(P1, P3), 0.0);
  }

  @Test
  final void testCheckDipDirection() {
    // strike is due east (90)
    LocationList locs = LocationList.of(P1, P4);

    // test strike out of range
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkDipDirection(locs, -1.0);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkDipDirection(locs, 360.0);
    });

    // test compatible strike - only rejects dip direction if it is equal to
    // strike or strike + 180
    assertEquals(90.1, Faults.checkDipDirection(locs, 90.1));
    assertEquals(180.0, Faults.checkDipDirection(locs, 180));
    assertEquals(269.9, Faults.checkDipDirection(locs, 269.9));
    assertEquals(270.1, Faults.checkDipDirection(locs, 270.1));
    assertEquals(0.0, Faults.checkDipDirection(locs, 0.0));
    assertEquals(89.9, Faults.checkDipDirection(locs, 89.9));

    // test dip direction == strike or strike + 180
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkDipDirection(locs, 90.0);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Faults.checkDipDirection(locs, 270.0);
    });
  }

}
