package gov.usgs.earthquake.nshmp.calc;

import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGV;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P01;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P02;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P03;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P05;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P075;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P1;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P15;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P2;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P25;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P3;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P4;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P5;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P75;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA10P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA1P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA1P5;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA2P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA3P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA4P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA5P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA7P5;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT;
import static gov.usgs.earthquake.nshmp.model.SourceType.SLAB;
import static gov.usgs.earthquake.nshmp.model.SourceType.ZONE;
import static gov.usgs.earthquake.nshmp.model.TectonicSetting.ACTIVE_CRUST;
import static gov.usgs.earthquake.nshmp.model.TectonicSetting.SUBDUCTION;
import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.NavigableMap;
import java.util.Set;
import java.util.TreeMap;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.model.SourceType;
import gov.usgs.earthquake.nshmp.model.TectonicSetting;

class CalcConfigTests {

  static final Path RESOURCES = Path.of("src/test/resources/calc");
  static final CalcConfig DEFAULTS;
  static final CalcConfig EXTENDS;
  static final CalcConfig EXTENDS_EMPTY; // same as DEFAULTS

  static {
    try {
      DEFAULTS = CalcConfig.defaults().build();
      EXTENDS = CalcConfig.defaults()
          .extend(CalcConfig.from(RESOURCES.resolve("calc-config-extends.json")))
          .build();
      EXTENDS_EMPTY = CalcConfig.defaults()
          .extend(CalcConfig.from(RESOURCES.resolve("calc-config-extends-empty.json")))
          .build();

    } catch (IOException ioe) {
      throw new RuntimeException();
    }
  }

  static final Set<Imt> IMTS = EnumSet.of(
      PGV, PGA,
      SA0P01, SA0P02, SA0P03, SA0P05, SA0P075,
      SA0P1, SA0P15, SA0P2, SA0P25, SA0P3, SA0P4, SA0P5, SA0P75,
      SA1P0, SA1P5, SA2P0, SA3P0, SA4P0, SA5P0, SA7P5,
      SA10P0);

  static final Set<Imt> IMTS_EXTENDS = EnumSet.of(PGA, SA0P2, SA1P0);

  @TempDir
  static Path testFolder;

  @Test
  void createAndWrite() throws IOException {
    CalcConfig copy = CalcConfig.copyOf(DEFAULTS).build();
    assertEquals(DEFAULTS.toString(), copy.toString());

    DEFAULTS.write(testFolder);
    Path written = testFolder.resolve(CalcConfig.FILE_NAME);
    CalcConfig config = CalcConfig.from(written).build();
    assertEquals(DEFAULTS.toString(), config.toString());

    // tests non-null service URL serialization
    EXTENDS.write(testFolder);
    written = testFolder.resolve(CalcConfig.FILE_NAME);
    config = CalcConfig.from(written).build();
    assertEquals(EXTENDS.toString(), config.toString());
  }

  @Test
  void publicOverrides() throws IOException {
    Set<Imt> imts = Set.of(PGA, SA0P2, SA1P0);
    double distance = 50.0;
    double timespan = 100.0;
    CalcConfig config = CalcConfig.defaults()
        .imts(imts)
        .distance(distance)
        .timespan(timespan)
        .build();
    assertEquals(imts, config.hazard.imts);
    assertEquals(distance, config.rate.distance);
    assertEquals(timespan, config.rate.timespan);
    assertEquals(ValueFormat.POISSON_PROBABILITY, config.rate.valueFormat);
  }

  @Test
  void testHazardMember() {
    CalcConfig.Hazard def = DEFAULTS.hazard;
    assertEquals(ExceedanceModel.TRUNCATION_3SIGMA_UPPER, def.exceedanceModel);
    assertEquals(3.0, def.truncationLevel);
    assertEquals(IMTS, def.imts);
    assertEquals(EnumSet.noneOf(TectonicSetting.class), def.tectonicSettings);
    assertEquals(EnumSet.noneOf(SourceType.class), def.tectonicSettings);
    assertEquals(5.0, def.gmmDampingRatio);
    assertEquals(false, def.gmmDampingSigma);
    assertEquals(ValueFormat.ANNUAL_RATE, def.valueFormat);

    Map<Imt, XySequence> imlMap = def.modelCurves();
    Map<Imt, XySequence> imlMapExpect = new EnumMap<>(Imt.class);
    for (Imt imt : IMTS) {
      double[] imls = imlsForImt(imt);
      imlMapExpect.put(imt, XySequence.create(imls, new double[imls.length]));
    }
    assertEquals(imlMap, imlMapExpect);

    /* Test MPRS IMLs. */
    Map<Imt, XySequence> curves = DEFAULTS.hazard.modelCurves();
    assertArrayEquals(IMLS_PGV_TEST, curves.get(PGV).xValues().toArray());
    assertArrayEquals(IMLS_SA0P01_TEST, curves.get(PGA).xValues().toArray());
    assertArrayEquals(IMLS_SA0P01_TEST, curves.get(SA0P01).xValues().toArray());
    assertArrayEquals(IMLS_SA0P02_TEST, curves.get(SA0P02).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P03).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P05).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P075).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P1).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P15).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P2).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P25).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P3).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P4).xValues().toArray());
    assertArrayEquals(IMLS_SA0P5_TEST, curves.get(SA0P5).xValues().toArray());
    assertArrayEquals(IMLS_SA2P0_TEST, curves.get(SA0P75).xValues().toArray());
    assertArrayEquals(IMLS_SA2P0_TEST, curves.get(SA1P0).xValues().toArray());
    assertArrayEquals(IMLS_SA2P0_TEST, curves.get(SA1P5).xValues().toArray());
    assertArrayEquals(IMLS_SA2P0_TEST, curves.get(SA2P0).xValues().toArray());
    assertArrayEquals(IMLS_SA3P0_TEST, curves.get(SA3P0).xValues().toArray());
    assertArrayEquals(IMLS_SA4P0_TEST, curves.get(SA4P0).xValues().toArray());
    assertArrayEquals(IMLS_SA5P0_TEST, curves.get(SA5P0).xValues().toArray());
    assertArrayEquals(IMLS_SA7P5_TEST, curves.get(SA7P5).xValues().toArray());
    assertArrayEquals(IMLS_SA10P0_TEST, curves.get(SA10P0).xValues().toArray());

    assertThrows(UnsupportedOperationException.class, () -> {
      CalcConfig.defaults()
          .extend(CalcConfig.from(RESOURCES.resolve("calc-config-extends-bad-imt.json")))
          .build();
    });

    Map<Imt, XySequence> logCurves = DEFAULTS.hazard.logModelCurves();
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_PGV_TEST, IMLS_PGV_TEST.length)),
        logCurves.get(PGV).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P01_TEST, IMLS_SA0P01_TEST.length)),
        logCurves.get(PGA).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P01_TEST, IMLS_SA0P01_TEST.length)),
        logCurves.get(SA0P01).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P02_TEST, IMLS_SA0P02_TEST.length)),
        logCurves.get(SA0P02).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P03).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P05).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P075).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P1).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P15).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P2).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P25).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P3).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P4).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA0P5_TEST, IMLS_SA0P5_TEST.length)),
        logCurves.get(SA0P5).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA2P0_TEST, IMLS_SA2P0_TEST.length)),
        logCurves.get(SA0P75).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA2P0_TEST, IMLS_SA2P0_TEST.length)),
        logCurves.get(SA1P0).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA2P0_TEST, IMLS_SA2P0_TEST.length)),
        logCurves.get(SA1P5).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA2P0_TEST, IMLS_SA2P0_TEST.length)),
        logCurves.get(SA2P0).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA3P0_TEST, IMLS_SA3P0_TEST.length)),
        logCurves.get(SA3P0).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA4P0_TEST, IMLS_SA4P0_TEST.length)),
        logCurves.get(SA4P0).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA5P0_TEST, IMLS_SA5P0_TEST.length)),
        logCurves.get(SA5P0).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA7P5_TEST, IMLS_SA7P5_TEST.length)),
        logCurves.get(SA7P5).xValues().toArray());
    assertArrayEquals(
        DoubleData.ln(Arrays.copyOf(IMLS_SA10P0_TEST, IMLS_SA10P0_TEST.length)),
        logCurves.get(SA10P0).xValues().toArray());

    def = EXTENDS.hazard;
    assertEquals(ExceedanceModel.TRUNCATION_OFF, def.exceedanceModel);
    assertEquals(2.0, def.truncationLevel);
    assertEquals(IMTS_EXTENDS, def.imts);
    assertEquals(EnumSet.of(ACTIVE_CRUST, SUBDUCTION), def.tectonicSettings);
    assertEquals(EnumSet.of(FAULT, ZONE, SLAB), def.sourceTypes);
    assertEquals(3.0, def.gmmDampingRatio);
    assertEquals(true, def.gmmDampingSigma);
    assertEquals(ValueFormat.POISSON_PROBABILITY, def.valueFormat);

    assertArrayEquals(
        IMLS_PGA_CUSTOM,
        def.modelCurves().get(PGA).xValues().toArray());

    def = EXTENDS_EMPTY.hazard;
    assertEquals(ExceedanceModel.TRUNCATION_3SIGMA_UPPER, def.exceedanceModel);
    assertEquals(3.0, def.truncationLevel);
    assertEquals(IMTS, def.imts);
    assertEquals(EnumSet.noneOf(TectonicSetting.class), def.tectonicSettings);
    assertEquals(EnumSet.noneOf(SourceType.class), def.tectonicSettings);
    assertEquals(5.0, def.gmmDampingRatio);
    assertEquals(false, def.gmmDampingSigma);
    assertEquals(ValueFormat.ANNUAL_RATE, def.valueFormat);
  }

  @Test
  void testDisaggMember() {
    CalcConfig.Disagg def = DEFAULTS.disagg;
    assertEquals(2475, def.returnPeriod);
    assertEquals(0.0, def.bins.rMin);
    assertEquals(1000.0, def.bins.rMax);
    assertEquals(20.0, def.bins.Δr);
    assertEquals(4.4, def.bins.mMin);
    assertEquals(9.4, def.bins.mMax);
    assertEquals(0.2, def.bins.Δm);
    assertEquals(-3.0, def.bins.εMin);
    assertEquals(3.0, def.bins.εMax);
    assertEquals(0.5, def.bins.Δε);
    assertEquals(1.0, def.contributorLimit);

    def = EXTENDS.disagg;
    assertEquals(475, def.returnPeriod);
    assertEquals(10.0, def.bins.rMin);
    assertEquals(100.0, def.bins.rMax);
    assertEquals(10.0, def.bins.Δr);
    assertEquals(5.0, def.bins.mMin);
    assertEquals(8.0, def.bins.mMax);
    assertEquals(0.1, def.bins.Δm);
    assertEquals(-2.0, def.bins.εMin);
    assertEquals(2.0, def.bins.εMax);
    assertEquals(1.0, def.bins.Δε);
    assertEquals(0.1, def.contributorLimit);
  }

  @Test
  void testRateMember() {
    CalcConfig.Rate def = DEFAULTS.rate;
    assertEquals(4.2, def.bins.mMin);
    assertEquals(9.4, def.bins.mMax);
    assertEquals(0.1, def.bins.Δm);

    assertEquals(20.0, def.distance);
    assertEquals(DistributionFormat.INCREMENTAL, def.distributionFormat);
    assertEquals(ValueFormat.ANNUAL_RATE, def.valueFormat);
    assertEquals(30.0, def.timespan);

    def = EXTENDS.rate;
    assertEquals(5.0, def.bins.mMin);
    assertEquals(8.0, def.bins.mMax);
    assertEquals(0.2, def.bins.Δm);

    assertEquals(40.0, def.distance);
    assertEquals(DistributionFormat.CUMULATIVE, def.distributionFormat);
    assertEquals(ValueFormat.POISSON_PROBABILITY, def.valueFormat);
    assertEquals(20.0, def.timespan);

    def = EXTENDS_EMPTY.rate;
    assertEquals(4.2, def.bins.mMin);
    assertEquals(9.4, def.bins.mMax);
    assertEquals(0.1, def.bins.Δm);

    assertEquals(20.0, def.distance);
    assertEquals(DistributionFormat.INCREMENTAL, def.distributionFormat);
    assertEquals(ValueFormat.ANNUAL_RATE, def.valueFormat);
    assertEquals(30.0, def.timespan);
  }

  @Test
  void testSiteMember() {
    CalcConfig.SiteDefaults def = DEFAULTS.site;
    assertEquals(760.0, def.vs30);
    assertEquals(true, def.vsInferred);
    assertEquals(Double.NaN, def.z1p0);
    assertEquals(Double.NaN, def.z2p5);

    def = EXTENDS.site;
    assertEquals(530.0, def.vs30);
    assertEquals(false, def.vsInferred);
    assertEquals(0.3, def.z1p0);
    assertEquals(1.0, def.z2p5);

    def = EXTENDS_EMPTY.site;
    assertEquals(760.0, def.vs30);
    assertEquals(true, def.vsInferred);
    assertEquals(Double.NaN, def.z1p0);
    assertEquals(Double.NaN, def.z2p5);
  }

  @Test
  void testOutputMember() {
    List<Integer> defaultReturnPeroiods = List.of(475, 975, 2475);

    CalcConfig.Output def = DEFAULTS.output;
    assertEquals(Path.of("hazout"), def.directory);
    assertEquals(Set.of(DataType.TOTAL, DataType.MAP), def.dataTypes);
    assertEquals(defaultReturnPeroiods, def.returnPeriods);

    def = EXTENDS.output;
    assertEquals(Path.of("custom"), def.directory);
    // also tests automatic addition of TOTAL and MAP
    assertEquals(Set.of(DataType.TOTAL, DataType.MAP, DataType.GMM), def.dataTypes);
    assertEquals(List.of(100, 200), def.returnPeriods);

    def = EXTENDS_EMPTY.output;
    assertEquals(Path.of("hazout"), def.directory);
    assertEquals(Set.of(DataType.TOTAL, DataType.MAP), def.dataTypes);
    assertEquals(defaultReturnPeroiods, def.returnPeriods);
  }

  @Test
  void testPerformanceMember() {
    CalcConfig.Performance def = DEFAULTS.performance;
    assertEquals(true, def.optimizeGrids);
    assertEquals(true, def.smoothGrids);
    assertEquals(1000, def.systemPartition);
    assertEquals(ThreadCount.ALL, def.threadCount);

    def = EXTENDS.performance;
    assertEquals(false, def.optimizeGrids);
    assertEquals(false, def.smoothGrids);
    assertEquals(500, def.systemPartition);
    assertEquals(ThreadCount.ONE, def.threadCount);

    def = EXTENDS_EMPTY.performance;
    assertEquals(true, def.optimizeGrids);
    assertEquals(true, def.smoothGrids);
    assertEquals(1000, def.systemPartition);
    assertEquals(ThreadCount.ALL, def.threadCount);
  }

  static double[] imlsForImt(Imt imt) {
    if (imt.isSA() || imt == Imt.PGA) {
      return imlMap.ceilingEntry(imt).getValue();
    } else if (imt == Imt.PGV) {
      return IMLS_PGV_TEST;
    } else {
      /* IMLs not yet assigned for IMT. */
      throw new UnsupportedOperationException();
    }
  }

  static final double[] IMLS_PGA_CUSTOM = {
      0.0025, 0.0075, 0.0169, 0.0380, 0.0854, 0.192, 0.432, 0.973, 2.19, 4.92
  };

  static final double[] IMLS_SA0P01_TEST = {
      0.00233, 0.00350, 0.00524, 0.00786, 0.0118, 0.0177, 0.0265, 0.0398, 0.0597, 0.0896, 0.134,
      0.202, 0.302, 0.454, 0.680, 1.02, 1.53, 2.30, 3.44, 5.17 };
  static final double[] IMLS_SA0P02_TEST = {
      0.00283, 0.00424, 0.00637, 0.00955, 0.0143, 0.0215, 0.0322, 0.0483, 0.0725, 0.109, 0.163,
      0.245, 0.367, 0.551, 0.826, 1.24, 1.86, 2.79, 4.18, 6.27 };
  static final double[] IMLS_SA0P5_TEST = {
      0.00333, 0.00499, 0.00749, 0.0112, 0.0169, 0.0253, 0.0379, 0.0569, 0.0853, 0.128, 0.192,
      0.288, 0.432, 0.648, 0.972, 1.46, 2.19, 3.28, 4.92, 7.38 };
  static final double[] IMLS_SA2P0_TEST = {
      0.00250, 0.00375, 0.00562, 0.00843, 0.0126, 0.0190, 0.0284, 0.0427, 0.0640, 0.0960, 0.144,
      0.216, 0.324, 0.486, 0.729, 1.09, 1.64, 2.46, 3.69, 5.54 };
  static final double[] IMLS_SA3P0_TEST = {
      0.00200, 0.00300, 0.00449, 0.00674, 0.0101, 0.0152, 0.0228, 0.0341, 0.0512, 0.0768, 0.115,
      0.173, 0.259, 0.389, 0.583, 0.875, 1.31, 1.97, 2.95, 4.43 };
  static final double[] IMLS_SA4P0_TEST = {
      0.00133, 0.00200, 0.00300, 0.00449, 0.00674, 0.0101, 0.0152, 0.0228, 0.0341, 0.0512, 0.0768,
      0.115, 0.173, 0.259, 0.389, 0.583, 0.875, 1.31, 1.97, 2.95 };
  static final double[] IMLS_SA5P0_TEST = {
      0.000999, 0.00150, 0.00225, 0.00337, 0.00506, 0.00758, 0.0114, 0.0171, 0.0256, 0.0384, 0.0576,
      0.0864, 0.130, 0.194, 0.292, 0.437, 0.656, 0.984, 1.48, 2.21 };
  static final double[] IMLS_SA7P5_TEST = {
      0.000499, 0.000749, 0.00112, 0.00169, 0.00253, 0.00379, 0.00569, 0.00853, 0.0128, 0.0192,
      0.0288, 0.0432, 0.0648, 0.0972, 0.146, 0.219, 0.328, 0.492, 0.738, 1.11 };
  static final double[] IMLS_SA10P0_TEST = {
      0.000333, 0.000499, 0.000749, 0.00112, 0.00169, 0.00253, 0.00379, 0.00569, 0.00853, 0.0128,
      0.0192, 0.0288, 0.0432, 0.0648, 0.0972, 0.146, 0.219, 0.328, 0.492, 0.738 };

  static final double[] IMLS_PGV_TEST = {
      0.237, 0.355, 0.532, 0.798, 1.19, 1.80, 2.69, 4.04, 6.06, 9.09, 13.6, 20.5, 30.7,
      46.0, 69.0, 103.0, 155.0, 233.0, 349.0, 525.0 };

  /* NavigableMap facilitates assigning IMLs to ranges of spectral periods. */
  static final NavigableMap<Imt, double[]> imlMap = new TreeMap<>(Map.of(
      SA0P01, IMLS_SA0P01_TEST,
      SA0P02, IMLS_SA0P02_TEST,
      SA0P5, IMLS_SA0P5_TEST,
      SA2P0, IMLS_SA2P0_TEST,
      SA3P0, IMLS_SA3P0_TEST,
      SA4P0, IMLS_SA4P0_TEST,
      SA5P0, IMLS_SA5P0_TEST,
      SA7P5, IMLS_SA7P5_TEST,
      SA10P0, IMLS_SA10P0_TEST));

}
