package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.OptionalDouble;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;

class DelimitedDataTests {

  private static final Path DATA_PATH = Paths.get("src/test/resources/data");

  private static final Path TEST_DATA_CSV = DATA_PATH.resolve("delimited-data-tests.csv");
  private static final Path TEST_DATA = DATA_PATH.resolve("delimited-data-tests.txt");
  private static final Path TEST_DATA_CSV_CRLF = DATA_PATH.resolve("delimited-data-tests-CRLF.csv");
  private static final Path TEST_DATA_CRLF = DATA_PATH.resolve("delimited-data-tests-CRLF.txt");
  private static final Path TEST_DATA_EMPTY = DATA_PATH.resolve("delimited-data-tests-empty.csv");
  private static final Path TEST_DATA_BAD_HEADER_1_CSV =
      DATA_PATH.resolve("delimited-data-tests-bad-header-1.csv");
  private static final Path TEST_DATA_BAD_HEADER_2_CSV =
      DATA_PATH.resolve("delimited-data-tests-bad-header-2.csv");
  private static final Path TEST_DATA_BAD_RECORD_CSV =
      DATA_PATH.resolve("delimited-data-tests-bad-record.csv");
  private static final Path TEST_DATA_BAD_RECORD =
      DATA_PATH.resolve("delimited-data-tests-bad-record.txt");
  private static final Path TEST_DATA_CSV_NO_COMMENTS =
      DATA_PATH.resolve("delimited-data-tests-no-comments.csv");

  @TempDir
  static Path tmp;

  @Test
  void testCreateExceptions() {
    assertThrows(NullPointerException.class, () -> {
      List<String> lines = null;
      DelimitedData.comma(lines);
    });

    assertThrows(NullPointerException.class, () -> {
      Path path = null;
      DelimitedData.comma(path);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DelimitedData.comma(DATA_PATH.resolve("no-file.csv"));
    });

    assertThrows(IllegalStateException.class, () -> {
      DelimitedData.comma(TEST_DATA_EMPTY);
    });

    assertThrows(IllegalStateException.class, () -> {
      DelimitedData.comma(TEST_DATA_BAD_HEADER_1_CSV);
    });

    assertThrows(IllegalStateException.class, () -> {
      DelimitedData.comma(TEST_DATA_BAD_HEADER_2_CSV);
    });

    assertThrows(UncheckedIOException.class, () -> {
      /*
       * It's difficult to test IOException throws by Files.lines() nested in a
       * Supplier lambda. Create temporary valid csv file that will get through
       * the Files.exists() check and header reading, but then delete it prior
       * to calling records().
       */
      Path tmpFile = tmp.resolve("disappearing-file.csv");
      Files.copy(TEST_DATA_CSV, tmpFile);
      DelimitedData dd = DelimitedData.comma(tmpFile);
      Files.delete(tmpFile);
      dd.records();
    });

  }

  private static final Map<Integer, String> COLUMN_KEYMAP = Map.of(
      0, "lon", 1, "lat", 2, "a", 3, "b", 4, "c", 5, "d");
  // private static final Set<String> COLUMN_KEYS = Set.of("lon", "lat", "a",
  // "b", "c", "d");
  private static final List<String> EXPECTED_STRINGS = List.of(
      "[-100, 40, 1, 2, 3, 4]",
      "[-110, 30, , , true, false]",
      "[-120.5, 20.12345, 5.2, 1e-4, test, me]");
  private static final List<List<Object>> EXPECTED_VALUES = List.of(
      List.of(-100., 40., 1., 2., 3., 4.),
      List.of(-110., 30., OptionalDouble.empty(), OptionalDouble.empty(), true, false),
      List.of(-120.5, 20.12345, 5.2, 1e-4, "test", "me"));

  @Test
  void testColumnKeysCsv() throws IOException {
    DelimitedData csvPath = DelimitedData.comma(TEST_DATA_CSV);
    DelimitedData csvLines = DelimitedData.comma(Files.readAllLines(TEST_DATA_CSV));
    assertEquals(Set.copyOf(COLUMN_KEYMAP.values()), csvPath.columnKeys());
    assertEquals(Set.copyOf(COLUMN_KEYMAP.values()), csvLines.columnKeys());
    assertEquals(csvPath.columnKeys(), csvLines.columnKeys());
  }

  @Test
  void testColumnKeysWhitespace() throws IOException {
    DelimitedData dataPath = DelimitedData.whitespace(TEST_DATA);
    DelimitedData dataLines = DelimitedData.whitespace(Files.readAllLines(TEST_DATA_CRLF));
    assertEquals(Set.copyOf(COLUMN_KEYMAP.values()), dataPath.columnKeys());
    assertEquals(Set.copyOf(COLUMN_KEYMAP.values()), dataLines.columnKeys());
    assertEquals(dataPath.columnKeys(), dataLines.columnKeys());
  }

  @Test
  void testComments() throws IOException {
    DelimitedData data = DelimitedData.comma(TEST_DATA_CSV);
    List<String> comments = data.comments();
    assertEquals(3, comments.size());
    List<String> expected = Files.readAllLines(TEST_DATA_CSV).stream()
        .filter(line -> line.startsWith("#"))
        .collect(Collectors.toList());
    assertEquals(expected, comments);

    // file containing no comment lines should return Optional.empty()
    data = DelimitedData.comma(TEST_DATA_CSV_NO_COMMENTS);
    assertTrue(data.comments().isEmpty());
  }

  /*
   * Collect given {@code DelimitedData} to a list of {@code Record}s using
   * proper try-with statement to ensure resources are closed
   */
  private List<Record> toList(DelimitedData d) {
    try (Stream<Record> s = d.records()) {
      return s.collect(Collectors.toList());
    }
  }

  @Test
  void testRecords() throws IOException {
    List<Record> recordsPath = toList(DelimitedData.comma(TEST_DATA_CSV));
    List<Record> recordsLines = toList(DelimitedData.comma(Files.readAllLines(TEST_DATA_CSV_CRLF)));

    int recNo = 0;
    int i = 0;
    String col;
    // check doubles
    for (i = 0; i < COLUMN_KEYMAP.size(); i++) {
      double expected = (double) EXPECTED_VALUES.get(recNo).get(i);
      col = COLUMN_KEYMAP.get(i);
      assertEquals(expected, recordsPath.get(recNo).getDouble(col), 0.0);
      assertEquals(expected, recordsLines.get(recNo).getDouble(col), 0.0);
    }

    recNo = 1;
    // check booleans
    for (i = 4; i < 6; i++) {
      boolean expected = (boolean) EXPECTED_VALUES.get(recNo).get(i);
      col = COLUMN_KEYMAP.get(i);
      assertEquals(expected, recordsPath.get(recNo).getBoolean(col));
      assertEquals(expected, recordsLines.get(recNo).getBoolean(col));
    }

    i = 2;
    col = COLUMN_KEYMAP.get(i);
    // check OptionalDouble
    assertSame(OptionalDouble.empty(), recordsPath.get(recNo).getOptionalDouble(col));
    assertSame(OptionalDouble.empty(), recordsLines.get(recNo).getOptionalDouble(col));

    recNo = 2;
    // check doubles
    for (i = 0; i < COLUMN_KEYMAP.size() - 2; i++) {
      double expected = (double) EXPECTED_VALUES.get(recNo).get(i);
      col = COLUMN_KEYMAP.get(i);
      assertEquals(expected,
          recordsPath.get(recNo).getOptionalDouble(col).orElseThrow(), 0.0);
      assertEquals(expected,
          recordsLines.get(recNo).getOptionalDouble(col).orElseThrow(), 0.0);
    }
    // check strings
    assertEquals(EXPECTED_VALUES.get(recNo).get(4),
        recordsPath.get(recNo).get(COLUMN_KEYMAP.get(4)));
    assertEquals("test", recordsLines.get(recNo).get("c"));
  }

  @Test
  void testGet_invalidKey() {
    List<Record> records = toList(DelimitedData.comma(TEST_DATA_CSV));
    assertThrows(
        NoSuchElementException.class, () -> {
          records.get(0).getDouble("not_a_field");
        });
    assertThrows(
        NoSuchElementException.class, () -> {
          records.get(0).getBoolean("not_a_field");
        });
    assertThrows(
        NoSuchElementException.class, () -> {
          records.get(0).getOptionalDouble("not_a_field");
        });
    assertThrows(
        NoSuchElementException.class, () -> {
          records.get(0).get("not_a_field");
        });
  }

  @Test
  void testRecordsBadCsv() {
    try (Stream<Record> s = DelimitedData.comma(TEST_DATA_BAD_RECORD_CSV).records()) {
      assertThrows(IllegalStateException.class, () -> {
        s.collect(Collectors.toList());
      });
    }
  }

  @Test
  void testRecordsBadWhitespace() {
    try (Stream<Record> s = DelimitedData.whitespace(TEST_DATA_BAD_RECORD).records()) {
      assertThrows(IllegalStateException.class, () -> {
        s.collect(Collectors.toList());
      });
    }
  }

  @Test
  void testToString() {
    List<Record> recs = toList(DelimitedData.comma(TEST_DATA_CSV));
    for (int i = 0; i < recs.size(); i++) {
      assertEquals(EXPECTED_STRINGS.get(i), recs.get(i).toString());
    }
    recs = toList(DelimitedData.comma(TEST_DATA_CSV_CRLF));
    for (int i = 0; i < recs.size(); i++) {
      assertEquals(EXPECTED_STRINGS.get(i), recs.get(i).toString());
    }
  }

}
