package gov.usgs.earthquake.nshmp.fault;

import static org.junit.jupiter.api.Assertions.assertEquals;

import java.util.stream.DoubleStream;
import java.util.stream.Stream;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;

import com.google.common.collect.Range;

class FocalMechTest {

  // Rake bins
  static final Range<Double> REVERSE_RAKES = Range.closed(45., 135.);
  static final Range<Double> NORMAL_RAKES = Range.closed(-135., -45.);

  @Test
  final void testFromRakeDefault() {
    assertEquals(FocalMech.NORMAL, FocalMech.fromRake(FocalMech.NORMAL.rake()));
    assertEquals(FocalMech.REVERSE, FocalMech.fromRake(FocalMech.REVERSE.rake()));
    assertEquals(FocalMech.STRIKE_SLIP, FocalMech.fromRake(FocalMech.STRIKE_SLIP.rake()));
  }

  @ParameterizedTest(name = "{index} ==> {0}, {1}")
  @MethodSource("byRake")
  final void testFromRake(FocalMech expected, Double rake) {
    assertEquals(expected, FocalMech.fromRake(rake));

    // Simpler, without ParameterizedTest ...
    // List<Double> rakes = DoubleStream.iterate(-180.0,
    // d -> d <= 180.0,
    // d -> d + 5.0)
    // .boxed()
    // .collect(Collectors.toList());
    // for (Double rake : rakes) {
    // FocalMech expected = (REVERSE_RAKES.contains(rake)) ? FocalMech.REVERSE
    // : (NORMAL_RAKES.contains(rake)) ? FocalMech.NORMAL :
    // FocalMech.STRIKE_SLIP;
    // assertEquals(expected, FocalMech.fromRake(rake));
    // }
  }

  private static Stream<? extends Arguments> byRake() {
    return DoubleStream.iterate(-180.0, d -> d <= 180.0, d -> d + 5.0)
        .boxed()
        .map(d -> new Object[] { parseRake(d), d })
        .map(Arguments::of);
  }

  private static FocalMech parseRake(Double rake) {
    return (REVERSE_RAKES.contains(rake)) ? FocalMech.REVERSE
        : (NORMAL_RAKES.contains(rake)) ? FocalMech.NORMAL : FocalMech.STRIKE_SLIP;
  }

  @Test
  final void testDip() {
    // tests 'standard' dip
    assertEquals(50.0, FocalMech.NORMAL.dip());
    assertEquals(50.0, FocalMech.REVERSE.dip());
    assertEquals(90.0, FocalMech.STRIKE_SLIP.dip());
  }

  @Test
  final void testRake() {
    // tests 'standard' rake
    assertEquals(-90.0, FocalMech.NORMAL.rake());
    assertEquals(90.0, FocalMech.REVERSE.rake());
    assertEquals(0.0, FocalMech.STRIKE_SLIP.rake());
  }

  @Test
  final void testToString() {
    assertEquals("Normal", FocalMech.NORMAL.toString());
    assertEquals("Reverse", FocalMech.REVERSE.toString());
    assertEquals("Strike-Slip", FocalMech.STRIKE_SLIP.toString());
  }

}
