package gov.usgs.earthquake.nshmp.geo.json;

import static gov.usgs.earthquake.nshmp.geo.json.GeoJsonTests.TEST_LINE;
import static gov.usgs.earthquake.nshmp.geo.json.GeoJsonTests.TEST_MULTI_LINE;
import static gov.usgs.earthquake.nshmp.geo.json.GeoJsonTests.TEST_POINT;
import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson.Type;

class GeometryTests {

  // private static final Location TEST_POINT = Location.create(-117, 34);

  private static final Location POINT_WITH_DEPTH = Location.create(-117, 34, 1);

  // private static final LocationList TEST_LINE = LocationList.of(
  // Location.create(-117, 34),
  // Location.create(-118, 35),
  // Location.create(-116, 37),
  // Location.create(-117, 38));

  // private static final List<LocationList> TEST_MULTI_LINE =
  // Lists.newArrayList(
  // TEST_LINE,
  // LocationList.of(
  // Location.create(-117, 44),
  // Location.create(-118, 45),
  // Location.create(-116, 47),
  // Location.create(-117, 48)));

  private static final double[] POINT_COORDS =
      new double[] { TEST_POINT.longitude, TEST_POINT.latitude };

  private static final double[][] LINE_COORDS = new double[TEST_LINE.size()][];

  private static final double[][][] MULTI_LINE_COORDS =
      new double[TEST_MULTI_LINE.size()][TEST_LINE.size()][];

  private static final Geometry<double[]> POINT = new Geometry<>(Type.POINT, POINT_COORDS);

  private static final Geometry<double[][]> LINE_STRING =
      new Geometry<>(Type.LINE_STRING, LINE_COORDS);

  private static final Geometry<double[][][]> MULTI_LINE_STRING =
      new Geometry<>(Type.MULTI_LINE_STRING, MULTI_LINE_COORDS);

  private static final Geometry<double[][][]> POLYGON =
      new Geometry<>(Type.POLYGON,
          MULTI_LINE_COORDS);

  static {
    for (int i = 0; i < TEST_LINE.size(); i++) {
      LINE_COORDS[i] = new double[] {
          TEST_LINE.get(i).longitude,
          TEST_LINE.get(i).latitude };
    }

    for (int i = 0; i < TEST_MULTI_LINE.size(); i++) {
      LocationList locs = TEST_MULTI_LINE.get(i);
      for (int j = 0; j < locs.size(); j++) {
        MULTI_LINE_COORDS[i][j] = new double[] {
            locs.get(j).longitude,
            locs.get(j).latitude
        };
      }
    }
  }

  @Nested
  @DisplayName("Point Tests")
  class PointTests {

    @Test
    final void pointTestUOE() {
      assertThrows(UnsupportedOperationException.class, () -> {
        POINT.toLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        POINT.toMultiLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        POINT.toPolygonBorder();
      });
    }

    @Test
    final void pointTests() {
      assertArrayEquals(POINT_COORDS, POINT.coordinates, 0);
      assertEquals(TEST_POINT, POINT.toPoint());
      assertTrue(TEST_POINT.equals(POINT.toPoint()));
      assertEquals(Type.POINT, POINT.type);

      Geometry<double[]> actual = Geometry.point(TEST_POINT);
      assertArrayEquals(POINT_COORDS, actual.coordinates, 0);
      assertEquals(TEST_POINT, actual.toPoint());
      assertTrue(TEST_POINT.equals(actual.toPoint()));
      assertEquals(Type.POINT, actual.type);

      Geometry<double[]> depthActual = Geometry.point(POINT_WITH_DEPTH);
      assertArrayEquals(new double[] { -117, 34, 1 }, depthActual.coordinates);
      assertEquals(POINT_WITH_DEPTH, depthActual.toPoint());
    }
  }

  @Nested
  @DisplayName("LineString Tests")
  class LineStringTests {

    @Test
    final void lineStringTestUOE() {
      assertThrows(UnsupportedOperationException.class, () -> {
        LINE_STRING.toPoint();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        LINE_STRING.toMultiLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        LINE_STRING.toPolygonBorder();
      });
    }

    @Test
    final void lineStringTests() {
      assertArrayEquals(LINE_COORDS, LINE_STRING.coordinates);
      assertEquals(TEST_LINE, LINE_STRING.toLineString());
      assertTrue(TEST_LINE.equals(LINE_STRING.toLineString()));
      assertEquals(Type.LINE_STRING, LINE_STRING.type);

      Geometry<double[][]> actual = Geometry.lineString(TEST_LINE);
      assertArrayEquals(LINE_COORDS, actual.coordinates);
      assertEquals(TEST_LINE, actual.toLineString());
      assertTrue(TEST_LINE.equals(actual.toLineString()));
      assertEquals(Type.LINE_STRING, actual.type);
    }
  }

  @Nested
  @DisplayName("MultiLineString Tests")
  class MultiLineStringTests {

    @Test
    final void multiLineStringTestUOE() {
      assertThrows(UnsupportedOperationException.class, () -> {
        MULTI_LINE_STRING.toPoint();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        MULTI_LINE_STRING.toLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        MULTI_LINE_STRING.toPolygonBorder();
      });
    }

    @Test
    final void multiLineStringTests() {
      assertArrayEquals(MULTI_LINE_COORDS, MULTI_LINE_STRING.coordinates);
      assertEquals(TEST_MULTI_LINE, MULTI_LINE_STRING.toMultiLineString());
      assertTrue(TEST_MULTI_LINE.equals(MULTI_LINE_STRING.toMultiLineString()));
      assertEquals(Type.MULTI_LINE_STRING, MULTI_LINE_STRING.type);

      Geometry<double[][][]> actual = Geometry.multiLineString(TEST_MULTI_LINE);
      assertArrayEquals(MULTI_LINE_COORDS, actual.coordinates);
      assertEquals(TEST_MULTI_LINE, actual.toMultiLineString());
      assertTrue(TEST_MULTI_LINE.equals(actual.toMultiLineString()));
      assertEquals(Type.MULTI_LINE_STRING, actual.type);

      // .toMultiLineString() should return immutable list
      assertThrows(UnsupportedOperationException.class, () -> {
        var dummy = actual.toMultiLineString().add(TEST_LINE);
      });
    }
  }

  @Nested
  @DisplayName("Polygon Tests")
  class PolygonTests {

    @Test
    final void polygonTestUOE_toPoint() {
      assertThrows(UnsupportedOperationException.class, () -> {
        POLYGON.toPoint();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        POLYGON.toLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        POLYGON.toMultiLineString();
      });
    }

    @Test
    final void polygonTests() {
      assertEquals(Type.POLYGON, POLYGON.type);

      Geometry<double[][][]> actual = Geometry
          .polygon(TEST_MULTI_LINE.get(0), TEST_MULTI_LINE.get(1));
      assertEquals(TEST_MULTI_LINE.get(0), actual.toPolygonBorder());
      assertTrue(TEST_MULTI_LINE.get(0).equals(actual.toPolygonBorder()));
      assertEquals(Type.POLYGON, actual.type);
    }
  }

}
