package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.DIP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RAKE;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RJB;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RX;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VSINF;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.WIDTH;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.Z1P0;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.Z2P5;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZHYP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.HashMap;
import java.util.Map;
import java.util.stream.Stream;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtensionContext;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.ArgumentsProvider;
import org.junit.jupiter.params.provider.ArgumentsSource;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Builder;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Field;

class GmmInputTest {

  private static final double dip = DIP.defaultValue;
  private static final double Mw = MW.defaultValue;
  private static final double rake = RAKE.defaultValue;
  private static final double rJB = RJB.defaultValue;
  private static final double rRup = RRUP.defaultValue;
  private static final double rX = RX.defaultValue;
  private static final double vs30 = VS30.defaultValue;
  private static final boolean vsInf = VSINF.defaultValue > 0.0;
  private static final double width = WIDTH.defaultValue;
  private static final double z1p0 = Z1P0.defaultValue;
  private static final double z2p5 = Z2P5.defaultValue;
  private static final double zHyp = ZHYP.defaultValue;
  private static final double zTor = ZTOR.defaultValue;

  private static final String ALREADY_SET = "Field %s already set";
  private static final Map<Field, String> VALUES = new HashMap<>();

  static {
    VALUES.put(DIP, Double.toString(dip));
    VALUES.put(MW, Double.toString(Mw));
    VALUES.put(RAKE, Double.toString(rake));
    VALUES.put(RJB, Double.toString(rJB));
    VALUES.put(RRUP, Double.toString(rRup));
    VALUES.put(RX, Double.toString(rX));
    VALUES.put(VS30, Double.toString(vs30));
    VALUES.put(VSINF, Boolean.toString(vsInf));
    VALUES.put(WIDTH, Double.toString(width));
    VALUES.put(Z1P0, Double.toString(z1p0));
    VALUES.put(Z2P5, Double.toString(z2p5));
    VALUES.put(ZHYP, Double.toString(zHyp));
    VALUES.put(ZTOR, Double.toString(zTor));
  }

  @ParameterizedTest(name = "({index}) {0}={1}")
  @ArgumentsSource(AlreadySet.class)
  final void builderAlreadySet(Field id, String s) {
    Exception exception = assertThrows(IllegalStateException.class, () -> {
      Builder gmmBuilder = GmmInput.builder().withDefaults();
      gmmBuilder.set(id, s);
      gmmBuilder.set(id, s);
    });

    assertEquals(String.format(ALREADY_SET, id.toString()), exception.getMessage());
  }

  static class AlreadySet implements ArgumentsProvider {
    @Override
    public Stream<? extends Arguments> provideArguments(ExtensionContext context) throws Exception {
      return VALUES.keySet().stream().map(key -> Arguments.of(key, VALUES.get(key)));
    }
  }

  @Test
  final void builderFromCopy() {
    GmmInput model = GmmInput.builder().withDefaults().build();
    GmmInput gmmCopy = GmmInput.builder().fromCopy(model).build();

    assertEquals(gmmCopy.dip, model.dip, 0);
    assertEquals(gmmCopy.Mw, model.Mw, 0);
    assertEquals(gmmCopy.rake, model.rake, 0);
    assertEquals(gmmCopy.rJB, model.rJB, 0);
    assertEquals(gmmCopy.rRup, model.rRup, 0);
    assertEquals(gmmCopy.rX, model.rX, 0);
    assertEquals(gmmCopy.vs30, model.vs30, 0);
    assertEquals(gmmCopy.vsInf, model.vsInf);
    assertEquals(gmmCopy.width, model.width, 0);
    assertEquals(gmmCopy.z1p0, model.z1p0, 0);
    assertEquals(gmmCopy.z2p5, model.z2p5, 0);
    assertEquals(gmmCopy.zHyp, model.zHyp, 0);
    assertEquals(gmmCopy.zTor, model.zTor, 0);
  }

  @Test
  final void builderSet() {
    double val = 10.0;
    boolean bool = true;

    Builder gmmBuilder = GmmInput.builder();
    gmmBuilder.set(DIP, Double.toString(val));
    gmmBuilder.set(MW, Double.toString(val));
    gmmBuilder.set(RAKE, Double.toString(val));
    gmmBuilder.set(RJB, Double.toString(val));
    gmmBuilder.set(RRUP, Double.toString(val));
    gmmBuilder.set(RX, Double.toString(val));
    gmmBuilder.set(VS30, Double.toString(val));
    gmmBuilder.set(VSINF, Boolean.toString(bool));
    gmmBuilder.set(WIDTH, Double.toString(val));
    gmmBuilder.set(Z1P0, Double.toString(val));
    gmmBuilder.set(Z2P5, Double.toString(val));
    gmmBuilder.set(ZHYP, Double.toString(val));
    gmmBuilder.set(ZTOR, Double.toString(val));
    GmmInput gmm = gmmBuilder.build();

    assertEquals(gmm.dip, val, 0);
    assertEquals(gmm.Mw, val, 0);
    assertEquals(gmm.rake, val, 0);
    assertEquals(gmm.rJB, val, 0);
    assertEquals(gmm.rRup, val, 0);
    assertEquals(gmm.rX, val, 0);
    assertEquals(gmm.vs30, val, 0);
    assertEquals(gmm.vsInf, bool);
    assertEquals(gmm.width, val, 0);
    assertEquals(gmm.z1p0, val, 0);
    assertEquals(gmm.z2p5, val, 0);
    assertEquals(gmm.zHyp, val, 0);
    assertEquals(gmm.zTor, val, 0);

    Exception exception = assertThrows(IllegalStateException.class, () -> {
      GmmInput.builder().set(MW, "true");
    });
    assertEquals("Unhandled field: " + MW.toString(), exception.getMessage());
  }

  @Test
  final void builderWithDefaults() {
    GmmInput gmm = GmmInput.builder().withDefaults().build();

    assertEquals(gmm.dip, dip, 0);
    assertEquals(gmm.Mw, Mw, 0);
    assertEquals(gmm.rake, rake, 0);
    assertEquals(gmm.rJB, rJB, 0);
    assertEquals(gmm.rRup, rRup, 0);
    assertEquals(gmm.rX, rX, 0);
    assertEquals(gmm.vs30, vs30, 0);
    assertEquals(gmm.vsInf, vsInf);
    assertEquals(gmm.width, width, 0);
    assertEquals(gmm.z1p0, z1p0, 0);
    assertEquals(gmm.z2p5, z2p5, 0);
    assertEquals(gmm.zHyp, zHyp, 0);
    assertEquals(gmm.zTor, zTor, 0);
  }

  @Test
  final void fieldFromString() {
    Field mwField = Field.fromString("mw");

    assertEquals(MW, mwField);
    assertEquals(MW.defaultValue, mwField.defaultValue, 0);
    assertEquals(MW.id, mwField.id);
    assertEquals(MW.info, mwField.info);
    assertEquals(MW.label, mwField.label);
    assertEquals(MW.units, mwField.units);

    assertThrows(IllegalArgumentException.class, () -> {
      Field.fromString("test");
    });
  }

  @Test
  final void gmmInput() {
    GmmInput gmm = GmmInput.builder()
        .mag(Mw)
        .rJB(rJB)
        .rRup(rRup)
        .rX(rX)
        .dip(dip)
        .width(width)
        .zTor(zTor)
        .zHyp(zHyp)
        .rake(rake)
        .vs30(vs30)
        .vsInf(vsInf)
        .z1p0(z1p0)
        .z2p5(z2p5)
        .build();

    assertEquals(gmm.dip, dip, 0);
    assertEquals(gmm.Mw, Mw, 0);
    assertEquals(gmm.rake, rake, 0);
    assertEquals(gmm.rJB, rJB, 0);
    assertEquals(gmm.rRup, rRup, 0);
    assertEquals(gmm.rX, rX, 0);
    assertEquals(gmm.vs30, vs30, 0);
    assertEquals(gmm.vsInf, vsInf);
    assertEquals(gmm.width, width, 0);
    assertEquals(gmm.z1p0, z1p0, 0);
    assertEquals(gmm.z2p5, z2p5, 0);
    assertEquals(gmm.zHyp, zHyp, 0);
    assertEquals(gmm.zTor, zTor, 0);
  }

  @Test
  final void gmmInputEquals() {
    double v = 25.0;
    Builder gmmBuilder = GmmInput.builder();
    GmmInput gmm = gmmBuilder.withDefaults().build();
    GmmInput gmmCopy = GmmInput.builder().fromCopy(gmm).build();

    assertTrue(gmm.equals(gmmCopy));
    assertTrue(gmm.equals(gmm));
    assertFalse(gmm.equals(null));
    assertFalse(gmm.equals(new Object()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().dip(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().mag(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().rake(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().rJB(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().rRup(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().rX(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().vs30(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().vsInf(false).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().width(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().z1p0(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().z2p5(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().zHyp(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().zTor(v).build()));
  }

  @Test
  final void gmmInputHashCode() {
    GmmInput gmm = GmmInput.builder().withDefaults().build();
    GmmInput gmmCopy = GmmInput.builder().fromCopy(gmm).build();
    GmmInput gmmDiff = GmmInput.builder().withDefaults()
        .mag(10.0).z1p0(10.0).z2p5(10.0).build();

    assertEquals(gmm.hashCode(), gmmCopy.hashCode());
    assertNotEquals(gmm.hashCode(), gmmDiff.hashCode());
  }

  @Test
  final void gmmInputToString() {
    GmmInput gmm = GmmInput.builder().withDefaults().build();
    GmmInput gmmCopy = GmmInput.builder().fromCopy(gmm).build();
    String toString = gmm.toString();
    String toStringCopy = gmmCopy.toString();

    assertTrue(toString.contains("dip="));
    assertTrue(toString.contains("Mw="));
    assertTrue(toString.contains("rake="));
    assertTrue(toString.contains("rJB="));
    assertTrue(toString.contains("rRup="));
    assertTrue(toString.contains("rX="));
    assertTrue(toString.contains("vs30="));
    assertTrue(toString.contains("vsInf="));
    assertTrue(toString.contains("width="));
    assertTrue(toString.contains("z1p0="));
    assertTrue(toString.contains("z2p5="));
    assertTrue(toString.contains("zHyp="));
    assertTrue(toString.contains("zTor="));
    assertEquals(toString, toStringCopy);
  }

}
