package gov.usgs.earthquake.nshmp.internal;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NavigableMap;
import java.util.NoSuchElementException;
import java.util.TreeMap;
import java.util.function.Function;

import org.junit.jupiter.api.Test;

class ParsingTests {

  private enum TestEnum {
    ABC,
    DEF,
    G_HI_J;

    @Override
    public String toString() {
      return this.name().toLowerCase();
    }
  }

  // Consider insertion-order preserving version of a map for these expected
  // results?
  private static final String ENUM_WEIGHT_STRING_1 = " [ G_HI_J : 0.3, ABC : 0.5, DEF: 0.2]";
  private static final Map<TestEnum, Double> ENUM_WEIGHT_MAP_1 =
      Map.of(TestEnum.G_HI_J, 0.3, TestEnum.ABC, 0.5, TestEnum.DEF, 0.2);
  private static final String ENUM_WEIGHT_STRING_2 = "[ABC : 0.7 , DEF : 0.3 ]";
  private static final Map<TestEnum, Double> ENUM_WEIGHT_MAP_2 =
      Map.of(TestEnum.ABC, 0.7, TestEnum.DEF, 0.3);
  private static final String ENUM_WEIGHT_STRING_BAD_WTS = "[ABC : 0.6 , DEF : 0.3 ]";
  private static final String ENUM_WEIGHT_STRING_BAD_NUM = "[ABC : nul , DEF : 0.3 ]";
  private static final String ENUM_WEIGHT_STRING_BAD_ENTRY = "[ABC : nul , DEF ]";

  private static final String VALUE_VALUE_WEIGHT_MAP_STRING =
      " [6.5 :: [1.0 : 0.8, 5.0 : 0.2]; 10.0 :: [1.0 : 0.2, 5.0 : 0.8]] ";
  private static final String VALUE_VALUE_WEIGHT_MAP_STRING_BAD_WTS =
      " [6.5 :: [1.0 : 0.8, 5.0 : 0.2]; 10.0 :: [1.0 : 0.2, 5.0 : 0.9]] ";
  private static final String VALUE_VALUE_WEIGHT_MAP_STRING_BAD_NUM =
      " [6.5 :: [1.0 : 0.8, 5.0 : 0.2]; TEN :: [1.0 : 0.2, 5.0 : 0.9]] ";
  private static final String VALUE_VALUE_WEIGHT_MAP_STRING_BAD_ENTRY =
      " [6.5 :: [1.0 : 0.8, 5.0 : 0.2]; 10.0 :: 3.0] ";
  private static final NavigableMap<Double, Map<Double, Double>> VALUE_VALUE_WEIGHT_MAP;// =

  static {
    // Can this be constructed in one line above?
    // Is there an equivalent to `Map.copyOf(x)` to make a copy of the
    // maps when assigning to the static final variable?
    NavigableMap<Double, Map<Double, Double>> x = new TreeMap<Double, Map<Double, Double>>();
    NavigableMap<Double, Double> y = new TreeMap<Double, Double>();
    y.put(1.0, 0.8);
    y.put(5.0, 0.2);
    x.put(6.5, Map.copyOf(y));

    NavigableMap<Double, Double> z = new TreeMap<Double, Double>();
    z.put(1.0, 0.2);
    z.put(5.0, 0.8);
    x.put(10.0, Map.copyOf(z));
    VALUE_VALUE_WEIGHT_MAP = x;
    // x.remove(6.5); // this makes test fail
  }

  @Test
  final void testEnumValueMapToString() {
    /*
     * Parsing.enumValueMapToString() appears to return the string sorted by
     * keys (insertion order not preserved) and without spaces around
     * delimiters, but this is FRAGILE!
     */
    assertEquals(ENUM_WEIGHT_STRING_1.replaceAll(" ", ""),
        Parsing.enumValueMapToString(ENUM_WEIGHT_MAP_1));
    assertEquals(ENUM_WEIGHT_STRING_2.replaceAll(" ", ""),
        Parsing.enumValueMapToString(ENUM_WEIGHT_MAP_2));
  }

  @Test
  final void testStringToEnumWeightMap() {
    assertEquals(ENUM_WEIGHT_MAP_1,
        Parsing.stringToEnumWeightMap(ENUM_WEIGHT_STRING_1, TestEnum.class));
    assertEquals(ENUM_WEIGHT_MAP_2,
        Parsing.stringToEnumWeightMap(ENUM_WEIGHT_STRING_2, TestEnum.class));

    assertThrows(IllegalArgumentException.class, () -> {
      Parsing.stringToEnumWeightMap(ENUM_WEIGHT_STRING_BAD_WTS, TestEnum.class);
    });
    assertThrows(NumberFormatException.class, () -> {
      Parsing.stringToEnumWeightMap(ENUM_WEIGHT_STRING_BAD_NUM, TestEnum.class);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Parsing.stringToEnumWeightMap(ENUM_WEIGHT_STRING_BAD_ENTRY, TestEnum.class);
    });
  }

  @Test
  final void testStringToValueValueWeightMap() {
    assertEquals(VALUE_VALUE_WEIGHT_MAP,
        Parsing.stringToValueValueWeightMap(VALUE_VALUE_WEIGHT_MAP_STRING));

    assertThrows(IllegalArgumentException.class, () -> {
      Parsing.stringToValueValueWeightMap(VALUE_VALUE_WEIGHT_MAP_STRING_BAD_WTS);
    });
    assertThrows(NumberFormatException.class, () -> {
      Parsing.stringToValueValueWeightMap(VALUE_VALUE_WEIGHT_MAP_STRING_BAD_NUM);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Parsing.stringToValueValueWeightMap(VALUE_VALUE_WEIGHT_MAP_STRING_BAD_ENTRY);
    });

  }

  @Test
  final void testSplitsAndJoin() {
    String str = "Test string";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.SPACE), Parsing.Delimiter.SPACE));

    str = "test";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.SPACE), Parsing.Delimiter.SPACE));

    // duplicate delimiters do not produce an empty field
    str = "test_double__underscore_delimited";
    assertNotEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.UNDERSCORE),
            Parsing.Delimiter.UNDERSCORE));
    str = "Test string 1 with   extra spaces";
    assertNotEquals(str, Parsing
        .join(Parsing.split(str, Parsing.Delimiter.SPACE), Parsing.Delimiter.SPACE));

    str = "test string with μ 𝛽 θν° unicode";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.SPACE), Parsing.Delimiter.SPACE));

    str = "test:colon:delimited";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.COLON), Parsing.Delimiter.COLON));

    str = "test,comma,delimited";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.COMMA), Parsing.Delimiter.COMMA));

    str = "test-dash-delimited";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.DASH), Parsing.Delimiter.DASH));

    str = "test.period.delimited";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.PERIOD), Parsing.Delimiter.PERIOD));

    str = "test/slash/delimited";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.SLASH), Parsing.Delimiter.SLASH));

    str = "test_underscore_delimited";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.UNDERSCORE),
            Parsing.Delimiter.UNDERSCORE));

    str = "1 2 3 4";
    assertEquals(str,
        Parsing.join(Parsing.split(str, Parsing.Delimiter.SPACE), Parsing.Delimiter.SPACE));
    assertEquals(str,
        Parsing.join(Parsing.splitToList(str, Parsing.Delimiter.SPACE), Parsing.Delimiter.SPACE));
    assertNotEquals(str,
        Parsing.join(Parsing.splitToDoubleList(str, Parsing.Delimiter.SPACE),
            Parsing.Delimiter.SPACE));

    str = "1.0 2.0 3.0 4.0";
    assertEquals(str,
        Parsing.join(Parsing.splitToDoubleList(str, Parsing.Delimiter.SPACE),
            Parsing.Delimiter.SPACE));
    str = "1.1 2.02 3.3 4.0";
    assertEquals(str,
        Parsing.join(Parsing.splitToDoubleList(str, Parsing.Delimiter.SPACE),
            Parsing.Delimiter.SPACE));
    // trailing zeros are not reproduced
    str = "1.10 2.02 3.3 4.0";
    assertNotEquals(str,
        Parsing.join(Parsing.splitToDoubleList(str, Parsing.Delimiter.SPACE),
            Parsing.Delimiter.SPACE));

    // doubles below 1e-3 are returned in scientific notation, with capital E
    str = "1.0E-4 2.02 3.3 4.0";
    assertEquals(str,
        Parsing.join(Parsing.splitToDoubleList(str, Parsing.Delimiter.SPACE),
            Parsing.Delimiter.SPACE));
    str = "1.0E-3 2.02 3.3 4.0";
    assertNotEquals(str,
        Parsing.join(Parsing.splitToDoubleList(str, Parsing.Delimiter.SPACE),
            Parsing.Delimiter.SPACE));

    List<Double> listDoubles = List.of(1e-4, 2.02, 3.3);
    assertEquals(listDoubles, Parsing.splitToDoubleList(
        Parsing.join(listDoubles, Parsing.Delimiter.COLON), Parsing.Delimiter.COLON));
  }

  @Test
  final void testEnumsToString() {
    // System.out.println("Print enum as default: " + TestEnum.ABC);
    // System.out.println("Pring enum as name(): " + TestEnum.ABC.name());
    // System.out.println("Print enum toString(): " + TestEnum.ABC.toString());
    // System.out.println();

    // This should be more robust, create expected string with alternative
    // method?
    List<TestEnum> v = List.of(TestEnum.values());
    String enumString = "[ABC,DEF,G_HI_J]";
    assertEquals(enumString, Parsing.enumsToString(v, TestEnum.class));
  }

  @Test
  final void testEnumLabels() {
    assertEquals("abc", Parsing.enumLabelWithSpaces(TestEnum.ABC, false));
    assertEquals("Abc", Parsing.enumLabelWithSpaces(TestEnum.ABC, true));
    assertEquals("g hi j", Parsing.enumLabelWithSpaces(TestEnum.G_HI_J, false));
    assertEquals("G Hi J", Parsing.enumLabelWithSpaces(TestEnum.G_HI_J, true));

    assertEquals("abc", Parsing.enumLabelWithDashes(TestEnum.ABC, false));
    assertEquals("Abc", Parsing.enumLabelWithDashes(TestEnum.ABC, true));
    assertEquals("g-hi-j", Parsing.enumLabelWithDashes(TestEnum.G_HI_J, false));
    assertEquals("G-Hi-J", Parsing.enumLabelWithDashes(TestEnum.G_HI_J, true));
  }

  @Test
  final void testToDoubleArray() {
    double[] xArray = { 1.0, 2.0, 3.0 };
    String xString = "[1.0, 2.0, 3.0]";
    String xStringNoSpace = "[1.0,2.0,3.0]";
    String xStringCurly = " { 1.0, 2.0, 3.0 } ";
    String xStringBadNumber = "[1.0, string, 3.0]";

    assertArrayEquals(xArray, Parsing.toDoubleArray(xString));
    assertArrayEquals(xArray, Parsing.toDoubleArray(xStringNoSpace));
    assertArrayEquals(xArray, Parsing.toDoubleArray(xStringCurly));
    assertThrows(NumberFormatException.class, () -> {
      Parsing.toDoubleArray(xStringBadNumber);
    });
  }

  @Test
  final void testToStringMethods() {
    List<Double> doubles = List.of(1e-5, 2e-4, 3e-3, 0.0, 1.0001, 10.0);
    String fmt_f = "%f";
    String fmt_4f = "%.4f";
    String fmt_3e = "%.3e";
    String expected_f = "[0.000010,0.000200,0.003000,0.000000,1.000100,10.000000]";
    String expected_f0 = "[0.000010,0.000200,0.003000,0.0,1.000100,10.000000]";
    // String expected_4f = "[0.0000,0.0002,0.0030,0.0000,1.0001,10.0000]";
    String expected_4f0 = "[0.0000,0.0002,0.0030,0.0,1.0001,10.0000]";
    // String expected_3e =
    // "[1.000e-05,2.000e-04,3.000e-03,0.000e+00,1.000e+00,1.000e+01]";
    String expected_3e0 = "[1.000e-05,2.000e-04,3.000e-03,0.0,1.000e+00,1.000e+01]";

    assertEquals(expected_f0, Parsing.toString(doubles, fmt_f));
    assertEquals(expected_4f0, Parsing.toString(doubles, fmt_4f));
    assertEquals(expected_3e0, Parsing.toString(doubles, fmt_3e));

    assertEquals(expected_f, Parsing.toString(doubles, fmt_f, ",", true, false));

    assertEquals(Parsing.trimEnds(expected_f), Parsing.toString(doubles, fmt_f, ",", false, false));
    assertEquals(expected_f.replaceAll(",", "|"),
        Parsing.toString(doubles, fmt_f, "|", true, false));
    assertEquals(expected_f0,
        Parsing.toString(doubles.stream().mapToDouble(Double::doubleValue), fmt_f, ",", true,
            true));
  }

  @Test
  final void testIntListToRangeString() {
    List<Integer> values = List.of(1, 2, 3, 4, 10, 19, 18, 17, 16);
    String valuesToString = "[1, 2, 3, 4, 10, 19, 18, 17, 16]";
    String valuesRangeString = "1:4,10,19:16";

    assertEquals(valuesToString, values.toString());
    assertEquals(valuesRangeString, Parsing.intListToRangeString(values));

    assertEquals(valuesRangeString,
        Parsing.intListToRangeString(Parsing.rangeStringToIntList(valuesRangeString)));
    assertEquals(values,
        Parsing.rangeStringToIntList(Parsing.intListToRangeString(values)));

    assertEquals("1", Parsing.intListToRangeString(List.of(1)));
    assertEquals("1:2", Parsing.intListToRangeString(List.of(1, 2)));
    assertEquals("1,3", Parsing.intListToRangeString(List.of(1, 3)));

    // Throw exception if list is empty
    assertThrows(IllegalArgumentException.class, () -> {
      Parsing.intListToRangeString(List.of());
    });

  }

  @Test
  final void testStripZeros() {
    assertEquals("10.0", Parsing.stripZeros("10.00"));
    assertEquals("1.0", Parsing.stripZeros("1.00"));
    assertEquals("1.01", Parsing.stripZeros("1.0100"));
    assertEquals("1.00e-01", Parsing.stripZeros(String.format("%.2e", 0.1)));
    assertEquals("1.00E-01", Parsing.stripZeros(String.format("%.2E", 0.1)));
    assertEquals("1.00e+00", Parsing.stripZeros(String.format("%.2e", 1.0)));
    assertEquals("1.00E+00", Parsing.stripZeros(String.format("%.2E", 1.0)));
    assertEquals("1.00e+10", Parsing.stripZeros(String.format("%.2e", 1e+10)));
  }

  @Test
  final void testTrimEnds() {
    assertEquals("X", Parsing.trimEnds("   [X] "));
    assertThrows(IllegalArgumentException.class, () -> {
      Parsing.trimEnds("  X  ");
    });
  }

  @Test
  final void testQueryMethods() {
    Map<String, String> params = Map.of("key1", "val1", "key2", "val2");
    assertEquals(params, Parsing.queryToMap(Parsing.mapToQuery(params)));
    // Do we need to test special characters? (HTML codes?)
  }

  @Test
  final void testReadInt() {
    // Parsing.readInt() is not used in code base
    String string = "0 1      2 3    4  5";
    for (int i = 0; i <= 5; i++) {
      assertEquals(i, Parsing.readInt(string, i));
    }
    assertThrows(IndexOutOfBoundsException.class, () -> {
      Parsing.readInt(string, 10);
    });
  }

  @Test
  final void testReadDouble() {
    String string = "0.0   1.0  2.0 3.0  4   5.0\n";
    for (double i = 0.0; i <= 5.0; i++) {
      assertEquals(i, Parsing.readDouble(string, (int) i));
    }
    assertThrows(IndexOutOfBoundsException.class, () -> {
      Parsing.readDouble(string, 10);
    });
  }

  @Test
  final void testStripComment() {
    // Parsing.stripComment() is not used in code base
    String base_string = "base string";
    String the_comment = "a comment";
    String string_with_comment = base_string + "# " + the_comment;
    String string_with_space_comment = base_string + " # " + the_comment;
    assertEquals(base_string, Parsing.stripComment(base_string, '#'));
    assertEquals(base_string, Parsing.stripComment(string_with_comment, '#'));
    // Parsing.stripComment() does not trim space at the end of the cleaned
    // string
    assertNotEquals(base_string, Parsing.stripComment(string_with_space_comment, '#'));
  }

  @Test
  final void testToLineList() {
    // Parsing.toLineList() is not used in code base
    List<String> lineList = List.of("line 1", "line 2", "line 3");
    List<String> longerList = new ArrayList<>(lineList);
    Iterator<String> it = lineList.iterator();
    // read all elements from iterator
    assertEquals(lineList, Parsing.toLineList(it, 3));
    // no more elements if iterator is read again
    assertThrows(NoSuchElementException.class, () -> {
      Parsing.toLineList(it, 1);
    });

    longerList.add("line4");
    Iterator<String> itl = longerList.iterator();
    assertEquals(lineList, Parsing.toLineList(itl, 3));

    assertThrows(NoSuchElementException.class, () -> {
      Parsing.toLineList(itl, 5);
    });

  }

  @Test
  final void testFormatDoubleFunction() {
    Function<Double, String> fmtFunc = Parsing.formatDoubleFunction("%.3f");
    assertEquals("1.001", fmtFunc.apply(1.0011));
    assertEquals("0.0", fmtFunc.apply(0.00000000));

    // compact (clean) zero-values are only written if the original value is
    // zero, not if it is zero after formatting
    assertEquals("0.000", fmtFunc.apply(0.000000001));
  }
}
