package gov.usgs.earthquake.nshmp.mfd;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.util.Arrays;
import java.util.List;
import java.util.stream.DoubleStream;

import org.junit.jupiter.api.Test;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.GutenbergRichter;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.Single;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Type;

class MfdsTests {

  @Test
  void testCheckRate() {
    assertEquals(1.0, Mfds.checkRate(1.0));
    assertThrows(IllegalArgumentException.class, () -> {
      Mfds.checkRate(-1.0);
    });
  }

  @Test
  void testCheckValues() {
    assertThrows(IllegalArgumentException.class, () -> {
      Mfds.checkValues(DoubleStream.of(-5.0), DoubleStream.of(1.0));
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Mfds.checkValues(DoubleStream.of(5.0), DoubleStream.of(-1.0));
    });
    assertDoesNotThrow(() -> {
      Mfds.checkValues(DoubleStream.of(5.0), DoubleStream.of(1.0));
    });
  }

  @Test
  void testCheckMfdProperty() {
    assertThrows(IllegalArgumentException.class, () -> {
      Mfds.checkGaussianMagnitudeCount(0);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Mfds.checkGaussianMagnitudeCount(22);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Mfds.checkGaussianSigma(0.0);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Mfds.checkGaussianSigma(0.51);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Mfds.checkGaussianSigmaCount(-1);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      Mfds.checkGaussianSigmaCount(6);
    });
  }

  @Test
  void testPropsToString() {
    assertEquals(
        Type.SINGLE.name() + " {dummy}",
        Mfds.propsToString(Type.SINGLE, "dummy"));
  }

  private static final double[] COMBINE_M = {
      5.05, 5.15, 5.25, 5.35, 5.45, 5.55, 5.65, 5.75, 5.85, 5.95, // GR
      6.05, 6.15, 6.25, 6.35, 6.45, 6.55, 6.65, 6.75, // GR
      6.76, 6.808, // GAUSS
      6.85, // GR
      6.856, 6.904,
      6.95, // GR
      6.952, 6.96,
      7.0, // GR GAUSS
      7.008, 7.048, // GAUSS
      7.05, // GR
      7.056, 7.096, 7.104, 7.144, // GAUSS
      7.15, // GR
      7.152, 7.192, 7.2, 7.24, 7.248, // GAUSS
      7.25, // GR
      7.296, 7.344, // GAUSS
      7.35, // GR
      7.392, 7.44, // GAUSS
      7.45 }; // GR

  private static final double[] COMBINE_R = {
      0.9803760319471205,
      0.7787403628225511,
      0.6185754577093839,
      0.4913519513660598,
      0.39029472815693284,
      0.31002212243909005,
      0.24625932524251715,
      0.1956107351042815,
      0.15537912990850308,
      0.12342202997321594,
      0.09803760319471205,
      0.0778740362822551,
      0.0618575457709384,
      0.04913519513660598,
      0.039029472815693286,
      0.031002212243909005,
      0.024625932524251715,
      0.019561073510428153,
      4.438109698488586E-5,
      9.117799957055695E-5,
      0.015537912990850309,
      1.5962281648018392E-4,
      2.3812925993219852E-4,
      0.012342202997321593,
      3.0272161935547244E-4,
      2.219054849244293E-5,
      3.2793441535340477E-4,
      4.5588999785278476E-5,
      3.0272161935547244E-4,
      0.009803760319471205,
      7.981140824009196E-5,
      2.3812925993219852E-4,
      1.1906462996609926E-4,
      1.5962281648018392E-4,
      0.007787403628225511,
      1.5136080967773622E-4,
      9.117799957055695E-5,
      1.6396720767670238E-4,
      4.438109698488586E-5,
      1.5136080967773622E-4,
      5.623413251903491E-4,
      1.1906462996609926E-4,
      7.981140824009196E-5,
      4.466835921509635E-4,
      4.5588999785278476E-5,
      2.219054849244293E-5,
      3.548133892335753E-4 };

  @Test
  void testCombine() {
    Mfd mfd1 = new Single(7.0, 0.002).toGaussianBuilder(11, 0.12, 2).build();
    Mfd mfd2 = new Single(7.2, 0.001).toGaussianBuilder(11, 0.12, 2).build();
    Mfd mfd3 = new GutenbergRichter(5.0, 1.0, 0.1, 5.0, 7.2).toBuilder().build();
    Mfd mfd4 = new GutenbergRichter(4.0, 1.0, 0.1, 5.0, 7.5).toBuilder().build();
    XySequence actual = Mfds.combine(List.of(mfd1, mfd2, mfd3, mfd4)).data();
    XySequence expected = XySequence.create(COMBINE_M, COMBINE_R);
    assertEquals(expected, actual);
  }

  private static final double[] GR_RATE_M = {
      5.05, 5.15, 5.25, 5.35, 5.45,
      5.55, 5.65, 5.75, 5.85, 5.95 };

  private static final double[] GR_RATE_R = {
      0.009120108393559097,
      0.007585775750291836,
      0.00630957344480193,
      0.005248074602497723,
      0.004365158322401657,
      0.00363078054770101,
      0.003019951720402013,
      0.002511886431509577,
      0.0020892961308540407,
      0.0017378008287493728 };

  @Test
  void testGutenbergRichterRate() {
    double a = 2;
    double b = 0.8;
    double[] actual = Arrays.stream(GR_RATE_M)
        .map(m -> Mfds.gutenbergRichterRate(a, b, m))
        .toArray();
    assertArrayEquals(GR_RATE_R, actual);
  }

  private static final double COMBINE_MO_RATE = 4.2389375463986755E18;

  @Test
  void testMomentRate() {
    Mfd mfd = Mfd.create(XySequence.create(COMBINE_M, COMBINE_R));
    assertEquals(COMBINE_MO_RATE, Mfds.momentRate(mfd));
  }

  private static final double PARETO_RATE = 3.6159044829897224E-5;

  @Test
  void testParetoRate() {
    double Mt = Earthquakes.magToMoment(4.5);
    double M = Earthquakes.magToMoment(7.0);
    double Mcm = Earthquakes.magToMoment(6.5);
    double b = 0.8;
    double β = b / 1.5;
    double actual = Mfds.paretoRate(Mt, M, β, Mcm);
    assertEquals(PARETO_RATE, actual);
  }
}
