package gov.usgs.earthquake.nshmp.model;

import static java.lang.Math.abs;
import static java.util.stream.Collectors.toMap;
import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Logger;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.MethodSource;

import com.google.common.util.concurrent.MoreExecutors;

import gov.usgs.earthquake.nshmp.calc.CalcConfig;
import gov.usgs.earthquake.nshmp.calc.Hazard;
import gov.usgs.earthquake.nshmp.calc.HazardCalcs;
import gov.usgs.earthquake.nshmp.calc.HazardExport;
import gov.usgs.earthquake.nshmp.calc.Site;
import gov.usgs.earthquake.nshmp.calc.Sites;
import gov.usgs.earthquake.nshmp.calc.ThreadCount;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Imt;

class LoaderTests {

  /*
   * These tests ensure that the basic model structure is in place. Model
   * components (e.g. model-config.json, gmm-tree.json) are tested by their
   * respective deserializing classes.
   */

  private static final double TOLERANCE = 1e-8;

  static final Path MODEL_PATH = Paths.get("src/test/resources/model/test-model");
  static final Path RESULTS_PATH = MODEL_PATH.resolve("results");
  static final Path SITES_PATH = MODEL_PATH.resolve("test-sites.csv");

  static HazardModel model;
  static List<Site> sites;
  static Map<Location, Map<Imt, XySequence>> expecteds;
  static ExecutorService exec;

  @BeforeAll
  static void setUpBeforeClass() throws IOException {
    model = ModelLoader.load(MODEL_PATH);
    sites = Sites.fromCsv(SITES_PATH, model.config(), model.siteData());
    expecteds = loadExpecteds();
    exec = initExecutor(model.config().performance.threadCount);
  }

  @Test
  void emptyTestEclipseBug() {
    /*
     * Silly workaround for junt5 tests in eclipse where no tests are found
     * unless at least one @Test annotation exists. Doesn't fix fact that one
     * test below cannot be run on its own.
     */
  }

  @ParameterizedTest
  @MethodSource("siteStream")
  final void testLocation2(Site site) {
    Location loc = site.location();
    Map<Imt, XySequence> expected = expecteds.get(loc);
    Hazard actual = HazardCalcs.hazard(model, model.config(), site, exec);
    assertCurvesEqual(expected, actual, TOLERANCE);
  }

  private static Stream<Site> siteStream() {
    return sites.stream();
  }

  private static void assertCurvesEqual(Map<Imt, XySequence> expected, Hazard actual, double tol) {
    expected.entrySet().forEach(
        e -> assertCurveEquals(
            e.getValue(),
            actual.curves().get(e.getKey()),
            tol));
  }

  private static void assertCurveEquals(XySequence expected, XySequence actual, double tol) {

    // IMLs close but not exact due to exp() transform
    assertArrayEquals(
        expected.xValues().toArray(),
        actual.xValues().map(Math::exp).toArray(),
        tol);

    double[] expectedYs = expected.yValues().toArray();
    double[] actualYs = actual.yValues().toArray();

    // y-value difference relative to tolerance
    assertArrayEquals(expectedYs, actualYs, tol);

    // y-value difference relative to tolerance
    for (int i = 0; i < expectedYs.length; i++) {
      String message = String.format(
          "arrays differ at [%s] expected:<[%s]> but was:<[%s]>",
          i, expectedYs[i], actualYs[i]);
      assertTrue(compare(expectedYs[i], actualYs[i], tol), message);
    }
  }

  private static boolean compare(double expected, double actual, double tolerance) {
    return abs(actual - expected) / expected < tolerance ||
        Double.valueOf(expected).equals(Double.valueOf(actual));
  }

  /* read curves then transpose */
  private static Map<Location, Map<Imt, XySequence>> loadExpecteds() throws IOException {
    // consider centralized curve file processing
    Map<Imt, Map<Location, XySequence>> curves = Files.walk(RESULTS_PATH)
        .filter(LoaderTests::isCurveFile)
        .collect(toMap(
            LoaderTests::imtFromPath,
            LoaderTests::readCurves));
    return transpose(curves);
  }

  private static boolean isCurveFile(Path path) {
    return path.getFileName().toString().equals("curves.csv");
  }

  private static Imt imtFromPath(Path path) {
    return Imt.valueOf(path.getParent().getFileName().toString());
  }

  private static Map<Location, XySequence> readCurves(Path path) {
    try {
      List<String> lines = Files.readAllLines(path);
      double[] imls = toValues(lines.get(0));
      return lines.stream()
          .skip(1)
          .map(line -> toCurve(line, imls))
          .collect(toMap(
              Entry::getKey,
              Entry::getValue));
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  private static Entry<Location, XySequence> toCurve(String line, double[] xs) {
    Site site = toSite(line);
    double[] ys = toValues(line);
    return Map.entry(site.location(), XySequence.create(xs, ys));
  }

  private static Site toSite(String line) {
    String[] s = Arrays.stream(line.split(","))
        .map(String::trim)
        .limit(3)
        .toArray(String[]::new);
    String name = s[0];
    Location loc = Location.create(
        Double.valueOf(s[1]),
        Double.valueOf(s[2]));
    return Site.builder()
        .name(name)
        .location(loc)
        .build();
  }

  private static double[] toValues(String line) {
    return Arrays.stream(line.split(","))
        .map(String::trim)
        .skip(3)
        .mapToDouble(Double::valueOf)
        .toArray();
  }

  private static Map<Location, Map<Imt, XySequence>> transpose(
      Map<Imt, Map<Location, XySequence>> mapIn) {

    Map<Location, Map<Imt, XySequence>> mapOut = new HashMap<>();
    for (Entry<Imt, Map<Location, XySequence>> imtEntry : mapIn.entrySet()) {
      Imt imt = imtEntry.getKey();
      for (Entry<Location, XySequence> locEntry : imtEntry.getValue().entrySet()) {
        Location loc = locEntry.getKey();
        XySequence xy = locEntry.getValue();
        mapOut.computeIfAbsent(loc, k -> new EnumMap<>(Imt.class)).put(imt, xy);
      }
    }
    return mapOut;
  }

  public static void main(String[] args) throws IOException {
    model = ModelLoader.load(MODEL_PATH);
    List<Site> sites = Sites.fromCsv(SITES_PATH, model.config(), model.siteData());
    calcWrite(model, model.config(), sites);
  }

  /* For creating/updating results */
  private static Path calcWrite(
      HazardModel model,
      CalcConfig config,
      List<Site> sites) throws IOException {

    ExecutorService exec = initExecutor(config.performance.threadCount);
    Logger log = Logger.getLogger(LoaderTests.class.getName());
    HazardExport handler = HazardExport.create(model, config, sites, log);
    for (Site site : sites) {
      Hazard hazard = HazardCalcs.hazard(model, config, site, exec);
      handler.write(hazard);
    }
    handler.expire();
    exec.shutdown();
    return handler.outputDir();
  }

  private static ExecutorService initExecutor(ThreadCount threadCount) {
    return (threadCount == ThreadCount.ONE)
        ? MoreExecutors.newDirectExecutorService()
        : Executors.newFixedThreadPool(threadCount.value());
  }

}
